/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
//getcwd:
#ifdef WIN32	//MSVC only
#include <direct.h>
#else
#include <sys/unistd.h>
#endif

#include "system/system.h"
#include "system/singletons.h"
#include "system/exception.h"
#include "base/base.h"

#include "inspiration/world.h"
#include "inspiration/spawnobj.h"

#include "impl.h"
#include "wndresize.h"
#include "func.h"
#include "menulib.h"

using namespace std;

class EditorFuncImpl {
public:
	EditorImpl *ed;
	
	EditorFuncImpl(EditorImpl *ed_);
	
	void newWorld();
	void loadWorld();
	void saveWorld();
	void saveWorldAs();
	void resize();
	void showWindow(int wnd, int show);	//0 = hide, 1 = show, 2 = toggle

protected:
	bool doSaveWorldDialog();
	bool doOpenWorldDialog();

	//char editor_filename[256];
	char editor_full_filename[256];
};

EditorFuncImpl::EditorFuncImpl(EditorImpl *ed_) : ed(ed_) {
	//memset(editor_filename, 0, sizeof(editor_filename));
	memset(editor_full_filename, 0, sizeof(editor_full_filename));
}


bool EditorFuncImpl::doOpenWorldDialog() {
	string filename;
	if (!dialog.open("world", filename)) return false;
	strncpy(editor_full_filename, filename.c_str(), sizeof(editor_full_filename));
	return true;
}

bool EditorFuncImpl::doSaveWorldDialog() {
	string filename;
	if (!dialog.save("world", filename)) return false;
	strncpy(editor_full_filename, filename.c_str(), sizeof(editor_full_filename));
	return true;
}

void EditorFuncImpl::newWorld() {
/*
this will wipe our textures as well
which makes importing textures a pain
it'd be nice to either
(a) only clear ents and reset tiles and tags to a default
(b) create texture and tag templates separately and import them per world
	.... like a "this is your cave level template" and "this is your bridge level template"
*/	
	worldSys.newWorld();
	//..then assert by the end of it we're now attached to the same world
	//in fact, we should be making this assertion for loadWorld and saveWorld too ...
	
	//instead just clear spawnobjs and tiles?
	//spawnObjSys.newSystem();
	
	//or how about loading all in the tex search paths into the map texs by default?
	
}

void EditorFuncImpl::loadWorld() {

	//store the current dir
	char cwd[256];
	getcwd(cwd, sizeof(cwd));
	bool success = doOpenWorldDialog();
	chdir(cwd);

	if (success) {
		if (!worldSys.loadWorld(editor_full_filename)) {
			//TODO - popup?
			con << "unable to load to file " << editor_full_filename << "\n";
		}
	}
}

void EditorFuncImpl::saveWorldAs() {
	//store the current dir
	char cwd[256];
	getcwd(cwd, sizeof(cwd));
	bool success = doSaveWorldDialog();
	chdir(cwd);
	 
	if (success) {
		if (!worldSys.saveWorldAs(editor_full_filename)) {
			//TODO - popup?
			con << "unable to save to file " << editor_full_filename << "\n";
		}
	}
}

void EditorFuncImpl::saveWorld() {

	if (!worldSys.saveWorld()) {
		//TODO - popup?
		con << "unable to save to world\n";
	}
}

//void EditorFunc_Undo() {
//	if (landscapeHistory.Undo()) {
//		con.print("undo success\n");
//	} else {
//		con.print("undo failed\n");
//	}
//	editorWndTile.refresh();
//}
//
//void EditorFunc_Redo() {
//	if (landscapeHistory.Redo()) {
//		con.print("redo success\n");
//	} else {
//		con.print("redo failed\n");
//	}
//	editorWndTile.refresh();
//}

//very very parallel to above
void EditorFuncImpl::showWindow(int wnd, int show) {
	if (wnd < 0 || wnd >= EW_COUNT) return;
	MenuEditorWindow *hwnd = (*editor->editWnd[wnd]);
	if (!hwnd) return;

	if (show < 0 || show > 1) {
		hwnd->setVisible(!hwnd->getVisible());
	} else {
		hwnd->setVisible(show != 0);
	}
}

//make a menu alternative for this:

void EditorFuncImpl::resize() {
	//if we get a nonzero result then resize the landscape
	resize_params_t params;
	params.width = ed->world->getMapSize().x;
	params.height = ed->world->getMapSize().y;
	params.layers = ed->world->getMapSize().z;
	params.tag_tile_width = ed->world->tags->tagsize.x;
	params.tag_tile_height = ed->world->tags->tagsize.y;

	ed->resizeWnd.resize(&params);

}

EditorFunc::EditorFunc(EditorImpl *ed_) : impl(new EditorFuncImpl(ed_)) {}
EditorFunc::~EditorFunc() { delete impl; }
void EditorFunc::newWorld() { impl->newWorld(); }
void EditorFunc::loadWorld() { impl->loadWorld(); }
void EditorFunc::saveWorld() { impl->saveWorld(); }
void EditorFunc::saveWorldAs() { impl->saveWorldAs(); }
void EditorFunc::resize() { impl->resize(); }
void EditorFunc::showWindow(int wnd, int show) { impl->showWindow(wnd, show); }
