/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <math.h>

#include "system/gl.h"
#include "system/singletons.h"
#include "system/memory.h"
#include "base/base.h"

#include "inspiration/world.h"
#include "inspiration/spawnobj.h"
#include "inspiration/main.h"	//system_deltatime

#include "impl.h"
#include "wndtile.h"
#include "wndent.h"
#include "wndtag.h"
#include "input.h"

using namespace std;

class EditorInputImpl {
protected:
	vec2f editor_mouse, editor_mouse_last;

	//state inputs
	void editTags();
	void surfacePaint();
	void surfaceFill();
	void surfaceMove();
	void entCreate();
	void entSelect();

	//aux functions
	bool fillTile(int x, int y, int key, int init_key);
	void fillColumn(int x, int y, int key, int init_key);
	void fillSurface(int x, int y, int key);

	float snapCoord(float x);
	box2f getSpawnObjBBox(SpawnObj *id);
	SpawnObj *findSpawnObjIntersect();
	
public:
	EditorImpl *ed;
	
	EditorInputImpl(EditorImpl *ed_);
	
	void getInput();
	void updateMouse(const vec2f &v);
};

EditorInputImpl::EditorInputImpl(EditorImpl *ed_) : ed(ed_) {}

float EditorInputImpl::snapCoord(float x) {
	if (!editor->snapEnabled) return x;

	float div = x / editor->snapDist;
	float fl = (float)floor(div);
	float ce = fl + 1;

	return editor->snapDist * (div - fl < ce - div ? fl : ce);
}

box2f EditorInputImpl::getSpawnObjBBox(SpawnObj *id) {
	assert(id);
	box2f bbox = id->getTypeBBox();
	id->rw("bbox") >> bbox;
	return bbox + editor->getSpawnObjPos(id);
}

SpawnObj *EditorInputImpl::findSpawnObjIntersect() {
	SpawnObj *bestobj = NULL;

	for (int i = 0; i < spawnObjSys.count(); i++) {
		SpawnObj *s = spawnObjSys[i];
		assert(s);
		box2f so_bbox = getSpawnObjBBox(s);
		if (so_bbox.contains(editor_mouse)) {
			bestobj = s;
		}
	}
	return bestobj;
}

bool EditorInputImpl::fillTile(int x, int y, int key, int init_key) {
	if (x < 0 || x >= ed->world->getMapSize().x ||
		y < 0 || y >= ed->world->getMapSize().y ||
		editor->layer < 0 || editor->layer >= ed->world->getMapSize().z)
	{
		return false;
	}
	
	//if all the painted values match what we initially found then go through with it!
	if (ed->world->keys->getIndex(ed->world->getBackupMap()->riskyGet(x,y,editor->layer)) != init_key) return false;	//already got it here
	//do the painting
	ed->world->getMap()->rw(x,y,editor->layer) = ed->world->keys->safeGet(key);
	ed->world->getBackupMap()->rw(x,y,editor->layer) = ed->world->keys->safeGet(key);
	return true;
}

//prereq: the <x,y> tile of the column is filled
//does: fills all to the left and right with matching init_* properties
//fills all above & below
void EditorInputImpl::fillColumn(int x, int y, int key, int init_key) {
	//column boundaries, inclusive
	int left_bound, right_bound;

	//seek left
	for (left_bound = x-1; left_bound >= 0; left_bound--) {
		if (!fillTile(left_bound, y, key, init_key)) break;
	}
	left_bound++;

	//seek right
	for (right_bound = x+1; right_bound < ed->world->getMapSize().x; right_bound++) {
		if (!fillTile(right_bound, y, key, init_key)) break;
	}
	right_bound--;

	//seek across, filling below
	if (y > 0) {
		for (x = left_bound; x <= right_bound; x++) {
			//if we can fill the column
			if (fillTile(x, y-1, key, init_key)) {
				//then do so
				fillColumn(x, y-1, key, init_key);
			}
		}
	}
	//and filling above...
	if (y < ed->world->getMapSize().y-1) {
		for (x = left_bound; x <= right_bound; x++) {
			//if we can fill the column
			if (fillTile(x, y+1, key, init_key)) {
				//then do so
				fillColumn(x,y+1, key, init_key);
			}
		}
	}
}

void EditorInputImpl::fillSurface(int x, int y, int key) {
	if (x < 0 || x >= ed->world->getMapSize().x ||
		y < 0 || y >= ed->world->getMapSize().y ||
		editor->layer < 0 || editor->layer >= ed->world->getMapSize().z) return;
		
	int init_key = ed->world->keys->getIndex(ed->world->getBackupMap()->riskyGet(x,y, editor->layer));//grab the initial contents
	if (init_key == key) return;				//if we already got it here then return
	ed->world->getMap()->rw(x,y,editor->layer) = ed->world->keys->safeGet(key);			//fill that initial tile
	ed->world->getBackupMap()->rw(x,y,editor->layer) = ed->world->keys->safeGet(key);			//fill that initial tile
	fillColumn(x, y, key, init_key);		//and follow suit with the rest of the column
}



void EditorInputImpl::editTags() {
	vec2i tile(
		floor_ftoi(editor_mouse.x / (float)ed->world->tags->tagsize.x),
		floor_ftoi(editor_mouse.y / (float)ed->world->tags->tagsize.y));
	if (tile.x < 0 || tile.x >= ed->world->tags->getSize().x || tile.y < 0 || tile.y >= ed->world->tags->getSize().y) return;
	if (mouse.getButtonState(MouseHandler::BUTTON_LEFT) == BUTTON_STATE_RELEASE) {
		if (kbd.keystate[Key::LSHIFT] == BUTTON_STATE_DOWN ||
			kbd.keystate[Key::RSHIFT] == BUTTON_STATE_DOWN)
		{
			int newkey = (*ed->world->tags)(VEC2ELEM(tile));
			if (newkey != editor->tagKey) {
				editor->tagKey = newkey;
				ed->tagWnd.refreshComboTagKeySelection();
			}
		} else {
			(*ed->world->tags)(VEC2ELEM(tile)) = editor->tagKey;
		}
	}
}


//currently its a messed up system
//if the user has a geom selected then nothing happens
//if the user has an image selected with no geom then nothing happens
//if the user has an image w/geom then we do the painting
void EditorInputImpl::surfacePaint() {
	int x = (int)editor_mouse.x;
	int y = (int)editor_mouse.y;
	if (mouse.getButtonState(MouseHandler::BUTTON_LEFT) == BUTTON_STATE_DOWN) {
		if (kbd.keystate[Key::LSHIFT] == BUTTON_STATE_DOWN ||
			kbd.keystate[Key::RSHIFT] == BUTTON_STATE_DOWN)
		{
			if (!(x < 0 || x >= ed->world->getMapSize().x ||
				y < 0 || y >= ed->world->getMapSize().y ||
				editor->layer < 0 || editor->layer >= ed->world->getMapSize().z)) 
			{
				editor->setMapKeySel(ed->world->keys->getIndex(ed->world->getBackupMap()->riskyGet(x, y, editor->layer)));
			}
		} else {
			int radius = editor->paintArea - 1;
			for (int j = y - radius; j <= y + radius; j++) {
				for (int i = x - radius; i <= x + radius; i++) {
					
					if (!(i < 0 || i >= ed->world->getMapSize().x ||
						j < 0 || j >= ed->world->getMapSize().y ||
						editor->layer < 0 || editor->layer >= ed->world->getMapSize().z)) 
					{
						//flow per second?  
						if (frand() < editor->paintDissolve) {
							ed->world->getMap()->rw(i, j, editor->layer) = ed->world->keys->safeGet(editor->mapKey);
							ed->world->getBackupMap()->rw(i, j, editor->layer) = ed->world->keys->safeGet(editor->mapKey);
						}
					}
				}
			}
		}
	}

	//show us the money!
	glDisable(GL_TEXTURE_2D);
	glColor3f(0,1,1);
	glLineWidth(2);
	glBegin(GL_LINE_LOOP);
	glVertex3f(x - editor->paintArea + 1, y - editor->paintArea + 1, editor->drawDepth);
	glVertex3f(x - editor->paintArea + 1, y + editor->paintArea, editor->drawDepth);
	glVertex3f(x + editor->paintArea, y + editor->paintArea, editor->drawDepth);
	glVertex3f(x + editor->paintArea, y - editor->paintArea + 1, editor->drawDepth);
	glEnd();
	glLineWidth(1);
	glEnable(GL_TEXTURE_2D);
}

void EditorInputImpl::surfaceFill() {
	int x = (int)editor_mouse.x;
	int y = (int)editor_mouse.y;
	if (mouse.getButtonState(MouseHandler::BUTTON_LEFT) == BUTTON_STATE_DOWN) {
		if (kbd.keystate[Key::LSHIFT] == BUTTON_STATE_DOWN ||
			kbd.keystate[Key::RSHIFT] == BUTTON_STATE_DOWN)
		{
			if (!(x < 0 || x >= ed->world->getMapSize().x ||
				y < 0 || y >= ed->world->getMapSize().y ||
				editor->layer < 0 || editor->layer >= ed->world->getMapSize().z)) 
			{
				editor->setMapKeySel(ed->world->keys->getIndex(ed->world->getBackupMap()->riskyGet(x, y, editor->layer)));
			}
			return;
		}
	} else {
		if (mouse.getButtonState(MouseHandler::BUTTON_LEFT) == BUTTON_STATE_RELEASE) {
			fillSurface(x, y, editor->mapKey);
		}
	}
}

void EditorInputImpl::surfaceMove() {
	static box2i box;
	static vec2i seldown;
	static vec2f down;	//needs to be float for box shifting
	static bool dragging = false;
	
	//if we get a mouse button press . . .
	if (mouse.getButtonState(MouseHandler::BUTTON_LEFT) == BUTTON_STATE_PRESS) {
		//by default
		dragging = false;
		
		//if its inside the box then denote "moving!"
		down = editor_mouse;	//keep track of where we are
		if (((box2f)box).contains(editor_mouse)) {	//make sure we convert the box to floats before comparing it
			dragging = true;
		}
		//if its outside the box then denote "selecting!"
		else {
			//round to the nearest for the selection box
			seldown.set(
				(int)floor(down.x + 0.5f),
				(int)floor(down.y + 0.5f));
			box.min = seldown;
			box.max = seldown;
		}
	}
	//if we are dragging the mouse
	else if (mouse.getButtonState(MouseHandler::BUTTON_LEFT) == BUTTON_STATE_DOWN) {
		//if we are in the moving state then 
		if (dragging) {
			//now here's a hack:
			vec2i &imin = box.min;
			vec2i &imax = box.max;
			
			vec2f fd = editor_mouse - down;
			vec2i d(fd.x < 0 ? -(int)(-fd.x) : (int)(fd.x),
					fd.y < 0 ? -(int)(-fd.y) : (int)(fd.y));
			vec2i size = box.size();
			//if we got greater than one block movement along this axis
			for (int i = 0; i < 2; i++) {
				if (d(i)) {
					int oldpos = imin(i);
					int newpos = oldpos + d(i);
					if (newpos < 0) newpos = 0;
					if (newpos + size(i) > ed->world->getMapSize()(i)) newpos = ed->world->getMapSize()(i) - size(i);
					d(i) = newpos - oldpos;		//correct our d.x (if we use it later then change this code)
					//make sure we have somewhere to move into & do the shift
					if (newpos != oldpos) {
						int absd = d(i) < 0 ? -d(i) : d(i);		//get abs width of the temp buffer
						constMapKeyPtr *tbuf = NULL;
						if (editor->movePreserve) {
							tbuf = new constMapKeyPtr [absd];	//make a temp buffer to gather up the old data into
							if (!tbuf) {
								editor->movePreserve = 0;
								con.print("FAILED TO ALLOCATE OUR MOVE PRESERVATION BUFFER!\n");
							}
						}
						
						//get our specs depending on our direction
						int start, end, step,src,dst;
						if (d(i) < 0) {
							start = imin(i); end = imax(i); step = 1; src = imin(i) + d(i); dst = imax(i) + d(i);
						} else {
							start = imax(i)-1; end = imin(i)-1; step = -1; src = imax(i); dst = imin(i);
						}
						
						Map *maps[] = {ed->world->getMap(), ed->world->getBackupMap()};
						for (Map **m = maps; m < endof(maps); m++) {
						
							if (i == 0) {	//x-movement code.  has memcpy hacks
								constMapKeyPtr *map_row = &(*m)->rw(0, imin.y, editor->layer);
									
								//copy it in:
								for (int y = imin.y; y < imax.y; y++, map_row += ed->world->getMapSize().x) {
									if (editor->movePreserve) {
										//copy the sshift region into the temp buffer
										memcpy(tbuf, map_row + src, absd * sizeof(constMapKeyPtr));
									}
									//now copy the old data across - left-to-right to prevent overwrites
									for (int x = start; x != end; x += step) {
										map_row[x + d.x] = map_row[x];
									}
									if (editor->movePreserve) {
										//and now copy the buffer contents across
										memcpy(map_row + dst, tbuf, absd * sizeof(constMapKeyPtr));
									}
								}
							} else if (i == 1) {	//y-movement code
								constMapKeyPtr *map_col = &(*m)->rw(imin.x, 0, editor->layer);
								
								//copy it in:
								for (int x = imin.x; x < imax.x; x++, map_col++) {
									//copy the sshift region into the temp buffer
									int y;
									if (editor->movePreserve) {
										for (y = 0; y < absd; y++) {
											tbuf[y] = map_col[(src + y) * ed->world->getMapSize().x];
										}
									}
									//now copy the old data across - left-to-right to prevent overwrites
									for (y = start; y != end; y += step) {
										map_col[(y + d.y) * ed->world->getMapSize().x] = map_col[y * ed->world->getMapSize().x];
									}
									//and now copy the buffer contents across
									if (editor->movePreserve) {
										for (y = 0; y < absd; y++) {
											map_col[(dst + y) * ed->world->getMapSize().x] = tbuf[y];
										}
									}
								}
							}
						}
		
						if (editor->movePreserve) {
							delete tbuf;	//free the temp buffer
						}
						//now that we've shifted it, also shift the box over as well
						box.min(i) += d(i);
						box.max(i) += d(i);
					}
									
					//reset this axis, in terms of dragging-to-move
					down(i) = editor_mouse(i);
				}
			}
		} else {
			//get the box that wraps the rounded mouse and its rounded down
			vec2i cur(
				(int)floor(editor_mouse.x + 0.5f),
				(int)floor(editor_mouse.y + 0.5f));
			box.set(cur);
			box.stretch(seldown);
		}
	} else if (mouse.getButtonState(MouseHandler::BUTTON_LEFT) == BUTTON_STATE_RELEASE) {
	}
	
	if (box.min != box.max) {
		//draw the rectangle / do a rectangle select
		//how about some line patterns?
		glDisable(GL_TEXTURE_2D);
		glColor3f(0,1,1);
		glLineWidth(2);
		glBegin(GL_LINE_LOOP);
		glVertex3f(box.min.x, box.min.y, editor->drawDepth);
		glVertex3f(box.min.x, box.max.y, editor->drawDepth);
		glVertex3f(box.max.x, box.max.y, editor->drawDepth);
		glVertex3f(box.max.x, box.min.y, editor->drawDepth);
		glEnd();
		glLineWidth(1);
		glEnable(GL_TEXTURE_2D);
	}
}

void EditorInputImpl::entCreate() {

	//delete any selected:
	if (kbd.keystate[Key::DEL] == BUTTON_STATE_PRESS ||
		kbd.keystate[Key::BACKSPACE] == BUTTON_STATE_PRESS)
	{
		editor->deleteEntity();
	}

	switch (mouse.getButtonState(MouseHandler::BUTTON_LEFT)) {
	case BUTTON_STATE_PRESS:		//find new vertex

		//change methods and deselect all
		//but what about the mouse down box / dragging issues?
//		editor->setMethod(EDITOR_ENT_SELECT);

		//mouse x and mouse y:

		//create a ent instance
		editor->spawnObj = spawnObjSys.createByName(editor->spawnType.c_str() );
		if (!editor->spawnObj) {
			con.print("failed to create spawn object\n");
			break;
		}
			
		editor->spawnObj->rw("pos") << vec2f(
			snapCoord(editor_mouse.x),
			snapCoord(editor_mouse.y));


		//deselect all other entities
		for (int i = 0; i < spawnObjSys.count(); i++) {
			SpawnObj *s = spawnObjSys[i];
			assert(s);
			s->sel = false;
		}
		//and select our new one ... ?
		// that makes it easy to change fields after creating them
		//but not selecting them means if the user changes the ent type
		// the newly created ent's type won't change as well
		//editor->spawnObj->sel = true;

		editorWndEnt.refreshSelected();
		
		//spawn something up front too
//dont bother
		if (!editor->spawnObj->createEnt()) {
			con << "EditorInputImpl::entCreate: spawnObjCreateEntity failed to create object\n";
		}
			
		break;
	}
}

void EditorInputImpl::entSelect() {

	//delete any selected:
	if (kbd.keystate[Key::DEL] == BUTTON_STATE_PRESS ||
		kbd.keystate[Key::BACKSPACE] == BUTTON_STATE_PRESS)
	{
		editor->deleteEntity();
	}

	//for left click + drag to select - this adds units without clearing selection
	//for right click + drag to rotate - this scales instead
	bool shiftMod = kbd.keystate[Key::LSHIFT] == BUTTON_STATE_DOWN || kbd.keystate[Key::RSHIFT] == BUTTON_STATE_DOWN;
	//for left click + drag to select - this subtracts units without clearing selection
	bool ctrlMod = kbd.keystate[Key::LCTRL] == BUTTON_STATE_DOWN || kbd.keystate[Key::RCTRL] == BUTTON_STATE_DOWN;
	
	//select/translate
	{
		static bool moving = false;	//moving = true <=> we are moving seelcted objects, false <=> we are rectangl-selecting objects
		static vec2f down;
		static vec2f drag_last;
		static vec2f dragDelta;

		//from mouse click to entity
		static bool dragged = false;

		switch (mouse.getButtonState(MouseHandler::BUTTON_LEFT)) {
		case BUTTON_STATE_PRESS:
			down = editor_mouse;

			editor->spawnObj = findSpawnObjIntersect();
			editorWndEnt.refreshSelected();
			dragged = false;
			if (editor->spawnObj) {

				//if we clicked down on an unselected entity
				//then assert we're not moving
				//and instead consider adding it to the selected set:
				if (!editor->spawnObj->sel) {
					//if we aren't adding the ent to the set
					if (!shiftMod && !ctrlMod) {
						//then instead remove all other ents
						for (int i = 0; i < spawnObjSys.count(); i++) {
							SpawnObj *s = spawnObjSys[i];
							assert(s);
							s->sel = false;
						}
					}
				}

				//select our clicked guy (or remove it if we're using the sub mod)
				editor->spawnObj->sel = !ctrlMod;

				moving = true;
				//drag_last = editor->getSpawnObjPos(spawnObj);
				drag_last.x = snapCoord(editor_mouse.x);
				drag_last.y = snapCoord(editor_mouse.y);
				dragDelta = 0.f;
			} else {
				moving = false;
			}
			break;

		case BUTTON_STATE_DOWN:		//drag the new vertex
			if (moving) {
				if (!!mouse.getMove()) {

					//snap the current selected entity to the grid
					//adjust all others accordingly

					vec2f newpos(snapCoord(editor_mouse.x), snapCoord(editor_mouse.y));
					//accum here
					dragDelta += newpos - drag_last;
					//dissipate later
					drag_last = newpos;
					
					
					vec2f offsetTheRest;
					if (editor->spawnObj) {	//assert?
						vec2f oldpos = editor->getSpawnObjPos(editor->spawnObj);
						vec2f newpos(
							snapCoord(oldpos.x + dragDelta.x),
							snapCoord(oldpos.y + dragDelta.y));
						
						//move all else by the snapped movement of the first entity
						offsetTheRest = newpos - oldpos;
						//dissipate delta vector
						dragDelta -= offsetTheRest;
						
						editor->spawnObj->rw("pos") << newpos;
					}

					//TODO - use a vector on mouseDown, and HERE only cycle through it 
					for (int i = 0; i < spawnObjSys.count(); i++) {
						SpawnObj *s = spawnObjSys[i];
						assert(s);
						if (!s->sel) continue;
						if (s == editor->spawnObj) continue;
						s->rw("pos") << (editor->getSpawnObjPos(s) + offsetTheRest);
					}
					
					//we only need to update pos if an ent is selected
					if (editor->spawnObj) editorWndEnt.refreshSelected();
				}
			} else {
				if (!!mouse.getMove()) {
					dragged = true;
				}
				
				box2f box(editor_mouse);
				box.stretch(down);
				
				//draw the rectangle / do a rectangle select
				glDisable(GL_TEXTURE_2D);
				glColor3f(0,1,0);
				glLineWidth(2);
				glBegin(GL_LINE_LOOP);
				glVertex3f(box.min.x, box.min.y, editor->drawDepth);
				glVertex3f(box.min.x, box.max.y, editor->drawDepth);
				glVertex3f(box.max.x, box.max.y, editor->drawDepth);
				glVertex3f(box.max.x, box.min.y, editor->drawDepth);
				glEnd();
				glLineWidth(1);
				glEnable(GL_TEXTURE_2D);
			}
			break;
		case BUTTON_STATE_RELEASE:
			if (!moving) {
			
				box2f box(editor_mouse);
				box.stretch(down);

				for (int i = 0; i < spawnObjSys.count(); i++) {
					SpawnObj *s = spawnObjSys[i];
					assert(s);

					//photoshop-style
					//shift to add
					//control to subtract

					//if no modifier keys then deselect all else
					if (!kbd.keystate[Key::LSHIFT] && 
						!kbd.keystate[Key::RSHIFT] && 
						!kbd.keystate[Key::LCTRL] &&
						!kbd.keystate[Key::RCTRL])
					{
						s->sel = false;
						if (s == editor->spawnObj) {
							editor->spawnObj = NULL;
						}
					}

					if (getSpawnObjBBox(s).touches(box)) {
						if (kbd.keystate[Key::LCTRL] || kbd.keystate[Key::RCTRL]) {
							s->sel = false;
						} else {
							s->sel = true;
						}
						editor->spawnObj = s;
					}
				}

				editor->spawnType = "";
				if (editor->spawnObj) editor->spawnObj->rw("classname") >> editor->spawnType;

				editorWndEnt.refreshSelected();
			}
			break;
		}
	}

	//rotate/scale
	{
		static bool moving = false;	//moving = true <=> we are moving seelcted objects, false <=> we are rectangl-selecting objects
		static vec2f down;
		static vec2f pivot;

		//from mouse click to entity
		static bool rotating = false;
		static bool scaling = false;

		switch (mouse.getButtonState(MouseHandler::BUTTON_RIGHT)) {
		case BUTTON_STATE_PRESS:
			if (editor->spawnObj) {
				editor->spawnObj->rw("pos") >> pivot;
				down = editor_mouse;
				
				rotating = false;
				scaling = false;
				if (shiftMod) {
					scaling = true;
				} else {
					rotating = true;	//can only rotate if we have at least one ent selected
				}
			}
			break;

		case BUTTON_STATE_DOWN:		//drag the new vertex
			if (rotating) {
				if (!!mouse.getMove()) {
					vec2f d1 = down - pivot;
					vec2f d2 = editor_mouse - pivot;
					down = editor_mouse;

					//calc rotation from [ent pos to down] to [ent pos to editor_mouse]
					//from 
					//apply that to our base entity.  calc ent rot from start.  apply delta to all ents(?)
					float sinom = vec2det(d1, d2) / (d1.len() * d2.len());
					
					if (sinom < -1) sinom = -1;
					if (sinom > 1) sinom = 1;
					float omega = asin(sinom) * 180.f / M_PI;
					//this gives the delta angle from the base position

					//TODO - use a vector on mouseDown, and HERE only cycle through it 
					for (int i = 0; i < spawnObjSys.count(); i++) {
						SpawnObj *s = spawnObjSys[i];
						assert(s);
						if (!s->sel) continue;
						//if (s == spawnObj) continue;
						float angle = 0;
						s->rw("angle") >> angle;
						s->rw("angle") << (angle + omega);
					}
					
					//we only need to update pos if an ent is selected
					if (editor->spawnObj) editorWndEnt.refreshSelected();
				}
			}
			if (scaling) {	//these should be mutually exclusive
				if (!!mouse.getMove()) {
					vec2f d1 = down - pivot;
					vec2f d2 = editor_mouse - pivot;
					down = editor_mouse;
					
					float ratio = d2.len() / d1.len();
					for (int i = 0; i < spawnObjSys.count(); i++) {
						SpawnObj *s = spawnObjSys[i];
						assert(s);
						if (!s->sel) continue;
						
						vec2f scale(1.f);
						s->rw("scale") >> scale;
						s->rw("scale") << (scale * ratio);
						
						//scale the bbox from its original too?
						//sure
						box2f bbox;
						if (!(s->rw("bbox") >> bbox)) bbox = s->getTypeBBox();;
						s->rw("bbox") << (bbox * ratio);
					}

					//we only need to update pos if an ent is selected
					if (editor->spawnObj) editorWndEnt.refreshSelected();
				}
			}
			break;
		}
	}
	
	//z-order
	if (mouse.getDZ()) {
		for (int i = 0; i < spawnObjSys.count(); i++) {
			SpawnObj *s = spawnObjSys[i];
			assert(s);
			if (!s->sel) continue;
			
			float zBias = 0;
			s->rw("zBias") >> zBias;
			s->rw("zBias") << (zBias + .1f * (float)mouse.getDZ());

		}
		//we only need to update pos if an ent is selected
		if (editor->spawnObj) editorWndEnt.refreshSelected();
	}
}

void EditorInputImpl::getInput() {

	if (!editor->active) return;

	//should these just fall through, or should they be allowed to return 'true' as well?
	switch (editor->method) {
	case EDITOR_EDIT_TAGS:		editTags();							break;
	case EDITOR_SURFACE_PAINT:	surfacePaint();					break;
	case EDITOR_SURFACE_FILL:	surfaceFill();						break;
	case EDITOR_SURFACE_MOVE:	surfaceMove();						break;
	case EDITOR_ENT_CREATE:		entCreate();						break;
	case EDITOR_ENT_SELECT:		entSelect();						break;
	}
}

void EditorInputImpl::updateMouse(const vec2f &v) {
	editor_mouse_last = editor_mouse;
	editor_mouse = v;
}

// ************************* EditorInput ************************* 

EditorInput::EditorInput(EditorImpl *ed_) : impl(new EditorInputImpl(ed_)) {}
EditorInput::~EditorInput() { delete impl; }
void EditorInput::getInput() { impl->getInput(); }
void EditorInput::updateMouse(const vec2f &v) { impl->updateMouse(v); }
