/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include "system/singletons.h"
#include "base/base.h"
#include "menu/sys.h"

#include "inspiration/world.h"
#include "inspiration/maptex.h"

#include "menulib.h"
#include "impl.h"
#include "func.h"
#include "wndent.h"

using namespace pwn;

// ************************* MenuEditorWindow ************************* 

	//dont allow close through menu_window_t - instead just set the visibility to zero
MenuEditorWindow::MenuEditorWindow(const vec2f &_pos, const vec2f &_size)
: super(_pos, _size, false)
{
	allowHide = true;
}

bool MenuEditorWindow::getVisible() {
	return editor.isActive() && super::getVisible();
}

// ************************* MenuEditorWindow ************************* 

MenuEditorObject::MenuEditorObject(const vec2f &_pos, const vec2f &_size)
: super(_pos, _size) { }

bool MenuEditorObject::getVisible() {
	return editor.isActive() && super::getVisible();
}

// ************************* MenuEditorRoot ************************* 

static const char *hotkeyChars[15] = {
	"q", "w", "e", "r", "t",
	"a", "s", "d", "f", "g",
	"z", "x", "c", "v", "b",
};

class MenuAnim : public MenuEditorObject {
	typedef MenuEditorObject super;
public:
	AnimState anim;
	
	MenuAnim(const vec2f &_pos, const vec2f &_size) : super(_pos, _size) {}
	
	virtual bool display() {
		if (!super::display()) return false;
		
		//render the anim
		animSys.render(
			size.x * .5f, size.y * .5f,
			0.f,	//depth
			0.f,
			2.f, 2.f, 
			vec4uc(255,255,255,255),
			0,
			getBlend(0),
			false,
			false,
			false,
			anim.getAbsFrame(),
			true
		);		
		return true;
	}
};

class HelperWidget : public MenuEditorObject {
	typedef MenuEditorObject super;
protected:
	gc::Pointer<Text> key;
	gc::Pointer<Object> tex;
	gc::Pointer<MenuAnim> anim;
public:

	HelperWidget(const vec2f &_pos, const char *str)
	:	super(_pos, vec2f(3,2)),
		key(createPtr<Text>()),
		tex(createPtr<Object>()),
		anim(createPtr<MenuAnim>())
	{
		key = new Text(vec2f(), str);
		addChild(key);
		key->setBorderColor(0);

		anim = new MenuAnim(vec2f(1,0), vec2f(2));
		addChild(anim);
		anim->setBorderColor(0);
		anim->setBackgroundColor(0);

		tex = new Object(vec2f(1,0), vec2f(2));
		addChild(tex);
		tex->setBorderColor(0);
		tex->setBackgroundColor(0xffffffff);
	}

	virtual void setTex(int id) {
		tex->setTextureID(id);
		tex->setBackgroundColor(id ? 0xffffffff : 0);
	}
	virtual void setAnim(const AnimState &animState) {
		anim->anim = animState;
		anim->setBackgroundColor(anim->anim.group ? 0xffffffff : 0);
	}
	virtual void setSelected(bool sel) {
		setBorderColor(sel ? 0xff00ffff : 0);
	}
};

//very editor-specific I suppose . . . not much of a library-ish class . . .
//so i'll make it singleton!

class MenuEditorRoot : public MenuEditorObject {
	typedef MenuEditorObject super;
public:
	gc::Pointer<Object> editTags, surfPaint, surfFill, surfMove, surfMovePres, entCreate, entSelect;
	
	//helper stuff
	gc::Pointer<Object> helper;
	std::vector<gc::Pointer<Object> > modeList;
	gc::Pointer<Text> curtypetext;

	//and per-menu stuff:
	gc::Pointer<Object> hotkeys;
	std::vector<gc::Pointer<HelperWidget> > hotkeyList;
	
	EditorImpl *ed;

	MenuEditorRoot(EditorImpl *ed_);
	virtual eventResponse_t mouseEvent(MouseEvent e, const vec2f &v);
};

MenuEditorRoot::MenuEditorRoot(EditorImpl *ed_)
:	super(vec2f(0,0),menuSys.getSize()),
	editTags(createPtr<Object>()),
	surfPaint(createPtr<Object>()),
	surfFill(createPtr<Object>()),
	surfMove(createPtr<Object>()),
	surfMovePres(createPtr<Object>()),
	entCreate(createPtr<Object>()),
	entSelect(createPtr<Object>()),
	helper(createPtr<Object>()),
	curtypetext(createPtr<Text>()),
	hotkeys(createPtr<Object>()),
	ed(ed_)
{
	visible = true;

//start buttons up, out of bounds
#define ROOT_MAKE_CHILD(var, str)	\
	var = new Text(vec2f(0,-2), str); addChild(var);
	ROOT_MAKE_CHILD(editTags, "Tags")
	ROOT_MAKE_CHILD(surfPaint, "Paint")
	ROOT_MAKE_CHILD(surfFill, "Fill")
	ROOT_MAKE_CHILD(surfMovePres, "Move(p)")
	ROOT_MAKE_CHILD(surfMove, "Move(n)")
	ROOT_MAKE_CHILD(entCreate, "MakeEnt")
	ROOT_MAKE_CHILD(entSelect, "Sel Ent")

	backgroundColor = 0;//0x3F3F0000;	//light fade dark blue

	
	//while we're here lets add some helper text
	//can we add a textbox and just add all the text to it
	//then have it display as transparent but with colored text?
	//why not?!
	{
		MenuEditorObject *b = new MenuEditorObject(vec2f(2),vec2f(1));
		b->setBorderColor(0);
		b->setBackgroundColor(0);
		Text *m = NULL;
		float y = 1;
		//for some odd reason the text is getting cut at the spaces .... hhmmm...
		m = new Text(vec2f(1,y++), "1: create entity"); b->addChild(m); m->setBorderColor(0); m->setBackgroundColor(0); modeList.push_back(createPtr<Object>(m));
		m = new Text(vec2f(1,y++), "2: select entity"); b->addChild(m); m->setBorderColor(0); m->setBackgroundColor(0); modeList.push_back(createPtr<Object>(m));
		m = new Text(vec2f(1,y++), "3: paint tiles"); b->addChild(m); m->setBorderColor(0); m->setBackgroundColor(0); modeList.push_back(createPtr<Object>(m));
		m = new Text(vec2f(1,y++), "4: fill tiles"); b->addChild(m); m->setBorderColor(0); m->setBackgroundColor(0); modeList.push_back(createPtr<Object>(m));
		m = new Text(vec2f(1,y++), "5: move tiles"); b->addChild(m); m->setBorderColor(0); m->setBackgroundColor(0); modeList.push_back(createPtr<Object>(m));
		m = new Text(vec2f(1,y++), "6: move w/pres"); b->addChild(m); m->setBorderColor(0); m->setBackgroundColor(0); modeList.push_back(createPtr<Object>(m));
		m = new Text(vec2f(1,y++), "7: edit tags"); b->addChild(m); m->setBorderColor(0); m->setBackgroundColor(0); modeList.push_back(createPtr<Object>(m));
		y++;
		m = new Text(vec2f(1,y++), "f3: entity window"); b->addChild(m); m->setBorderColor(0); m->setBackgroundColor(0);
		m = new Text(vec2f(1,y++), "f4: tile window"); b->addChild(m); m->setBorderColor(0); m->setBackgroundColor(0);
		m = new Text(vec2f(1,y++), "f5: tag window"); b->addChild(m); m->setBorderColor(0); m->setBackgroundColor(0);
		m = new Text(vec2f(1,y++), "f6: texture window"); b->addChild(m); m->setBorderColor(0); m->setBackgroundColor(0);
		b->setSize(vec2f(20,y+2));
		addChild(b);
		helper = b;
	}
	{
		MenuEditorObject *h = new MenuEditorObject(vec2f(menuSys.getSize().x * 0.5 + 2, 4), vec2f(1));
		//how add all our shortcut things:
		HelperWidget *m = NULL;
		vec2f size(1);
		for (int j = 0; j < 3; j++) {
			for (int i = 0; i < 5; i++) {
				vec2f pos(4.1f*i + 2*j, 2.1f*j+ 2.f);
				m = new HelperWidget(pos, hotkeyChars[j*5+i]);
				h->addChild(m);
				m->setBorderColor(0);
				m->setVisible(false);
				hotkeyList.push_back(createPtr<HelperWidget>(m));
				if (pos.x + m->getSize().x > size.x) size.x = pos.x + m->getSize().x;
				if (pos.y + m->getSize().y > size.y) size.y = pos.y + m->getSize().y;
			}
		}
		h->setSize(size);
		h->setBackgroundColor(0x7f000000);
		h->setBorderColor(0);
		h->setVisible(false);
		addChild(h);
		hotkeys = h;
		
		//also add a text to it that shows the name of what we're creating
		//tied to the current selected ent huh? sounds tediuos sharing info between the two ...
		curtypetext = new Text(vec2f(2,0), "");
		h->addChild(curtypetext);
	}
}

	
	/*
	this next method is a threat
	it relies on a gl render of our group to be done
	but i was thinking of making a separate gl pathway (for the entity editor) anyways
	*/

//here's an idea: 
//on right-click on the root,
//pop-up all the dialogs, shrunk down to a 2x2 or so square, all spinning about the mouse
//upon click, have it expand back to its original position (or a position under the mouse)
//for now just use texts?
eventResponse_t MenuEditorRoot::mouseEvent(MouseEvent e, const vec2f &v) {
	//no matter what rotate our helper menu
//con << "mouse at " << v << "\n";
	if (v.x > size.x * 0.5) {
		helper->setPos(vec2f(2, helper->getPos().y));
		hotkeys->setPos(vec2f(size.x - hotkeys->getSize().x - 2, hotkeys->getPos().y));
	} else {
		helper->setPos(vec2f(size.x - helper->getSize().x - 2, helper->getPos().y));
		hotkeys->setPos(vec2f(2, hotkeys->getPos().y));
	}
	if (v.y > size.y * 0.5) {
		helper->setPos(vec2f(helper->getPos().x, 2));
		hotkeys->setPos(vec2f(hotkeys->getPos().x, 2));
	} else {
		helper->setPos(vec2f(helper->getPos().x, size.y - helper->getSize().y - 2));
		hotkeys->setPos(vec2f(hotkeys->getPos().x, size.y - hotkeys->getSize().y - 2));
	}
	
	//do our hotkeys for 1-7
	
	//only if the menu has no kbd focus ...
	//in a way this is like sending a keypress through all its parents to the focus child
	//	(thereby puttin this code in the editor root)
	//much in the same way mouse clicking works
	//and this would be an override ... the children (?) 
	//	no, this would get it on the bubble out, only if the children didn't override it
	if (!menuSys.keyWasTrapped()) {
		if (kbd.keystate['1'] == BUTTON_STATE_PRESS) editor->setMethod(EDITOR_ENT_CREATE);
		if (kbd.keystate['2'] == BUTTON_STATE_PRESS) editor->setMethod(EDITOR_ENT_SELECT);
		if (kbd.keystate['3'] == BUTTON_STATE_PRESS) editor->setMethod(EDITOR_SURFACE_PAINT);
		if (kbd.keystate['4'] == BUTTON_STATE_PRESS) editor->setMethod(EDITOR_SURFACE_FILL);
		if (kbd.keystate['5'] == BUTTON_STATE_PRESS) { editor->setMethod(EDITOR_SURFACE_MOVE); editor->movePreserve = true; }
		if (kbd.keystate['6'] == BUTTON_STATE_PRESS) { editor->setMethod(EDITOR_SURFACE_MOVE); editor->movePreserve = false; }
		if (kbd.keystate['7'] == BUTTON_STATE_PRESS) editor->setMethod(EDITOR_EDIT_TAGS);
	}

	//update colors of goods
	for (int i = 0; i < modeList.size(); i++) { modeList[i]->setFontColor(0xffffffff); }
	switch (editor->method) {
	case EDITOR_ENT_CREATE: modeList[0]->setFontColor(0xff0000ff); break;
	case EDITOR_ENT_SELECT: modeList[1]->setFontColor(0xff0000ff); break;
	case EDITOR_SURFACE_PAINT: modeList[2]->setFontColor(0xff0000ff); break;
	case EDITOR_SURFACE_FILL: modeList[3]->setFontColor(0xff0000ff); break;
	case EDITOR_SURFACE_MOVE:
		if (editor->movePreserve) modeList[4]->setFontColor(0xff0000ff);
		else modeList[5]->setFontColor(0xff0000ff);
		break;
	case EDITOR_EDIT_TAGS: modeList[6]->setFontColor(0xff0000ff); break;
	}

	switch (editor->method) {
	case EDITOR_ENT_CREATE:
	case EDITOR_ENT_SELECT:
		hotkeys->setVisible(true);
		{
			static int start = 0;
			//cycle through spawn types
			//and display them
			for (int hotkeyNo = 0; hotkeyNo < hotkeyList.size(); hotkeyNo++) {
				int index = hotkeyNo + start * hotkeyList.size();
				if (index < spawnTypeSys.count()) {
					std::string curType = spawnTypeSys.getNameForIndex(index);

					//i should probably route this through the helperwidget display function ...
					//so i get scaling and what not ...
					int anim_group = spawnTypeSys.getGroupIndex(index);
#pragma warning instead in the render call animSys.render(...etc) or something
#pragma warning or maybe I should make a gl based anim renderer?
					hotkeyList[hotkeyNo]->setTex(0);
					hotkeyList[hotkeyNo]->setAnim(AnimState(anim_group,0,0));
					hotkeyList[hotkeyNo]->setSelected(curType == editor->spawnType);

					int keycheck = hotkeyChars[hotkeyNo][0];
					if (!menuSys.keyWasTrapped() && kbd.keystate[keycheck] == BUTTON_STATE_PRESS) {	//for some odd reason state_press isn't showing up ...
						editor->spawnType = curType;

						//make this part of editorWndEnt.refreshTypeCurrent()
						curtypetext->setText(curType);
						curtypetext->setSize(vec2f(curtypetext->getText().length(), curtypetext->getSize().y));
						
						//change all selected spawnobjs to this type
						for (int i = 0; i < spawnObjSys.count(); i++) {
							SpawnObj *s = spawnObjSys[i];
							assert(s);
							if (!s->sel) continue;
							s->rw("classname") << curType;
						}
						
						editorWndEnt.refreshTypeCurrent();
					}

					hotkeyList[hotkeyNo]->setVisible(true);
				} else {
					hotkeyList[hotkeyNo]->setVisible(false);
				}
			}
			
			int keyCount = spawnTypeSys.count();
			int lastPage = (keyCount-1) / hotkeyList.size();
			if (!menuSys.keyWasTrapped() && kbd.keystate['['] == BUTTON_STATE_PRESS) start--;
			if (!menuSys.keyWasTrapped() && kbd.keystate[']'] == BUTTON_STATE_PRESS) start++;
			if (start > lastPage) start = 0;
			if (start < 0) start = lastPage;
		}
		break;
	case EDITOR_SURFACE_PAINT:
	case EDITOR_SURFACE_FILL:
	case EDITOR_SURFACE_MOVE:
		hotkeys->setVisible(true);
		{
			static int start = 0;
			//same but with types and stuff
			int mapkeyNo = -1;
			int hotkeyNo = -1;
			for (int mapkeyIndex = 0; mapkeyIndex < ed->world->keys->size(); mapkeyIndex++) {
				MapKey *k = NULL;
				if (mapkeyIndex) {	//let our empty tile pass through
					if (!ed->world->keys->isValid(mapkeyIndex)) continue;
					k = (*ed->world->keys)[mapkeyIndex];
					if (!k) continue;
				}
				mapkeyNo++;
				if (mapkeyNo < start * hotkeyList.size()) continue;	//before
				hotkeyNo++;
				if (hotkeyNo < hotkeyList.size()) {		//after

					//here's the deal
					//first we need to get map tag [i]
					//then we need to get its texture
					{
						int texID = 0;
						if (k) {
							MapTex *mt = k->tex;
							if (mt) {
								const TexLoad *t = texSys.getObj(mt->getCurrentImage());
								if (t) texID = t->tex.getTextureID();
							}
						}
						hotkeyList[hotkeyNo]->setTex(texID);
						hotkeyList[hotkeyNo]->setSelected(mapkeyIndex == editor->mapKey);
					}

					//what if its invalid?
					{
						AnimState animState;
						int texID = 0;
						if (k) {
#pragma warning same deal here.  we need a gl based renderer for animations
							animState = AnimState(animSys.getGroupIndex(k->groupname.c_str()), 0,0);
						}
						hotkeyList[hotkeyNo]->setAnim(animState);
					}

					int keycheck = hotkeyChars[hotkeyNo][0];
					if (!menuSys.keyWasTrapped() && kbd.keystate[keycheck] == BUTTON_STATE_PRESS) {	//for some odd reason state_press isn't showing up ...
						editor->setMapKeySel(mapkeyIndex);
					}

					hotkeyList[hotkeyNo]->setVisible(true);
				}
			}
			//invis any remaining...
			for (hotkeyNo++; hotkeyNo < hotkeyList.size(); hotkeyNo++) {
				hotkeyList[hotkeyNo]->setVisible(false);
			}

			//process cycles
			int keyCount = mapkeyNo;
			int lastPage = (keyCount-1) / hotkeyList.size();
			if (!menuSys.keyWasTrapped() && kbd.keystate['['] == BUTTON_STATE_PRESS) start--;
			if (!menuSys.keyWasTrapped() && kbd.keystate[']'] == BUTTON_STATE_PRESS) start++;
			if (start > lastPage) start = 0;
			if (start < 0) start = lastPage;

			//what was this gonna be for?
			//for (int i = 0; i < numberof(hotkeyChars); i++) {
		}
		break;
	default:
		hotkeys->setVisible(false);
	}

	//clicking a button sets the method
	if (selectedChild) {
		if (e.mrelease) {
			//if we clicked on something . . . hide all kids and process the input
			for (std::list<gc::Pointer<Object> >::iterator i = kids.begin(); i != kids.end(); ++i) {
				(*i)->setPos(vec2f(0,-2));
			}
		
			if (selectedChild.ptr() == editTags) {
				editor->setMethod(EDITOR_EDIT_TAGS);
				return RESPONSE_STOP;
			} else if (selectedChild.ptr() == surfPaint) {
				editor->setMethod(EDITOR_SURFACE_PAINT);
				return RESPONSE_STOP;
			} else if (selectedChild.ptr() == surfFill) {
				editor->setMethod(EDITOR_SURFACE_FILL);
				return RESPONSE_STOP;
			} else if (selectedChild.ptr() == surfMovePres) {
				editor->movePreserve = true;
				editor->setMethod(EDITOR_SURFACE_MOVE);
				return RESPONSE_STOP;
			} else if (selectedChild.ptr() == surfMove) {
				editor->movePreserve = false;
				editor->setMethod(EDITOR_SURFACE_MOVE);
				return RESPONSE_STOP;
			} else if (selectedChild.ptr() == entCreate) {
				editor->setMethod(EDITOR_ENT_CREATE);
				return RESPONSE_STOP;
			} else if (selectedChild.ptr() == entSelect) {
				editor->setMethod(EDITOR_ENT_SELECT);
				return RESPONSE_STOP;
			} else {
				con.print("got unknown middle child!\n");
				return RESPONSE_STOP;
			}
		}
	//if we didnt get a selected child
	} else {
		if (e.mpress) {
			//pop-up, around us, our buttons
			int index = 0;
			for (std::list<gc::Pointer<Object> >::iterator i = kids.begin(); i != kids.end(); ++i,++index) {
				float th = 2.f * M_PI * ((float)index + 0.4f) / (float)kids.size();
				vec2f r = vec2f((float)cos(th), (float)sin(th));
				(*i)->setPos(v + r * 3.f);
			}
			return RESPONSE_STOP;
		}
	}
	
	//right-click anywhere to pop up our current method's menu ... ?
	//nah, i'm overriding this so we can use right-click for object manipulation
#if 0
	if (e.rpress) {
		switch (method) {
		case EDITOR_EDIT_TAGS:
			editorFunc.showWindow(EW_TAG, 1);
			return RESPONSE_STOP;
		case EDITOR_SURFACE_PAINT:
		case EDITOR_SURFACE_FILL:
		case EDITOR_SURFACE_MOVE:
			editorFunc.showWindow(EW_TILE, 1);
			return RESPONSE_STOP;
		case EDITOR_ENT_CREATE:
		case EDITOR_ENT_SELECT:
			editorFunc.showWindow(EW_ENT, 1);
			return RESPONSE_STOP;
		}
	}
#endif
	return RESPONSE_CONTINUE;	//absolutely must return false for the editor/world interactions to take place
//	return super::doMouseClick(event,v);
//	case ID_WINDOW_WORLD:		editorFunc.showWindow(EW_MAIN, 2);		return true;
//	case ID_WINDOW_TEXTURE:		editorFunc.showWindow(EW_TEX, 2);		return true;
//	case ID_WINDOW_SCRIPT:		editorFunc.showWindow(EW_SCRIPT, 2);	return true;
}

MenuEditorObject *MenuEditorRootSys::create(EditorImpl *ed) {
	return new MenuEditorRoot(ed);
}
