/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <stdio.h>

#include "system/exception.h"
#include "base/base.h"

#include "inspiration/world.h"

#include "impl.h"
#include "wndtag.h"
#include "menulib.h"

//*************
//TAG WINDOW
//*************

using namespace std;
using namespace pwn;

// ************************* MenuTag prototype ************************* 

class MenuTag : public MenuEditorWindow, TextListener, SelectListener {
	typedef MenuEditorWindow super;
protected:
	EditorImpl *ed;
public:

	gc::Pointer<Text> methodTags;

	gc::Pointer<Select> selectTagKey;
	gc::Pointer<Text> createTagKey;
	gc::Pointer<Text> removeTagKey;
	
	gc::Pointer<Select> selectTagType;
	gc::Pointer<Text> createTagType;
	gc::Pointer<Text> removeTagType;

	gc::Pointer<Select> selectBGTex;
	
	//TODO: generalize vector inputs of 2d and 3d ...
	gc::Pointer<TextField> scrollX, scrollY;
	gc::Pointer<TextField> scaleX, scaleY;
	gc::Pointer<TextField> gravX, gravY;
	gc::Pointer<TextField> lightR, lightG, lightB;
	
	MenuTag(EditorImpl *);
	virtual eventResponse_t mouseEvent(MouseEvent e, const vec2f &v);
	virtual bool display();

	virtual void callback_onTextChange(TextField *);
	virtual void callback_userSetSelected(Select *, int newSel, int oldSel);
};

class EditorWndTagImpl {
protected:
	EditorImpl *ed;
public:
	gc::GlobalPointer<MenuTag> tagWnd;

	EditorWndTagImpl(EditorImpl *ed_)
	:	tagWnd(gc::sys->createPtr<MenuTag>()),
		ed(ed_)
	{}

	void init();
	void refresh();
	void refreshAllTextures();
	void refreshComboTagKeySelection();
	void refreshComboTagKeyContents();
	void refreshComboTagTypeSelection();
	void refreshComboTagTypeContents();
	void refreshTagType();
};

void EditorWndTagImpl::refreshComboTagKeySelection() {
	int tobesel = -1;
	//reverse mapping from key->id (sparse) to key's index (dense)
	int i = 0;
	for (map<short,MapTagKey*>::iterator it = ed->world->tagkeys->keys.begin(); it != ed->world->tagkeys->keys.end(); it++) {
		int id = it->first;
		if (id == editor->tagKey) tobesel = i;
		i++;
	}

	tagWnd->selectTagKey->setSelected(tobesel);
	//as a result for this, update anything dependant upon editor_tagkey_sel ?

		//i wonder if this is safe . . .
	refreshComboTagTypeSelection();
}

void EditorWndTagImpl::refreshComboTagKeyContents() {
	char name[256];
	int tobesel = -1;

	tagWnd->selectTagKey->resetContent();

	int i = 0;
	for (map<short,MapTagKey*>::iterator it = ed->world->tagkeys->keys.begin(); it != ed->world->tagkeys->keys.end(); it++) {
		int id = it->first;
		sprintf(name, "%.4x", id - MapTagKey::minRange);

		tagWnd->selectTagKey->addString(name);

		if (id == editor->tagKey) tobesel = i;
		i++;
	}
	refreshComboTagKeySelection();
}

void EditorWndTagImpl::refreshTagType() {

	char buffer[256];
	
	//is this even used?
	if (editor->tagKey == -1) {
		buffer[0] = 0;

		tagWnd->selectBGTex->setSelected(-1);
		tagWnd->scrollX->setText("");
		tagWnd->scrollY->setText("");
		tagWnd->scaleX->setText("");
		tagWnd->scaleY->setText("");
		tagWnd->gravX->setText("");
		tagWnd->gravY->setText("");
		tagWnd->lightR->setText("");
		tagWnd->lightG->setText("");
		tagWnd->lightB->setText("");
	}

	MapTagType tt;
	map<short,MapTagKey*>::iterator it = ed->world->tagkeys->keys.find(editor->tagKey);
	if (it != ed->world->tagkeys->keys.end()) {
		MapTagType *src = it->second->tagType;
		if (src) tt = *src;
	}

	tagWnd->selectBGTex->setSelected(ed->world->texs->getIndex(tt.tex));

	TextField *gravWnds[] = {
		tagWnd->gravX,
		tagWnd->gravY
	};
	editor->setFloatTextFields(tt.grav, gravWnds, numberof(gravWnds));

	TextField *lightWnds[] = {
		tagWnd->lightR,
		tagWnd->lightG,
		tagWnd->lightB,
	};
	editor->setFloatTextFields(tt.light, lightWnds, numberof(lightWnds));

	TextField *scrollWnds[] = {
		tagWnd->scrollX,
		tagWnd->scrollY,
	};
	editor->setFloatTextFields(tt.scroll, scrollWnds, numberof(scrollWnds));

	TextField *scaleWnds[] = {
		tagWnd->scaleX,
		tagWnd->scaleY,
	};
	editor->setFloatTextFields(tt.scale, scaleWnds, numberof(scaleWnds));
}

void EditorWndTagImpl::refreshComboTagTypeSelection() {

	tagWnd->selectTagType->setSelected(editor->tagType);

	refreshTagType();
}

void EditorWndTagImpl::refreshComboTagTypeContents() {

	tagWnd->selectTagType->resetContent();
	for (int i = 0; i < ed->world->tagtypes->tags.size(); i++) {
		tagWnd->selectTagType->addString(ed->world->tagtypes->tags[i]->name.c_str());
	}

	refreshComboTagTypeSelection();
}

void EditorWndTagImpl::refreshAllTextures() {
	editor->refreshTextureSelect(tagWnd->selectBGTex);
}

void EditorWndTagImpl::refresh() {
	refreshAllTextures();
	refreshComboTagKeyContents();	//tagKey
	refreshComboTagTypeContents();	//tagType, and all the text boxes beneath
}

// ************************* MenuTag ************************* 

MenuTag::MenuTag(EditorImpl *ed_)
:	super(vec2f(1,1),vec2f(24,21)),
	methodTags(createPtr<Text>()),
	selectTagKey(createPtr<Select>()),
	createTagKey(createPtr<Text>()),
	removeTagKey(createPtr<Text>()),
	selectTagType(createPtr<Select>()),
	createTagType(createPtr<Text>()),
	removeTagType(createPtr<Text>()),
	selectBGTex(createPtr<Select>()),
	scrollX(createPtr<TextField>()),
	scrollY(createPtr<TextField>()),
	scaleX(createPtr<TextField>()),
	scaleY(createPtr<TextField>()),
	gravX(createPtr<TextField>()),
	gravY(createPtr<TextField>()),
	lightR(createPtr<TextField>()),
	lightG(createPtr<TextField>()),
	lightB(createPtr<TextField>()),
	ed(ed_)
{
	visible = false;

	addChild(methodTags = new Text(vec2f(1,1), "Select Tags"));
	
	addChild(new Text(vec2f(1,3), "Key:"));
	addChild(selectTagKey = new Select(vec2f(1, 4), 10));
	selectTagKey->addSelectListener(this);
	addChild(createTagKey = new Text(vec2f(12,3), "Create"));
	addChild(removeTagKey = new Text(vec2f(12,4), "Remove"));

	addChild(new Text(vec2f(1,6), "Type:"));
	addChild(selectTagType = new Select(vec2f(1, 7), 10));
	selectTagType->addSelectListener(this);
	addChild(createTagType = new Text(vec2f(12,6), "Create"));
	addChild(removeTagType = new Text(vec2f(12,7), "Remove"));
	
	selectTagType->allowEdit = true;
	
	addChild(new Text(vec2f(1,9), "BG Tex:"));
	addChild(selectBGTex = new Select(vec2f(9,9), 10));
	selectBGTex->addSelectListener(this);
	
	addChild(new Text(vec2f(1,11), "Scroll:"));
	addChild(scrollX = new TextField(vec2f(9,11), 7));
	scrollX->addTextListener(this);
	addChild(scrollY = new TextField(vec2f(17,11), 7));
	scrollY->addTextListener(this);
	
	addChild(new Text(vec2f(1,13), "Scale:"));
	addChild(scaleX = new TextField(vec2f(9,13), 7));
	scaleX->addTextListener(this);
	addChild(scaleY = new TextField(vec2f(17,13), 7));
	scaleY->addTextListener(this);
	
	addChild(new Text(vec2f(1,15), "Gravity:"));
	addChild(gravX = new TextField(vec2f(2,16), 7));
	gravX->addTextListener(this);
	addChild(gravY = new TextField(vec2f(10,16), 7));
	gravY->addTextListener(this);
	
	addChild(new Text(vec2f(1,18), "Light:"));
	addChild(lightR = new TextField(vec2f(1,19), 7));
	lightR->addTextListener(this);
	addChild(lightG = new TextField(vec2f(9,19), 7));
	lightG->addTextListener(this);
	addChild(lightB = new TextField(vec2f(17,19), 7));
	lightB->addTextListener(this);
}

eventResponse_t MenuTag::mouseEvent(MouseEvent e, const vec2f &v) {
	
	//handle buttons here (Text)
	if (e.lpress) {
		if (selectedChild.ptr() == methodTags) {
			editor->setMethod(EDITOR_EDIT_TAGS);
			return RESPONSE_STOP;
		} else if (selectedChild.ptr() == createTagKey) {
			int newtagkey = ed->world->tagkeys->create();
			if (newtagkey == -1) {
				con.print("failed to create map tag key\n");
			} else {
				editor->tagKey = newtagkey;
				editor->tagType = -1;
				map<short,MapTagKey*>::iterator it = ed->world->tagkeys->keys.find(editor->tagKey);
				if (it != ed->world->tagkeys->keys.end()) editor->tagType = ed->world->tagtypes->getIndex(it->second->tagType);
				ed->tagWnd->refreshComboTagKeyContents();
				ed->tagWnd->refreshComboTagTypeSelection();
			}
			return RESPONSE_STOP;
		} else if (selectedChild.ptr() == removeTagKey) {
			//TODO - get the tag key from the combo
			editor->tagKey = ed->world->tagkeys->remove(editor->tagKey);
			editor->tagType = -1;
			map<short,MapTagKey*>::iterator it = ed->world->tagkeys->keys.find(editor->tagKey);
			if (it != ed->world->tagkeys->keys.end()) editor->tagType = ed->world->tagtypes->getIndex(it->second->tagType);
			ed->tagWnd->refreshComboTagKeyContents();
			ed->tagWnd->refreshComboTagTypeSelection();
			return RESPONSE_STOP;
		} else if (selectedChild.ptr() == createTagType) {
			int newtagtype = ed->world->tagtypes->create();
			if (newtagtype != -1) {
				map<short,MapTagKey*>::iterator it = ed->world->tagkeys->keys.find(editor->tagKey);
				if (it != ed->world->tagkeys->keys.end()) it->second->tagType = ed->world->tagtypes->tags[newtagtype];
				ed->tagWnd->refreshComboTagTypeContents();
			}
			return RESPONSE_STOP;
		} else if (selectedChild.ptr() == removeTagType) {
			short type = -1;
			map<short,MapTagKey*>::iterator it = ed->world->tagkeys->keys.find(editor->tagKey);
			if (it != ed->world->tagkeys->keys.end()) type = ed->world->tagtypes->getIndex(it->second->tagType);
			type = ed->world->tagtypes->remove(type);
			if (it != ed->world->tagkeys->keys.end()) {
				it->second->tagType = (type >= 0 && type < ed->world->tagtypes->tags.size())
					? ed->world->tagtypes->tags[type] : NULL;
			}
			ed->tagWnd->refreshComboTagTypeContents();
			return RESPONSE_STOP;
		}
	}
	
	//handle sliders here
	
	super::mouseEvent(e, v);
	return RESPONSE_STOP;
}

bool MenuTag::display() {
	//update colors
	methodTags->setFontColor((editor->method == EDITOR_EDIT_TAGS) ? 0xFF0000FF : 0xFFFFFFFF);
	//display rect
	return super::display();
}

void MenuTag::callback_onTextChange(TextField *m) {
	if (editor->tagType < 0 || editor->tagType >= ed->world->tagtypes->tags.size()) return;
	MapTagType *tt = ed->world->tagtypes->tags[editor->tagType];
	
	if (m == scrollX)		tt->scroll[0] = atof(m->getText().c_str());
	else if (m == scrollY)	tt->scroll[1] = atof(m->getText().c_str());
	else if (m == scaleX)	tt->scale[0] = atof(m->getText().c_str());
	else if (m == scaleY)	tt->scale[1] = atof(m->getText().c_str());
	else if (m == gravX)	tt->grav[0] = atof(m->getText().c_str());
	else if (m == gravY)	tt->grav[1] = atof(m->getText().c_str());
	else if (m == lightR)	tt->light[0] = atof(m->getText().c_str());
	else if (m == lightG)	tt->light[1] = atof(m->getText().c_str());
	else if (m == lightB)	tt->light[2] = atof(m->getText().c_str());
	else if (m == selectTagType) {
		//make sure the characters are valid
		if (editor->nameIsValid(m->getText().c_str())) {
			//check it to all other types /EXCEPT/ the current selected one
			//if its unique then set it
			int matchingname_tagtype = ed->world->tagtypes->findByName(m->getText().c_str());
			if (matchingname_tagtype == -1) {
				if (editor->tagType >= 0 && editor->tagType < ed->world->tagtypes->tags.size()) {
					ed->world->tagtypes->tags[editor->tagType]->name = m->getText();
				}
				//and replace the combo box contents
				ed->tagWnd->refreshComboTagTypeContents();
			}
		}
	}
}

void MenuTag::callback_userSetSelected(Select *m, int newSel, int oldSel) {
	if (newSel == oldSel) return;	//only on change
	if (m == ed->tagWnd->tagWnd->selectTagKey) {

		//anyone know any faster hex-string-to-int methods?
		char buffer[256];
		strcpy(buffer, m->items[newSel]->getText().c_str());
		sscanf(buffer, "%x", &editor->tagKey);
		
		//adding 2020 would mess up the -1, wouldn't it?  well so would the sscanf . . .
		editor->tagKey += MapTagKey::minRange;

		editor->tagType = -1;
		map<short,MapTagKey*>::iterator it = ed->world->tagkeys->keys.find(editor->tagKey);
		if (it != ed->world->tagkeys->keys.end()) {
			editor->tagType = ed->world->tagtypes->getIndex(it->second->tagType);
		}

		ed->tagWnd->refreshComboTagKeySelection();
	} else if (m == selectTagType) {
		editor->tagType = newSel;
		map<short,MapTagKey*>::iterator it = ed->world->tagkeys->keys.find(editor->tagKey);
		if (it != ed->world->tagkeys->keys.end()) {
			it->second->tagType = (editor->tagType >= 0 && editor->tagType < ed->world->tagtypes->tags.size())
				? ed->world->tagtypes->tags[editor->tagType] : NULL;
		}
		ed->tagWnd->refreshTagType();
	} else if (m == selectBGTex) {
		if (editor->tagType >= 0 && editor->tagType < ed->world->tagtypes->tags.size()) {
			ed->world->tagtypes->tags[editor->tagType]->tex = ed->world->texs->texs[newSel];
		}
	}
}

//notice - this usually doesnt get called this way =) time to change the windows system . . .
void EditorWndTagImpl::init() {
	(*editor->editWnd[EW_TAG]) = tagWnd = new MenuTag(ed);
	if (!tagWnd) throw Exception() << "EditorWndTagImpl::init failed";
	editor->rootmenu->addChild(tagWnd);
	//refresh();
}

EditorWndTag::EditorWndTag(EditorImpl *ed) : impl(new EditorWndTagImpl(ed)) {}
EditorWndTag::~EditorWndTag() { delete impl; }
void EditorWndTag::init() { impl->init(); }
void EditorWndTag::refresh() { impl->refresh(); }
void EditorWndTag::refreshAllTextures() { impl->refreshAllTextures(); }
void EditorWndTag::refreshComboTagKeySelection() { impl->refreshComboTagKeySelection(); }
