/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <string>
#include <sstream>
#include <vector>
#include <map>

#include <math.h>		//for degrees -> index
#include <string.h>
#include <assert.h>
#include <stdio.h>

#include "math/mat.h"

#include "base/base.h"	//con
#include "base/filebuffer.h"

#include "inspiration/draw.h"	//EGeometry
#include "inspiration/animation.h"
#include "inspiration/texloader.h"
#include "inspiration/filesys.h"
#include "inspiration/lua.h"
#include "inspiration/main.h"	//game_thistime
#include "inspiration/program.h"	//main_lua_state
#include "inspiration/strcasecmp.h"
#include "inspiration/game/main.h"

using namespace std;

class AnimSysImpl {
public:
	bool sessionInit();
	void sessionKill();

	void render(
		float x, float y, float z,
		float r, float sx, float sy,
		vec4uc color,
		unsigned int shader, Blend blendfunc, bool postFBO, bool useLighting,
		bool flip, int absFrame, bool immediate);

	int getGroupIndex(const char *group);
	bool getGroupName(int group, string &result);
	bool getSeqName(int group, int seq, std::string &result);
	int getAbsFrame(int groupIndex, int seqIndex, int frameIndex);
};

//this will be transformed into a class stored within the group struct
//that class will have a sprite obj ref in place of the filename
//should sprites really have bbox's? shouldn't joints?
//i suppose as long as joints have scales it shouldn't matter
//this way sprite aspect ratios will be preserved.
class AnimSprite {
public:
	AnimSprite() : bbox(vec2f(-.5f), vec2f(.5f)) {}
	string filename;
	box2f bbox;
	TexPtr image;
};

//if anything need be, sprite and parent can be converted to pointers
//this used to be for the rig - and hold the non-changing stuff
//but considering how interpolation works, i think it'll soon be merged with AnimJoint

class AnimKeyframe;

// when we C++ ize this one it'll be tricky
//because certain components might not change in a keyframe
//that means we will be doing our interpolation between dif keyframes for each component
//how do we keep from having to search through the whole seq every frame?
//we could precache, at each keyframe, a 'prev' and 'next' for each component
//and it's time block
// the current setup already keeps track of the cur start time and end time
//if we assert forward transitions always then we can keep one accumulation buffer with the state
// and only update the 'next' pointers with the keyframes as we get to them.
class AnimJoint {
public:
	//i have an issue with the default values
	//the default lua values are 1-based
	//from there they are offset to zero-based
	//i'd rather keep these defaults 0-based
	//so i'm gonna put the lua-specific 1-based defaults in all the lua read code
	//it just might be easier to store all our indexes in lua as 0-based
	//even if they don't match up in the lua code (which i'm never using for indexing)
	AnimJoint() : 
		angle(0), scale(1.f), sprite(0), parent(-1), uv(vec2f(0.f), vec2f(1.f)), flags(0),
		prevPos(0), prevAngle(0), prevScale(0), prevSprite(0), prevUV(0),
		nextPos(0), nextAngle(0), nextScale(0), nextSprite(0), nextUV(0)
	{ }

	vec3f pos;
	float angle;
	vec2f scale;
	int sprite;	//local ref in our AnimSprite struct
	box2f uv;	//texcoord range
	int parent;	//parent joint index. -1 == none. (always refs before itself so it is acyclic)
	
	//accum'd transform so far
	//for ease of skeletal transform caching or something
	mat22f rxform;	
	vec2f txform;

	//flags for each of our things that might interpolate
	union {
		struct {
			unsigned modPos : 1;
			unsigned modAngle : 1;
			unsigned modScale : 1;
			unsigned modSprite : 1;
			unsigned modUV : 1;
		};
		short flags;
	};
	
	//likewise we have pointers to which Keyframe held the last set value
	//if this frame's mod flag is set this'll point to this frame.
	//if not it'll point back to the first previous joint with the mod flag set
	//if none have it set then it'll point back to ... the base rig?  or the last mod flag frame?
	AnimKeyframe *prevPos, *prevAngle, *prevScale, *prevSprite, *prevUV;

	//this points to the next joint with the mod flag set
	//if none do then ... this loops back around?  or pairs up with the base rig?
	//probably loops back around . . .
	AnimKeyframe *nextPos, *nextAngle, *nextScale, *nextSprite, *nextUV;
	
	AnimJoint &operator=(const AnimJoint &src) {
		pos = src.pos;
		angle = src.angle;
		scale = src.scale;
		sprite = src.sprite;
		uv = src.uv;
		parent = src.parent;

		flags = src.flags;
		
		//should these copy or invalidate (null) ?
		prevPos = src.prevPos;
		prevAngle = src.prevAngle;
		prevScale = src.prevScale;
		prevSprite = src.prevSprite;
		prevUV = src.prevUV;
		
		nextPos = src.nextPos;
		nextAngle = src.nextAngle;
		nextScale = src.nextScale;
		nextSprite = src.nextSprite;
		nextUV = src.nextUV;
		
		return *this;
	}
};

class AnimKeyframe {
public:
	AnimKeyframe() : time(-1.f), dt(1.f) {}
	AnimKeyframe(float _time, int numJoints) : time(_time) { joints.resize(numJoints); }
	
	void reset(float _time, int numJoints) { time = _time; joints.resize(numJoints); }
	
	//maybe this should be duration rather than time ...
	//nah.  the final one would always have a zero duration.  that'd be repetative(?) meh.
	//of course lookups for the duration would be faster ...
	float time;
	float dt;

	vector<AnimJoint> joints;	//sized the same as the rig's size

	AnimKeyframe &operator=(const AnimKeyframe &src) {
		time = src.time;
		joints.resize(src.joints.size());
		for (int i = 0; i < joints.size(); i++) joints[i] = src.joints[i];
		return *this;
	}
};

class AnimSkelSeq {
public:
	string name;
/*
not just sort by time
we need an array of timeslices, each with their unique vector of keyframes (one per joint in the skeleton)
*/
	vector<AnimKeyframe> frames;	//sorted by time
	
	AnimSkelSeq() : name("idle") {}
	
	AnimSkelSeq &operator=(const AnimSkelSeq &src) {
		name = src.name;
		frames.resize(src.frames.size());
		for (int i = 0; i < frames.size(); i++) frames[i] = src.frames[i];
		return *this;
	}
};

//our new system
class AnimSkelGroup {
public:
	AnimSkelGroup(const char *_name) : name(_name) {}
	
	//always useful for occlusion's sake
	//i'll auto-calc this for the time being
	//but it sure would be nice to pre-store or something ... idunno ...
	box2f bbox;

	string name;
	vector<AnimSprite> sprites;	//localized references of what sprites are used.
	AnimKeyframe rig;			//base skeleton as defaults for any sequence
	vector<AnimSkelSeq> seqs;
	
	void finalize() {

		//if we get no readSeqs then we should throw in a default...
		if (!seqs.size()) {
			seqs.resize(1);	//default idle seq
			seqs[0].frames.resize(1);	//default single frame, single joint, etc
			seqs[0].frames[0].time = 0;	//touch
			seqs[0].frames[0].joints.resize(1);	//single joint in our frame, to match the rig
			//this joint's sprite will default to zero
			//but what happens if there is no sprite?  (i.e. the null frame) ?
		}
		
		//calc the bbox while we're here
		//start with a default in case no points get written
		bbox.min.set(0.f, 0.f);
		bbox.max.set(0.f, 0.f);
		bool boxInitdYet = false;
		for (int i = 0; i < rig.joints.size(); i++) {
			AnimJoint &joint = rig.joints[i];
			//I'm thoroughly confused by the system I implemented . . .
			//modSprite seems to be the frames where our sprite changes ...
			//what's the bbox otherwise?
			//if modSprite is set ... then find the sprite's bbox
			int spriteIndex = joint.sprite;
			if (spriteIndex < 0 || spriteIndex >= sprites.size()) {
				//complain
			} else {
				AnimSprite &sprite = sprites[spriteIndex];
				box2f sb = sprite.bbox;
				for (int c = 0; c < 4; c++) {
					vec2f corner(
						sprite.bbox[ (c>>0) & 1].x,
						sprite.bbox[ (c>>1) & 1].y);
					corner = vecScaleElem(corner, joint.scale);
					corner = vec2rot(corner, joint.angle);
					corner.x += joint.pos.x;
					corner.y += joint.pos.y;
					if (!boxInitdYet) {
						boxInitdYet = true;
						bbox.min.set(corner);
						bbox.max.set(corner);
					} else {
						bbox.stretch(corner);
					}
				}
			}
		}
		
		//cycle through all 
		//calc the prev & next joints
		for (int i = 0; i < seqs.size(); i++) {
			AnimSkelSeq *seq = &seqs[i];
			for (int j = 0; j < seq->frames.size(); j++) {
				AnimKeyframe *frame = &seq->frames[j];
				for (int k = 0; k < frame->joints.size(); k++) {
					AnimJoint *joint = &frame->joints[k];
					
					joint->prevPos = NULL;
					joint->prevAngle = NULL;
					joint->prevScale = NULL;
					joint->prevSprite = NULL;
					joint->prevUV = NULL;
					
					//now, for each entry in 'joint' ... cycle forward or back to find the prev/next most modified
					int j2 = j;
					while (j2 != (j+1)%seq->frames.size()) {
						AnimKeyframe *checkFrame = &seq->frames[j2];
						AnimJoint *checkJoint = &checkFrame->joints[k];
						if (!joint->prevPos && checkJoint->modPos) joint->prevPos = checkFrame;
						if (!joint->prevAngle && checkJoint->modAngle) joint->prevAngle = checkFrame;
						if (!joint->prevScale && checkJoint->modScale) joint->prevScale = checkFrame;
						if (!joint->prevSprite && checkJoint->modSprite) joint->prevSprite = checkFrame;
						if (!joint->prevUV && checkJoint->modUV) joint->prevUV = checkFrame;
						j2 = (j2 - 1 + seq->frames.size()) % seq->frames.size();
					}
					
					if (!joint->prevPos) joint->prevPos = frame;//&rig;
					if (!joint->prevAngle) joint->prevAngle = frame;//&rig;
					if (!joint->prevScale) joint->prevScale = frame;//&rig;
					if (!joint->prevSprite) joint->prevSprite = frame;//&rig;
					if (!joint->prevUV) joint->prevUV = frame;//&rig;

					
					joint->nextPos = NULL;
					joint->nextAngle = NULL;
					joint->nextScale = NULL;
					joint->nextSprite = NULL;
					joint->nextUV = NULL;
					
					j2 = (j + 1) % seq->frames.size();
					while (j2 != j) {
						AnimKeyframe *checkFrame = &seq->frames[j2];
						AnimJoint *checkJoint = &checkFrame->joints[k];
						if (!joint->nextPos && checkJoint->modPos) joint->nextPos = checkFrame;
						if (!joint->nextAngle && checkJoint->modAngle) joint->nextAngle = checkFrame;
						if (!joint->nextScale && checkJoint->modScale) joint->nextScale = checkFrame;
						if (!joint->nextSprite && checkJoint->modSprite) joint->nextSprite = checkFrame;
						if (!joint->nextUV && checkJoint->modUV) joint->nextUV = checkFrame;
						j2 = (j2 + 1) % seq->frames.size();
					}
					
					if (!joint->nextPos) joint->nextPos = frame;//&rig;
					if (!joint->nextAngle) joint->nextAngle = frame;//&rig;
					if (!joint->nextScale) joint->nextScale = frame;//&rig;
					if (!joint->nextSprite) joint->nextSprite = frame;//&rig;
					if (!joint->nextUV) joint->nextUV = frame;//&rig;
				}
				
				if (j == seq->frames.size()-1) {
					assert(seq->frames.size());
					if (seq->frames.size() == 1) {
						frame->dt = 1.f;
					} else {
						frame->dt = seq->frames[0].time;
					}
				} else {
					frame->dt = seq->frames[j+1%seq->frames.size()].time - frame->time;
				}
			}
		}

		for (int i = 0; i < sprites.size(); i++) {
			AnimSprite *sprite = &sprites[i];

			string result;
			if (!fileSys.search(sprite->filename, game.getSpritePaths(), result)) {
				con << "failed to find file " << sprite->filename << "\n";
				continue;
			}
			sprite->image = texSys.load(result.c_str(), TEXTURE_SPRITE);
			if (!sprite->image.isValid()) con << "failed to load texture at " << result << "\n";
		}
	}
};


// *************************** storage ***************************

static vector<AnimSkelGroup *> animSkelGroupTable;

//the old system had parallel arrays: group which tied into seq, and seq
//the new sys is all localized (sorry repetition!)
int getAnimationSequenceIndex(int groupIndex, const char *seqName) {
	if (groupIndex < 0 || groupIndex >= (int)animSkelGroupTable.size()) return 0;
	AnimSkelGroup *group = animSkelGroupTable[groupIndex];
	if (!seqName) return 0;	//for the 'null's
	for (int i = 0; i < group->seqs.size(); i++) {
		const AnimSkelSeq &s = group->seqs[i];
		if (s.name == seqName) return i;
	}
	//con.print("GetAnimationSequenceIndex() failed to find seq %s for group %s\n", seq_name, group->name.c_str());
	return 0;	//should we default to zero? or should we return an error value?
}

//how about we do validity checks only on setter instead?
//and assert it the rest of the way?
static bool animValidate(
	int *groupIndex, int *seqIndex, int *frameIndex,
	AnimSkelGroup **group, AnimSkelSeq **seq, AnimKeyframe **frame)
{
	if (*groupIndex < 0 || *groupIndex >= (int)animSkelGroupTable.size()) *groupIndex = 0;

	*group = NULL;
	*seq = NULL;
	*frame = NULL;

	if (animSkelGroupTable.size()) {
		*group = animSkelGroupTable[*groupIndex];

		if (*seqIndex < 0 || *seqIndex >= group[0]->seqs.size()) *seqIndex = 0;

		if (group[0]->seqs.size()) {
			*seq = &group[0]->seqs[*seqIndex];

			if (*frameIndex < 0 || *frameIndex >= seq[0]->frames.size()) *frameIndex = 0;

			if (*frameIndex >= 0 && *frameIndex < seq[0]->frames.size()) {
				*frame = &seq[0]->frames[*frameIndex];
				return true;
			}
			return false;
		}
		return false;
	}
	return false;
}

static const char *thisAnimGroup_name = NULL;
static int thisAnimGroup_firstSeq = -1;

static AnimSkelGroup *animGroupFindByName(const char *name) {
	for (int i = 0; i < (int)animSkelGroupTable.size(); i++) {
		AnimSkelGroup *group = animSkelGroupTable[i];
		if (group->name == name) return group;
	}
	return NULL;
}



///////////////// PF_animLoad - now the only lua animation interface ///////////////// 

/*
format:

sprites = {
	{
		file=$filename;
		bbox={...};
	};
	...
};
rig = {
	{
		pos={#x, #y};	-- optional, defaults to {0,0}
		angle=#theta;	-- optional, defaults to 0
		scale={#sx, #sy};	-- optional, defaults to {1,1}
		sprite=#spriteIndex;	-- one-based, optional, defaults to 1
		subtex={...};	-- optional, defaults to {0,0,1,1}
		parent=#parentJointIndex;	-- one-based, optional
	};
	...
};
seqs = {
	$sequenceName = {
		time=#time;
		joint=#jointIndex;	-- one-based
		pos={#x, #y};	-- optional:
		angle=#theta;	-- optional:
		scale={#sx, #sy};	-- optional:
		sprite=#spriteIndex;	-- optional:
		subtex={...};	-- optional:
	};
	...
};
*/

//load state, table index
static void animLoadFromTable(lua_State *newL, int baseTableIndex, AnimSkelGroup *group) {
	assert(group);

//keep all LuaTable's in a separate scope so when their dtor is called
//they unalloc the right one
//likewise every time we go to throw something, be sure to unlink all lua tables
//if only throw() called a different destructor than normal scope endings

	//now we pick out our necessary data
	LuaTable baseTable(newL, baseTableIndex);

	{
		LuaTable sprites;
		baseTable.req("sprites") >> sprites;
		for (sprites.startNext(); sprites.next();) {
			LuaTable sprite(newL);	//sprite is on top
			AnimSprite rs;
			sprite.req("file") >> rs.filename;
			sprite.rw("bbox") >> *(vec4f*)&rs.bbox;	///required? or default to [0,0]->[1,1] ?
			group->sprites.push_back(rs);	//index can be calc'd by offset in the area (index is 1-based, array is 0-based)
		}
	}

	{
		LuaTable rig;
		if (baseTable.rw("rig") >> rig) {
			for (rig.startNext(); rig.next();) {
				LuaTable joint(newL);
				AnimJoint rj;
				joint.rw("pos") >> rj.pos;
				joint.rw("angle") >> rj.angle;
				joint.rw("scale") >> rj.scale;
				rj.sprite = 1; joint.rw("sprite") >> rj.sprite; rj.sprite--;	//correct the index
				joint.rw("subtex") >> rj.uv;
				rj.parent = 1; joint.rw("parent") >> rj.parent; rj.parent--;	//correct the index
				group->rig.joints.push_back(rj);
			}
		} else {	//if no base is provided then make sure at least one is there
			group->rig.joints.resize(1);	//the default ctor will provide for a single-sprite no-skeleton animation
		}
		
		//hmm, rig ... prev and next should point to itself?  or null?
		//it defaults to null, so lets leave it there
		//no one reference ->prev*** ->prev*** or anything or it might break (due to the first deref maybe pointing to the rig)
	}

	//the seqs C++ struct will be the messiest
	//it will have a timestamp, flags for which variables changed, along with a new joint containing those changed entries

	{
		LuaTable seqs;
		if (baseTable.rw("seqs") >> seqs) {
			for (seqs.startNext(); seqs.next();) {
				if (lua_type(newL,-2) != LUA_TSTRING) throw Exception() << "expected seqs to be indexed by string";
				const char *key = lua_tostring(newL,-2);
				if (!key) throw Exception() << "got a null sequence name";
				AnimSkelSeq readSeq;
				readSeq.name = key;
				
				/*
				why don't we have a separate vector for jointIndex ?
				then the state info could just be a vector of integers for what loc we are at
				for each joint.  and we'd just need to interpolate between it and its next.
				but all anim states truly need is a float for the current location.
				which means we should just make a lookup table somewhere else to map
				from the float to the ... idunno ... maybe i'll cycle through each array every frame
				*/

				{
					LuaTable seq(newL);
					for (seq.startNext(); seq.next();) {

						if (lua_type(newL,-2) != LUA_TNUMBER) throw Exception() << "expected seq to be indexed by number";
						LuaTable keyframe(newL);
						
						//first read the time in
						//then find the keyframe it belongs in
						//then overwrite its joint data
						float time;
						keyframe.req("time") >> time;
						if (time < 0) throw Exception() << "time must be non-negative";
						
						//grab our timestep
						//but i still want constant time access ... don't make it a map ...
						int frameIndex = 0;
						for (; frameIndex < readSeq.frames.size(); frameIndex++) {
							AnimKeyframe &frame = readSeq.frames[frameIndex];
							if (frame.time == time) {
								break;
							}
							if (frame.time > time) {
								//then insert it before frame ...
								//this is a linked list operation
								//but resizes only happen on load
								//and using a LL would mean O(N) access time
								//i want to kee my O(1)
								readSeq.frames.resize(readSeq.frames.size()+1);
								//copy forward
								for (int i = readSeq.frames.size()-1; i > frameIndex; i--) {
									readSeq.frames[i] = readSeq.frames[i-1];
								}
								//this whole process asserts in-ordered nature
								//reset to -1 timestamp and to the rig's size
								readSeq.frames[frameIndex].reset(time, group->rig.joints.size());
								break;
							}
						}
						//if you got to the end and still found none ...
						if (frameIndex == readSeq.frames.size()) {
							readSeq.frames.resize(readSeq.frames.size()+1);
							readSeq.frames[frameIndex].reset(time, group->rig.joints.size());
						}
						//now here 
						
						//now frameIndex better be for a frame with matching time
						assert(readSeq.frames[frameIndex].time == time);

						int jointIndex = 1;
						keyframe.req("joint") >> jointIndex;
						jointIndex--;	//correct the index
						if (jointIndex < 0 || jointIndex >= group->rig.joints.size()) throw Exception() << "got an oob joint index: " << (jointIndex+1);
						
						AnimJoint &readDeltaJoints = readSeq.frames[frameIndex].joints[jointIndex];
						readDeltaJoints.modPos = keyframe.rw("pos") >> readDeltaJoints.pos;
						readDeltaJoints.modAngle = keyframe.rw("angle") >> readDeltaJoints.angle;
						readDeltaJoints.modScale = keyframe.rw("scale") >> readDeltaJoints.scale;
						if (readDeltaJoints.modSprite = keyframe.rw("sprite") >> readDeltaJoints.sprite) readDeltaJoints.sprite--;
						readDeltaJoints.modUV = keyframe.rw("subtex") >> readDeltaJoints.uv;
					}
				}
				
				/*
					todo - sort by name.  you should probably store individual vectors first.  then sort by name.
				*/
				
				group->seqs.push_back(readSeq);
			}
		}
		
		//todo - sort seq frames by time.
	}
}

int PF_animLoad(lua_State *baseL) {
	int result = 0;
	lua_State *newL = NULL;
	const char *alias = NULL;
	try {
		int argc = lua_gettop(baseL);
		const char *filename = NULL;
		bool loadingFromTable = false;	//better be true if filename is null
		if (argc == 1) {
			if (lua_type(baseL,1) != LUA_TSTRING) throw Exception() << "expected a string for first param, not " << lua_typename(baseL, lua_type(baseL,1));
			alias = filename = lua_tostring(baseL,1);
		} else {
			//if we get a table as our second param then it's to be interpretted as our state
			//otherwise we assume its a filename
			switch (lua_type(baseL,2)) {
			case LUA_TTABLE:
				alias = lua_tostring(baseL,1);
				//and we don't need a filename ... maybe we should just do the loading here
				//in such a situation, instead of creating a new state and running across its global for our goods
				//we'd instead use the baseL state and this table
				loadingFromTable = true;
				break;

			case LUA_TSTRING:
				alias = lua_tostring(baseL,1);
				filename = lua_tostring(baseL,2);
				break;
			default:
				alias = lua_tostring(baseL,1);
				throw Exception() << "expected a table or string as our second param, not " << lua_typename(baseL, lua_type(baseL,2));
			}
		}
		
		AnimSkelGroup *group = new AnimSkelGroup(alias);

		if (loadingFromTable) {
			animLoadFromTable(baseL, 2, group);
		} else {
			if (!filename) throw Exception() << "expected a filename";
			string result;
			if (!fileSys.search(filename, game.getSpritePaths(), result)) throw Exception() << "couldn't find file " << filename;
			
			//first see if it is a picture or not...
			//if it is then give it a single-frame [-.5,.5]^2 sprite sequence
			
			bool sfmeplz = false;
			int len = strlen(filename);
			if (len > 4) {
				sfmeplz = 
					!strcasecmp(filename + len - 4, ".bmp") ||
					!strcasecmp(filename + len - 4, ".tga") ||
					!strcasecmp(filename + len - 4, ".jpg") ||	//do i have .jpeg to be picked up by my image loader?
					!strcasecmp(filename + len - 4, ".png");
			} else if (len > 5) {
				sfmeplz = !strcasecmp(filename + len - 5, ".jpeg");
			}
			if (sfmeplz) {
				//here's our first sprite
				group->sprites.resize(1);
				group->sprites[0].filename = filename;
				group->rig.joints.resize(1);	//default ctor is good enuf
			} else {
			
				newL = lua_open();
				if (!newL) throw Exception() << "failed to create new lua state";

				{	//filebuffer scope of life
					FileBuffer fb;
					if (!fb.load(result.c_str())) throw Exception() << "failed to load file " << filename;
					if (!fb.getData()) throw Exception() << "had some problem reading the data of file " << filename;
					
					//now process the file here
					//and store it as the group in 'alias'
					
					//1) run it as a lua state
					//2) pick it apart

					const char *buffer = fb.getData();
					size_t bufferSize = fb.getSize();
					
					//oh shoot, we want a new lua state for this btw.  don't muddy up our global namespace.
					if (luaL_loadbuffer(newL, buffer, bufferSize, "anim")) {
						throw Exception() << "for loading animation " << filename << " :\n" << buffer << "\n\n"
							<< " we got some errors:\n"
							<< lua_tostring(newL, -1) << "\n"
							<< lua_tostring(newL, -2) << "\n";
					}
				}

				int base = lua_gettop(newL);
				lua_pushliteral(newL, "_TRACEBACK");
				lua_rawget(newL, LUA_GLOBALSINDEX);
				lua_insert(newL, base);
				if (lua_pcall(newL, 0, 0, 0)) throw Exception() << "got lua error:\n" << lua_tostring(newL, -1);	//display the error
				lua_remove(newL, base);

				//load state, table index
				animLoadFromTable(newL, LUA_GLOBALSINDEX, group);
				
				lua_close(newL);
				newL = NULL;	//so the throw catch doesn't dealloc it
			}
		}
		
		group->finalize();

		//here we store all the goods we just accumulated ...

		animSkelGroupTable.push_back(group);

		//return 'ok' for kicks
		lua_pushboolean(baseL,true);
		result = 1;
		
	} catch (Exception &t) {
		con << "PF_animLoad failed";
		if (alias) con << " on animation " << alias;
		con << ": " << t << "\n";
	}
	if (newL) lua_close(newL);
	//returning nothing means we failed
	return result;
}

///////////////// AnimState ///////////////// 

int AnimState::getAbsFrame() const {
	return animSys->getAbsFrame(group,seq,frame);
}

bool AnimState::setFromAbsFrame(int absFrame) {
	if (absFrame < 0) {
		//error bad absFrame
		return false;
	}

	group = absFrame % animSkelGroupTable.size();
	absFrame /= animSkelGroupTable.size();
	AnimSkelGroup *groupObj = animSkelGroupTable[group];
	
	seq = absFrame % groupObj->seqs.size();
	absFrame /= groupObj->seqs.size();
	AnimSkelSeq *seqObj = &groupObj->seqs[seq];
	
	frame = absFrame;
	if (frame >= seqObj->frames.size()) {	//this will fail of we were encoded wrong
		frame = 0;
		//error bad absFrame
		return false;
	}
	return true;
}

float AnimState::getFrameDelay() const {
	int group = this->group, seq = this->seq, frame = this->frame;		//new state so we dont mess up our goods
	AnimSkelGroup *groupPtr;
	AnimSkelSeq *seqPtr;
	AnimKeyframe *framePtr;
	if (!animValidate(&group, &seq, &frame, &groupPtr, &seqPtr, &framePtr)) {
		con.print("failed to validate group %d seq %d frame %d\n", group, seq, frame);
		return 1.f;	//if we failed to get a valid group/seq/frame
	}
	if (frame < 0 || frame >= seqPtr->frames.size()) {
		con.print("got an oob frame %d\n", frame);
		return 1.f;	//what about the last frame?
	}
	
	return seqPtr->frames[frame].dt;
}

bool AnimState::validateFrame() {
	AnimSkelGroup *groupPtr;
	AnimSkelSeq *seqPtr;
	AnimKeyframe *framePtr;
	return animValidate(&group, &seq, &frame, &groupPtr, &seqPtr, &framePtr);
}

bool AnimState::getBBox(box2f *store) const {
	int group = this->group, seq = this->seq, frame = this->frame;		//new state so we dont mess up our goods
	AnimSkelGroup *groupPtr;
	AnimSkelSeq *seqPtr;
	AnimKeyframe *framePtr;
	if (!animValidate(&group, &seq, &frame, &groupPtr, &seqPtr, &framePtr)) return false;	//if we failed to get a valid group/seq/frame
	//for the animation info
	//currently derived from the bbox
	*store = groupPtr->bbox;
	return true;
}

///////////////// AnimInfo ///////////////// 

void AnimInfo::setGroup(const char *groupName) { 
//doesnt force-set if we are on same group
	int newGroupIndex = animSys->getGroupIndex(groupName);
	if (newGroupIndex == group) return;			//no worries.  already done.
	group = newGroupIndex;			//set the new group index
	//dont use SetAnimationSequence because we want to force the frame index to reset, rather htan it being conditional
	seq = 0;							//and reset the rest.
	idleSeq = 0;
	frame = 0;
	frameNextTime = game_thistime + getFrameDelay();	//setup frameNextTime 
}

void AnimInfo::setSeq(const char *seqName) {
	setSeqIndex(getAnimationSequenceIndex(group, seqName));
}

void AnimInfo::setSeqIndex(int _seq) {
	if (_seq == seq) return;
	seq = _seq;
	frame = 0;
	frameNextTime = game_thistime + getFrameDelay();	//reset counter too
}

bool AnimInfo::setFromAbsFrame(int absFrame) {
	bool ret = super::setFromAbsFrame(absFrame);
	frameNextTime = game_thistime + getFrameDelay();
	return ret;
}

void AnimInfo::setIdleSeq(const char *idleSeqName) {
	setIdleSeqIndex(getAnimationSequenceIndex(group, idleSeqName));
}


void AnimInfo::setIdleSeqIndex(int _idleSeq) {
	idleSeq = _idleSeq;
}

void AnimInfo::incFrame() {
	if (group < 0 || group >= animSkelGroupTable.size()) return;
	AnimSkelGroup *g = animSkelGroupTable[group];
	
	if (seq < 0 || seq >= g->seqs.size()) {
		seq = idleSeq;
		frame = 0;
		return;	//oob seq ... should we switch to the idle seq?
	}
	AnimSkelSeq &s = g->seqs[seq];
	
	frame++;	
	if (frame >= s.frames.size()-1) {	//the last frame is a marker for interpolation and has a zero duration
		frame = 0;						//this isn't the case in zero-frame sprites, which won't be hit cuz this doesn't get hit til the frame is incremented
		seq = idleSeq;
	}
	
	float delay = getFrameDelay();
	
	
	//set the frameNextTime as well
	frameLastTime = game_thistime;
	frameNextTime = game_thistime + delay;
}


///////////////// AnimSys ///////////////// 

//to be called once and only once upon program construction
bool AnimSysImpl::sessionInit() {
	AnimSkelGroup *group = new AnimSkelGroup("null");
	group->finalize();
	
	//without adding the null group to the anim group table we can't easily hide our renderables...
	//when we do add it for some odd reason the lights all crap out
	animSkelGroupTable.push_back(group);

	//without adding a sprite the sprite renderer complains that it is invalid
	//but adding it is redundant since its a null sprite anyways ...
//		group->sprites.resize(1);
//		group->sprites[0].filename = "";
//		group->sprites[0].image = NULL;
	return true;
}

void AnimSysImpl::sessionKill() {
	VECTOR_DELETE_ALL(animSkelGroupTable);
}

/*
position, depth, rotation, scale will all be globally (last) applied
but until then we need to pick out the frame (which can now be a float ftw)
and from which do all our kinematics calculations 
*/
void AnimSysImpl::render(
	float x, float y, float z,
	float r, float sx, float sy,
	vec4uc color,
	unsigned int shader, Blend blendfunc, bool postFBO, bool useLighting,
	bool flip, int absFrame, bool immediate)
{
	//hack for the null frame
	if (!absFrame) return;

	//from absFrame we need to derive the frame, sequence, and group ... objects ... ?
	AnimState state;
	if (!state.setFromAbsFrame(absFrame)) {
		con << "bad abs frame\n";
		return;
	}
	AnimSkelGroup *groupObj = animSkelGroupTable[state.group];
	AnimSkelSeq *seqObj = &groupObj->seqs[state.seq];
	//find the frame that our time is within ... "absFrame" ...
	//this presents a dilemma ...
	//should absFrame's third value [group,seq,???] store the frame index or the time index?
	//now that frame indices are a bit more abstracted...
	//and storing the frame index would mean the animation system would increment by 1/dt depending on the frame ...
	//which would present subtle numerical inaccuracies
	AnimKeyframe *frameObj = &seqObj->frames[state.frame];
	
	vec2f rdir = vec2expd<float>(r);
	
	//flip
	float flipscale = flip ? -1.f : 1.f;

	//start with a matrix for our object orientation:
	mat22f objectrxform;
	vec2f objecttxform;
	objectrxform(0,0) = sx * rdir.x * flipscale;
	objectrxform(1,0) = sx * rdir.y;
	objectrxform(0,1) = -sy * rdir.y * flipscale;
	objectrxform(1,1) = sy * rdir.x;
	objecttxform.set(x,y);
	
	//assuming the parent joint always has a lower index ...
	for (int j = 0; j < frameObj->joints.size(); j++) {
		//for all the joints ...
		//we're gonna need ... for each AnimJoint
		//		we'll need the prev-most and next-most that have their mod flags set
		//		precache these as 'prev***' and 'next***'
		AnimJoint *joint = &frameObj->joints[j];
		
		//here's another issue ... parent coordinate frames.
		//it will really become big once I incorperate interpolation 
		//for each joint we'll need the parent joint info to have already been computed ...
		//the interpolated parent joint info ...
		//of the prev*** and next*** joints
		
		float curTime = frameObj->time;	//todo - store/find the current time
		
		//note that 'dt' will have to be stored/calc'd for each mod frame
		//we can't just subtract times in case they do loop around the end
		//give the rig a unit dt so we don't div by zero
		
		float posDt = joint->prevPos == joint->nextPos ? frameObj->dt : joint->nextPos->time - joint->prevPos->time;
		float posLerp = (curTime - joint->prevPos->time) / posDt;
		vec3f pos = vecLerp(joint->prevPos->joints[j].pos, joint->nextPos->joints[j].pos, posLerp);
		
#define LERP(a,b,t)	(b * t + a * (1. - t))
		float angleDt = joint->prevAngle == joint->nextAngle ? frameObj->dt : joint->nextAngle->time - joint->prevAngle->time;
		float angleLerp = (curTime - joint->prevAngle->time) / angleDt;
		float angle = LERP(joint->prevAngle->joints[j].angle, joint->nextAngle->joints[j].angle, angleLerp);
		
		float scaleDt = joint->prevScale == joint->nextScale ? frameObj->dt : joint->nextScale->time - joint->prevScale->time;
		float scaleLerp = (curTime - joint->prevScale->time) / scaleDt;
		vec2f scale = vecLerp(joint->prevScale->joints[j].scale, joint->nextScale->joints[j].scale, scaleLerp);
		
		float uvDt = joint->prevUV == joint->nextUV ? frameObj->dt : joint->nextUV->time - joint->prevUV->time;
		float uvLerp = (curTime - joint->prevUV->time) / uvDt;
		box2f uv(
			vecLerp(joint->prevUV->joints[j].uv.min, joint->nextUV->joints[j].uv.min, uvLerp),
			vecLerp(joint->prevUV->joints[j].uv.max, joint->nextUV->joints[j].uv.max, uvLerp));

		//why would you lerp sprites?  how about after 50% we switch?
		int spriteIndex = joint->prevSprite->joints[j].sprite;
		if (joint->prevSprite->joints[j].sprite != joint->nextSprite->joints[j].sprite) {
			float spriteDt = joint->prevSprite == joint->nextSprite ? frameObj->dt : joint->nextSprite->time - joint->prevSprite->time;
			float spriteLerp = (curTime - joint->prevSprite->time) / spriteDt;
			if (spriteLerp >= .5f) {
				spriteIndex = joint->nextSprite->joints[j].sprite;
			}
		}
		
		if (spriteIndex < 0 || spriteIndex >= groupObj->sprites.size()) {
			con << "bad sprite\n";
			continue;
		}
		
		//now construct the joint's local xform
		vec2f jointAngleDir = vec2expd<float>(angle);
		mat22f localrxform;
		vec2f localtxform;
		localrxform(0,0) = scale.x * jointAngleDir.x;
		localrxform(1,0) = scale.x * jointAngleDir.y;
		localrxform(0,1) = -scale.y * jointAngleDir.y;
		localrxform(1,1) = scale.y * jointAngleDir.x;
		localtxform = pos;
		
		//and right-mul the joint with its parent to get the new xform
		//if the joint has no parent then right-mul it to the object xform
		if (joint->parent >= 0 && joint->parent < frameObj->joints.size()) {
			AnimJoint *parentJoint = &frameObj->joints[joint->parent];
			joint->txform = operator*<2,2,float>(parentJoint->rxform, localtxform) + parentJoint->txform;
			joint->rxform = parentJoint->rxform * localrxform;
		} else {
			joint->txform = operator*<2,2,float>(objectrxform, localtxform) + objecttxform;
			joint->rxform = objectrxform * localrxform;
		}

		
		AnimSprite *spriteObj = &groupObj->sprites[spriteIndex];

		if (!immediate) {
			//now we can draw our joint
			EGeometry *g = renderSys.newGeom();

			//this is a whole new world now ... rescales have to be taken into account too
			g->dir = vec2expd<float>(r + angle);
			
			
			for (int i = 0; i < 4; i++) {
			
				//texcoord
				g->texcoord[i].set(
					uv[(i&1)^((i>>1)&1)].x,
					uv[(i&2)>>1].y);
					
				//joint coordinate frame

				//vtx
				g->vertex[i].set(
					spriteObj->bbox[(i&1)^((i>>1)&1)].x,
					spriteObj->bbox[(i&2)>>1].y);
					
				//transform by the joint coordinate space
				g->vertex[i] = operator*<2,2,float>(joint->rxform, g->vertex[i]) + joint->txform;
			}
			
			g->depth = pos.z + z;
			TexLoad *cm = texSys.getObj(spriteObj->image);
			g->colormap = cm ? cm->tex.getTextureID() : 0;
			g->bumpmap = cm ? cm->bumpmap.getTextureID() : 0;
			g->color = color;
			g->shader = shaderSys.getIndex(shader);
			g->blend = blendfunc;
			g->use_alpha_test = true;										//always use alpha testing with our sprites
			g->use_bump_mapping = BlendDFactor(blendfunc) != GL_ONE;		//only use bump mapping when we arent adding to the background (make this a drawSprite flag?
			g->use_lighting = useLighting;
			g->postFBO = postFBO;
			g->flip = flip ? -1 : 1;
			
			renderSys.geomDone(g);
		} else {
			//TODO - push glEnabled and the current texture2d
			//now gl based...
			TexLoad *cm = texSys.getObj(spriteObj->image);
			if (cm) {
				glBindTexture(GL_TEXTURE_2D, cm->tex.getTextureID());
			}
			glBegin(GL_QUADS);
			for (int i = 0; i < 4; i++) {
				glTexCoord2f(
					uv[(i&1)^((i>>1)&1)].x,
					uv[(i&2)>>1].y);
				vec2f v(
					spriteObj->bbox[(i&1)^((i>>1)&1)].x,
					spriteObj->bbox[(i&2)>>1].y);
				v = operator*<2,2,float>(joint->rxform, v) + joint->txform;
				glVertex3f(v.x, v.y, pos.z + z);
			}
			glEnd();
			if (cm) {
				glBindTexture(GL_TEXTURE_2D, 0);
			}
		}
	}
}
	
//how do you make a function accept pointers to const strings
//without making the parameter a const itself (thus making it non-writable) ?

int AnimSysImpl::getGroupIndex(const char *group) {
	if (!group) return 0;
	for (int i = 0; i < (int)animSkelGroupTable.size(); i++) {
		AnimSkelGroup *g = animSkelGroupTable[i];
		if (g->name == group) return i;
	}
	return 0;
}

bool AnimSysImpl::getGroupName(int group, string &result) { 
	result = "";
	if (group < 0 || group >= (int)animSkelGroupTable.size()) return false;
	AnimSkelGroup *g = animSkelGroupTable[group];
	result = g->name;
	return true;
}

//used by world.cpp for name writing to files
bool AnimSysImpl::getSeqName(int group, int seq, std::string &result) {
	result = "";
	if (group < 0 || group >= (int)animSkelGroupTable.size()) return false;
	AnimSkelGroup *g = animSkelGroupTable[group];
	if (seq < 0 || seq >= (int)g->seqs.size()) return false;
	result = g->seqs[seq].name;
	return true;
}
	
//factors ...
//this isn't 1-1
//there will be holes in the mapping: integers that don't map back to group/seq/frames ...
//but we can test for that
/*
encode:
	x = groupIndex + groupCount * (seqIndex + seqCount * frameIndex)
decode:
	assert(x>=0)
	groupIndex = x % groupCount
	x /= groupCount
	seqIndex = x % seqCount
	x /= seqCount
	frameIndex = x
	assert(frameIndex < frameCount)	//this will fail of we were encoded wrong
*/
int AnimSysImpl::getAbsFrame(int groupIndex, int seqIndex, int frameIndex) {
	//with no 'this' ref this might as well be static
	if (groupIndex < 0 || groupIndex >= animSkelGroupTable.size()) {
		con << "animSys.getAbsFrame got a bad groupIndex: " << groupIndex << "\n";
		return 0;
	}
	AnimSkelGroup *groupObj = animSkelGroupTable[groupIndex];
	assert(0 <= seqIndex && seqIndex < groupObj->seqs.size());
	AnimSkelSeq *seq = &groupObj->seqs[seqIndex];
	return groupIndex + animSkelGroupTable.size() * (seqIndex + groupObj->seqs.size() * frameIndex);
}

AnimSys::AnimSys() : impl(new AnimSysImpl()) {}
AnimSys::~AnimSys() { delete impl; }
bool AnimSys::sessionInit() { return impl->sessionInit(); }
void AnimSys::sessionKill() { impl->sessionKill(); }
void AnimSys::render(
		float x, float y, float z,
		float r, float sx, float sy,
		vec4uc color,
		unsigned int shader, Blend blendfunc, bool postFBO, bool useLighting,
		bool flip, int absFrame, bool immediate)
{ impl->render(x,y,z,r,sx,sy,color,shader,blendfunc,postFBO,useLighting,flip,absFrame,immediate); }
int AnimSys::getGroupIndex(const char *group) { return impl->getGroupIndex(group); }
bool AnimSys::getGroupName(int group, string &result) { return impl->getGroupName(group, result); }
bool AnimSys::getSeqName(int group, int seq, std::string &result) { return impl->getSeqName(group, seq, result); }
AnimSys animSys;

