/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <assert.h>

#include "base/base.h"
#include "system/memory.h"
#include "system/exception.h"

#include "entalloc.h"
#include "entconn.h"
#include "entity.h"
#include "client.h"
#include "spawnobj.h"	//spawnObjArrayUnlinkEnt
#include "game/main.h"
#include "game/client.h"

// **************** cvars ****************

void DumpEnts(int argc, char **argv);
void EntCount(int, char **) { con.print("entCount = %d\n", entAllocSys.getCount()); }

static CVarw *local_cvars[] = {
	new CVarw(0,						"dump_ents",	DumpEnts),
	new CVarw(0,						"entCount",		EntCount),
};

// **************** memory structures ****************

//ENT_MEM_SIZE is the size of the entity data, split into ENT_BLOCKS number of blocks sized ENT_BLOCK_SIZE
//for each block there is a 4-byte index set aside in the entity_block_offset[] array
//this has valid data from 0 to entity_count.  each valid data is the index of the first block
//of each entity.  THIS array will give you constant-time access to any entity.
//no guarentee on contiguous mapping from entity_block_offset[] to entity_data[]
//but there is a guarantee that all entitites which span multiple blocks will hold contiguous blocks.
//
//ENT_BLOCKFLAGS_SIZE is (ENT_BLOCKS / 8) * 2 to represent there being two bits per ENT_BLOCKS byte
//
//entptr_t clarification!
//entptr_t must always reference the same entity, regardless of creation/removal in the ent list
//therefore entptr_t's will reference block indices, not header indices
//therefore enumerations across entptr_t's are a bad idea
class EntBlockHeader {
public:
	int index;		//which block to start our entity at
	int size;		//number of bytes this entity requires.  div block size, round up, to get block count
};

//array byte sizes
#define ENTITY_BLOCK_HEADER_SIZE	(sizeof(EntBlockHeader) * ENTITY_BLOCK_COUNT)
#define ENTITY_BLOCKFLAGS_SIZE		(ENTITY_BLOCK_COUNT / 8) * 2			//1024 bytes for flags


class EntAllocSysImpl {
protected:

	//client state data
	//on the server, this holds the last state, and tests this state to any changes
	//on the client, this holds the latest data recieved
	char *cl_entity;

	//client info state data
	//same idea as cl_entity, but for client-specific stuff:
	char *cl_info;

	//the raw data within to hold all entities
	char *entData;

public:	//for DumpEnts
	//flags for each block of entity data
	//	- whether each block is filled or not
	//	- and whether it is the beginning of an entity structure or not
	char *entity_blockflags;

	//headers to each entity entry within the data structure
	//each header contains a start block index and a block count value
	EntBlockHeader *entity_block_header;	

protected:
	//the size of the wwwhhhhooolllee array
	//replacement for the old CL_ENTITY_SIZE and CL_INFO_SIZE defs, since they are no longer static
	int cl_entity_size;
	int cl_info_size;

public:
	EntAllocSysImpl();
	
	bool sysInit();
	void sysKill();
	void clearAll();
	void clearClientData();

	int getCount();
	char *getClientEntityArray();
	char *getClientInfoArray();

	char getBlockFlags(int block_index);
	const EntBlockHeader *getEntBlockHeader(int index);
	bool isActive(entptr_t e);
	Entity *entForIndex(int i);
	Entity *entForPtr(entptr_t ep);
	entptr_t ptrForEnt(Entity *e);

	void *newEnt(size_t size);
	void deleteEnt(void *_e);

	int entCount;
};

// **************** memory structure routines ****************

void DumpEnts(int argc, char **argv) {
	con.print("DumpEnts() begin\n");
	con.print("entCount: %d\n", entAllocSys->entCount);

	con.print("blockflags: ");
	int i;
	for (i = 0; i < ENTITY_BLOCK_COUNT; i++) {
		int bit_offset = (i & 3) << 1;
		int flags = (entAllocSys->entity_blockflags[i >> 2] & (3 << bit_offset)) >> bit_offset;
		con.print("{%d|%d},", (flags & 1), (flags & 2) >> 1);
	}
	con.print("\n");

	con.print("headers: ");
	for (i = 0; i < ENTITY_BLOCK_COUNT; i++) {
		con.print("{%d,%d},", entAllocSys->entity_block_header[i].index, entAllocSys->entity_block_header[i].size);
	}
	con.print("\n");

	con.print("DumpEnts() end\n");
}

// **************** serverside routines ****************

EntAllocSysImpl::EntAllocSysImpl() 
:	cl_entity(NULL),
	cl_info(NULL),
	entData(NULL),
	entity_blockflags(NULL),
	entity_block_header(NULL),
	cl_entity_size(0),
	cl_info_size(0)
{}

const EntBlockHeader *EntAllocSysImpl::getEntBlockHeader(int index) {
	if (index < 0 || index >= ENTITY_BLOCK_COUNT) return NULL;
	return entity_block_header + index;
}

char EntAllocSysImpl::getBlockFlags(int block_index) {
	int bit_offset = (block_index & 3) << 1;
	int byte_offset = block_index >> 2;
	return (entity_blockflags[byte_offset] >> bit_offset) & 3;
}

// **************** cl/sv misc routines ****************

BEGIN_CHILDCLASS(EntNull, Entity)
	EntNull() { clear(); }
	EntNull(SpawnObj *s) { clear(); }
	
	//clear all variables of Entity -- ESPECIALLY the active flag (if nothing else)
	void clear() {
		active = false;
		pos = 0.f;
		layer = 0;
		spawnobj = NULL;
	}
ENDCLASS()

static Entity entNull;

char *EntAllocSysImpl::getClientEntityArray() { return cl_entity; }
char *EntAllocSysImpl::getClientInfoArray() { return cl_info; }

Entity *EntAllocSysImpl::entForPtr(entptr_t ep) {

	//and dont forget to clear the null entity =) the ctor doesnt seem to be kicking in...
	//of course, this is engine-side, while the ctor is ONLY supposed to be executing game-side...
//	static Entity entNull;
//	static bool init_d = false;
//	if (!init_d) {
//		memset(&entNull, 0, sizeof(entNull));
//		init_d = true;
//	}

	int index = ep.i;
	if (index < 0 || index >= ENTITY_BLOCK_COUNT) 
//		return &entNull;	//this messes with our alloc system 
		return NULL;		//i wonder what this will do

	Entity *e = (Entity *)(entData + ep.i * ENTITY_BLOCK_SIZE);
	if (!e->active) return &entNull;

	//TODO - keep the 'block begin' bit and only return an entity if that bit is set!
	return e;
}

entptr_t EntAllocSysImpl::ptrForEnt(Entity *e) {
	entptr_t i = {-1};
	if (e) {
		char *ce = (char *)e;
		int index = ce - entData;
		if (index % ENTITY_BLOCK_SIZE) return i;
		index /= ENTITY_BLOCK_SIZE;
		if (index < 0 || index >= ENTITY_BLOCK_COUNT) return i;
		i.i = index;
	}
	return i;
}

bool EntAllocSysImpl::isActive(entptr_t e) {
//	if (e.i < 0 || e.i >= numberof(entity_list)) return false;
	if (e.i < 0 || e.i >= ENTITY_BLOCK_COUNT) return false;
	//TODO - keep the 'block begin' bit and only return an entity if that bit is set!
	return entForPtr(e)->active;
}

Entity *EntAllocSysImpl::entForIndex(int i) {
//still isnt a fully asserted call...
//	assert(i >= 0 && i < entAllocSys->entCount);
	if (i < 0 || i >= entCount) {
//		con << "entGetIndex got OOB entity " << i << "\n";
		return NULL;
	}
	entptr_t e = {entity_block_header[i].index};
	return entForPtr(e);
}

bool EntAllocSysImpl::sysInit() {

	con.print("allocating entData sized %d\n", ENTITY_DATA_SIZE);
	entData = (char *)MEMALLOCNAMED(ENTITY_DATA_SIZE, "entity data");
	if (!entData) return false;

	con.print("allocating entity_blockflags sized %d\n", ENTITY_BLOCKFLAGS_SIZE);
	entity_blockflags = (char *)MEMALLOCNAMED(ENTITY_BLOCKFLAGS_SIZE, "entity blockflags");
	if (!entity_blockflags) return false;

	con.print("allocating entity_block_header sized %d\n", ENTITY_BLOCK_HEADER_SIZE);
	entity_block_header = (EntBlockHeader *)MEMALLOCNAMED(ENTITY_BLOCK_HEADER_SIZE, "entity block header");
	if (!entity_block_header) return false;

#ifdef ONE_DELTA_ENTITY_BUFFER
	cl_entity_size = sizeof(ClientEntity) * ENTITY_BLOCK_COUNT;
	con.print("allocating cl_entity sized %d\n", cl_entity_size);
	cl_entity = (char *)MEMALLOCNAMED(cl_entity_size, "cl entity");
	if (!cl_entity) return false;
#else
//#define CL_ENTITY_SIZE				(sizeof(ClientEntity) * CL_ENTITY_NUM_PER_CLIENT * numberof(serverConnSys.connections))
	cl_entity_size = sizeof(ClientEntity) * CL_ENTITY_NUM_PER_CLIENT * numberof(serverConnSys.connections);
	con.print("allocating cl_entity sized %d\n", cl_entity_size);
	cl_entity = (char *)MEMALLOCNAMED(cl_entity_size, "cl entity");
	if (!cl_entity) return false;
#endif

	entAllocSys.clearAll();

	return true;
}

void EntAllocSysImpl::sysKill() {
	if (entData) memFree(entData);
	if (entity_blockflags) memFree(entity_blockflags);
	if (entity_block_header) memFree(entity_block_header);
	if (cl_entity) memFree(cl_entity);

	entData = NULL;
	entity_blockflags = NULL;
	entity_block_header = NULL;
	cl_entity = NULL;

	entAllocSys->entCount = 0;
}
	
void EntAllocSysImpl::clearAll() {
	assert(entData);
	assert(entity_blockflags);
	assert(entity_block_header);
	assert(cl_entity);

	//reset everything
	memset(entData, 0, ENTITY_DATA_SIZE);
	memset(entity_blockflags, 0, ENTITY_BLOCKFLAGS_SIZE);
	memset(entity_block_header, 0, ENTITY_BLOCK_HEADER_SIZE);
	memset(cl_entity, 0, cl_entity_size);

	entAllocSys->entCount = 0;
}

void EntAllocSysImpl::clearClientData() {
	memset(cl_entity, 0, cl_entity_size);
}

int EntAllocSysImpl::getCount() { 
	return entCount;
}

void *EntAllocSysImpl::newEnt(size_t size) {

	//will 'this' references work even before 'new' is called?
	//is the virtual function table pointer set yet?
	//if it is then would my memset(x,0,y) clear it?
	//if not, wouldn't the this->getClass() here fail?
	//
	//answer: the __vfptr is not yet set
	//this is because 'this' doesnt exist
	//this is true because 'operator new' is a static method
	//therefore just set up getClass() as a static method too

	assert(entAllocSys->entCount <= ENTITY_BLOCK_COUNT);

	int numblocks = size / ENTITY_BLOCK_SIZE;
	if (numblocks * ENTITY_BLOCK_SIZE < (int)size) numblocks++;	//round up

	int foundblocks = 0, firstblock = 0;

	//now hunt through the blockflags for 'numblocks' consecutive blocks
	for (int i = 0; i < ENTITY_BLOCK_COUNT; i++) {
		int bit_offset = (i & 3) << 1;
		int flags = (entity_blockflags[i >> 2] & (3 << bit_offset)) >> bit_offset;
		if (!flags) {
			if (!foundblocks) firstblock = i;
			foundblocks++;

			if (foundblocks == numblocks) {
				//firstblock is our block to use - set it and all the other blocks up to 'i'

				for (int j = firstblock; j <= i; j++) {
					int j_flags = j == firstblock ? 3 : 1;
					int j_bit_offset = (j & 3) << 1;
					entity_blockflags[j >> 2] &= ~(3 << j_bit_offset);	//clear the 'j' flag
					entity_blockflags[j >> 2] |= j_flags << j_bit_offset;		//and set it
				}

				entity_block_header[entAllocSys->entCount].index = firstblock;
				entity_block_header[entAllocSys->entCount].size = (int)size;
				entAllocSys->entCount++;

				char *ent_ptr = entData + firstblock * ENTITY_BLOCK_SIZE;

				//here is our most recent mess
				//we can't memset class data that uses inheritance
				//because this wipes our _vfptr table pointer away . . . hmm . . .
				//how can we zero all the data and still get back the entity that we want?
				//...
				//by overriding the class 'new' operator
				memset(ent_ptr, 0, numblocks * ENTITY_BLOCK_SIZE);
				return ent_ptr;
			}
		} else {
			foundblocks = 0;
		}
	}

//	DumpEnts(0,0);
	//I think the only way to handle operator new errors is via throw
	//...otherwise the bad pointer gets forwarded to the object's ctor
	throw Exception() << "Failed to create entity of size " << size;
	return NULL;
}

void EntAllocSysImpl::deleteEnt(void *_e) {
	Entity *e = (Entity *)_e;

	int j;

	if (!e) {
		con.print("called 'delete' on a null entity!\n");
		return;
	}

	if (!e->active) {
		con.print("called 'delete' on an inactive entity!\n");
		return;
	}

	e->active = 0;

	entptr_t ep = ptrForEnt(e);

	//and all spawn pointers:
	spawnObjSys.unlinkEnt(ep);

	//now... remove the active flags from the entity
	//entity_data
	//entity_blockflags
	//entity_block_header
	//entCount

	//find the block number for our entity
	int blockno = ((char *)e - entData) / ENTITY_BLOCK_SIZE;
	if (blockno < 0 || blockno >= ENTITY_BLOCK_COUNT) {
		con << "TRIED TO REMOVE AN OUT OF BOUNDS ENTITY BLOCK: "<< blockno <<" ptr "<< e <<"\n";
		return;
	}

//	con.print("removing entity from block no. %d\n", blockno);

	//find the header that points to that block (todo - store this in the Entity base class code)
	for (j = 0; j < entAllocSys->entCount; j++) {
		if (entity_block_header[j].index == blockno) {
			//clear all the flags

			//get the block count
			int size = entity_block_header[j].size;
			int count = size / ENTITY_BLOCK_SIZE;
			if (count * ENTITY_BLOCK_SIZE < size) count++;	//round up
			int start = entity_block_header[j].index;
			int end = start + count;

			for (int k = start; k < end; k++) {
				entity_blockflags[k >> 2] &= ~(3 << ((k & 3) << 1));
			}

			//and remove the header from the list...
			//this may cause problems if entptr_t's reference headers!
			//because header indices will move upon creation/removal of entities!
			//therefore have entptr_t's reference block indices!
			//therefore dont iterate across entptr_t numbers!

			for (; j < entAllocSys->entCount-1; j++) {
				entity_block_header[j].index = entity_block_header[j+1].index;
				entity_block_header[j].size = entity_block_header[j+1].size;
			}
			entity_block_header[j].index = 0;
			entity_block_header[j].size = 0;
			entAllocSys->entCount--;

			return;
		}
	}

	con.print("Failed to remove entity from block %d!\n", blockno);

	return;
}

EntAllocSys::EntAllocSys() : impl(new EntAllocSysImpl()) {}
EntAllocSys::~EntAllocSys() { delete impl; }
bool EntAllocSys::sysInit() { return impl->sysInit(); }
void EntAllocSys::sysKill() { impl->sysKill(); }
void EntAllocSys::clearAll() { impl->clearAll(); }
void EntAllocSys::clearClientData() { impl->clearClientData(); }
int EntAllocSys::getCount() { return impl->getCount(); }
char *EntAllocSys::getClientEntityArray() { return impl->getClientEntityArray(); }
char *EntAllocSys::getClientInfoArray() { return impl->getClientInfoArray(); }
char EntAllocSys::getBlockFlags(int block_index) { return impl->getBlockFlags(block_index); }
bool EntAllocSys::isActive(entptr_t e) { return impl->isActive(e); }
Entity *EntAllocSys::entForIndex(int i) { return impl->entForIndex(i); }
Entity *EntAllocSys::entForPtr(entptr_t e) { return impl->entForPtr(e); }
entptr_t EntAllocSys::ptrForEnt(Entity *e) { return impl->ptrForEnt(e); }
void *EntAllocSys::newEnt(size_t size) { return impl->newEnt(size); }
void EntAllocSys::deleteEnt(void *_e) { impl->deleteEnt(_e); }
EntAllocSys entAllocSys;
