/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <stdio.h>
#include <math.h>
#include <iostream>

#include "math/vec.h"
#include "system/gl.h"
#include "base/cvar.h"

#include "inspiration/game/main.h"
#include "inspiration/game/physics.h"
#include "inspiration/game/programmacros.h"
#include "inspiration/world.h"
#include "inspiration/entalloc.h"
#include "inspiration/main.h"

#ifdef WIN32
#include <float.h>
#define isnan _isnan
#endif

using namespace std;

static CVari collisionMaxTests("collisionMaxTests", 10);
static CVari debugDisplayCollision("debugDisplayCollision", 0);
static CVarf planePushBack("planePushBack", 0.00001f);
static CVarf baseFrictionScalar("baseFrictionScalar", 70.f);
static CVarf minVelocity("minVelocity", 0.00000001f);	//near zero
static CVarf maxVelocity("maxVelocity", 500.f);
static CVarf faceStepEpsilon("faceStepEpsilon", 0);//not helping us.			//0.0003125f;//0.03125f;	//0.0000003125;		//keep this very VERY small or else tracelines over far distances will fail
static CVarf collisionEpsilon("collisionEpsilon", 0.f);	//not helping either

static vec3f default_planes[4] = {
	vec3f(-1,0,0),
	vec3f(0,-1,0),
	vec3f(1,0,-1),
	vec3f(0,1,-1),
};


/*
recieves:
	delta_x, delta_y,
	this_xmin, this_ymin, this_xmax, this_ymax,
	xmin, ymin, xmax, ymax,
	steptime
returns:
	steptime
	colNormal
*/

bool Box_TestCollision(
	//movement
	const vec2f &delta,
	//this bounds
	const box2f &b,
	//other bounds
	const box2f &ob,
	//other bounds side flags
	char sides,
	//four plane equations, bitflags corresponding with 'sides'
	const plane2f *planes,
	int planeCount,
	//return
	vec2f *colNormal,
	float *steptime)
{
	bool collided = false;
	float del, mid_min, mid_max;
	
	for (int i = 0; i < 2; i++) {
		int j = !i;
		
		if (delta(i)) {
			if (delta(i) > 0)	//test + side of player to - side of map
				del = ob.min[i] - b.max[i];
			else				//test - side of player to + side of map
				del = ob.max[i] - b.min[i];

			//divide by player_vx so we get del in range of [0,steptime]
			//whereas dividing by movement_x would be in range [0,1]
	//		float frac = del / delta.x;
	
			//tweak to prevent insertion into surfaces
			float frac;
			if (del < 0) {
				frac = (del + faceStepEpsilon) / delta(i);
			} else {
				frac = (del - faceStepEpsilon) / delta(i);
			}

	//		//clamp adjusted value to [0,1]
	//		if (frac < 0.f) frac = 0.f;
	//		if (frac > 1.f) frac = 1.f;

			//if we have valid movement
			if (frac >= -collisionEpsilon && frac <= *steptime + collisionEpsilon) {
				//test and see if, after moving 'frac' fraction,
				//if the opposing dimensions align...
				mid_min = b.min[j] + delta[j] * frac;
				mid_max = b.max[j] + delta[j] * frac;
				
				//test <min,max> to <ymin,ymax>
				if (mid_min < ob.max[j] && mid_max > ob.min[j]) {
					
					//COLLISION!l!@:lk@!#XXOR

					if (i == 0) {
						if (delta(i) < 0 && (sides & bitflag(SIDE_RIGHT))) {
							*steptime = frac;
							colNormal->set(1,0);
							collided = true;
						}
						if (delta(i) > 0 && (sides & bitflag(SIDE_LEFT))) {
							*steptime = frac;
							colNormal->set(-1,0);
							collided = true;
						}
					} else {
						if (delta(i) < 0 && (sides & bitflag(SIDE_DOWN))) {
							*steptime = frac;
							colNormal->set(0,1);
							collided = true;
						}
						if (delta(i) > 0 && (sides & bitflag(SIDE_UP))) {
							*steptime = frac;
							colNormal->set(0,-1);
							collided = true;
						}
					}
				}
			}
		}
	}

	return collided;
}

// *********************** Orientable Bounding Box physics test

#ifdef OOBTEST
bool OBB_TestIntersection(
	const vec2f &p1, float r1, const box2f &b1,
	const vec2f &p2, float r2, const box2f &b2);

#include "system/system.h"	//only for the mouse button state . . .

BEGIN_CHILDCLASS(OOBTest, Collidable)
	float lasttouch;	//used for highlighting

	OOBTest(SpawnObj *s) : super(s) {
		if (!friction) friction = 1.f;	//not equal default . . .
		useGravity = false;
		lasttouch = 0;
	}
	
	virtual void touch(Collidable *o) {
		super::touch(o);
		lasttouch = game_thistime;
	}

	vec2f grabpos;
	vec2f mousepos, mouselastpos;
	
	bool gotdown;
	
	virtual void think() {
		super::think();
		if (!active) return;
		
		//only looks good if the bbox is centered . . .
		//i guess i need to fix that =) by adding an offset to the sprite? in relative space that is . . .
		scale = bbox.size();
		
		color.l = 0xFFFFFFFF;
		if (game_thistime - lasttouch < 0.5f) {
			color.l = 0xFF0000FF;
		}
		
		mouselastpos = mousepos;
		extern box2f camera_bbox;
		mousepos.x = camera_bbox.min.x * (1.f - mouse.posf.x) + camera_bbox.max.x * mouse.posf.x;
		mousepos.y = camera_bbox.max.y * (1.f - mouse.posf.y) + camera_bbox.min.y * mouse.posf.y;

		//forward transform
		vec2f worldgrabpos = vec2rot(grabpos, angle) + pos;

		vel.set(0.f);
		rotation = 0;
		vec2f del = mousepos - mouselastpos;

		if (mouse.buttonState[MouseHandler::BUTTON_LEFT] == BUTTON_STATE_PRESS ||
			mouse.buttonState[MouseHandler::BUTTON_RIGHT] == BUTTON_STATE_PRESS)
		{
			if (OBB_TestIntersection(
				pos, angle, bbox,
				mousepos, 0, box2f(0.f, 0.1f)))
			{
				gotdown = true;
				worldgrabpos = mousepos;
				grabpos = vec2rot(worldgrabpos - pos, -angle);			//inverse transform
			}
		}
		if (mouse.buttonState[MouseHandler::BUTTON_LEFT] == BUTTON_STATE_DOWN) {
			if (gotdown) {
				//pos += del;	//fully working direct movement:
				vel = (mousepos - worldgrabpos) / game_deltatime;
	//			}
	//			if (main_mouse->buttonState[MouseHandler::BUTTON_RIGHT] == MH_DOWN) {
				vec2f v1 = mousepos - pos;
				vec2f v2 = worldgrabpos - pos;
				float z = vec2det(v1,v2);
				rotation = -z * RADIAN_TO_DEGREE / (vec2len(v1) * vec2len(v2) * game_deltatime);
			}
		}
		if (mouse.buttonState[MouseHandler::BUTTON_LEFT] == BUTTON_STATE_UP &&
			mouse.buttonState[MouseHandler::BUTTON_RIGHT] == BUTTON_STATE_UP)
		{
			gotdown = false;
		}
	}
ENDCLASS()
//REGISTER_ENTITY(OOBTest)
#endif

// *********************** End Orientable Bounding Box physics test

//TODO - get this to work with the script somehow ...
//and traceline ...
enum {
	MAPTYPE_EMPTY,
	MAPTYPE_SOLID,
	MAPTYPE_LIQUID,
	MAPTYPE_COUNT
};

//traceline ignore filters
#define TR_IGNORE_ENTS		1
#define TR_IGNORE_WORLD		2

//utility function - only called externally 
static Collidable *TraceLine(
	const vec2f &start,
	const vec2f &del,
	float *dist,
	int layer,
	int flags,
	bool (*filter)(Entity *e))
{
	vec2f fulldel = del * *dist;
	//calc movement bounds
	box2i b( start );
	b.stretch( start + fulldel );
	
	World *world = worldSys.getWorld();
	if (world) {
		//used often enough - make it external?  define it upon world load?
		box2i worldbox(vec2i(), vec2i((const int*)world->getMapSize()) - 1);	//inclusive
	
		if (!b.touchesE(worldbox)) {	//clip
			flags |= TR_IGNORE_WORLD;
		} else {	//clamp
			b.shrink(worldbox);
		}
	} else {
		flags |= TR_IGNORE_WORLD;
	}

	int x,y;
	bool collision_occured = false;
	vec2f colNormal;
	
	box2f startbox(start, start + 0.1f);

	//cycle thru movement range from floor(min) to ceil(max) and test for collision
	if ((flags & TR_IGNORE_WORLD) == 0) {
		for (y = b.min.y; y <= b.max.y; y++) {
			for (x = b.min.x; x <= b.max.x; x++) {
				//if the movement of the player rect collides with the box <x,y>x<x+1,y+1> then stop the movement where it should be stopped
				const MapKey *mk = world->getMap()->riskyGet(x,y,layer);
				if (mk->type != MAPTYPE_SOLID) continue;
				
				//so how are we going to do this . . .
				//give the planes to the testcollision ...
				
				//retranslate the planes
				plane2f planes[4];
				int pi = 0;
				for (int i = 0; i < 4; i++) {
					if (mk->sides & bitflag(i)) {
						planes[pi] = planeTranslate<2,float>(mk->planes[i], vec2f(x,y));
						pi++;
					}
				}

				//test collision between <x,y>x<x+1,y+1> and the entity 'e'
				collision_occured |= Box_TestCollision(
					del,
					startbox,
					box2f(vec2f((float)x, (float)y), vec2f((float)x + 1.f, (float)y + 1.f)),
					mk->sides,
					planes,
					pi,
					&colNormal,
					dist);
			}
		}
	}

	Collidable *last_touch_ent = NULL;

	if ((flags & TR_IGNORE_ENTS) == 0) {
//conPrint("checking entities ... \n");
		//now cycle thru all entities in the area and test them to collision
		//BSP TREE ME!!!
		for (x = 0; x < entAllocSys.getCount(); x++) {
			Collidable *e = dynamic_cast<Collidable *>(entAllocSys.entForIndex(x));
			if (!e ||
				!e->active ||
				e->noPhysics ||
				e->collIgnoreEnts ||
				(filter && !filter(e)))
			{
//conPrint(" entity %d failed\n", x);
				continue;
			}

			bool hit = false;
			
			box2f wb = e->bbox + e->getPos();

#ifdef OOBTEST
			if (dynamic_cast<OOBTest *>(e)) {
				//inverse translate the space so that the box is at the origin
				vec2f s = start - e->pos;
				//inverse-rotate the line about the box
				s = vec2rot(s, -e->angle);
				vec2f d = vec2rot(del, -e->angle);
				//only then do a test collision . . . (assume the ent is at the origin)

				hit = Box_TestCollision(
					d,
					box2f(s, s + 0.1f),
					e->bbox,
					-1,
					default_planes, 4,
					&colNormal,
					dist);
			} else
#endif
			{
				plane2f planes[4] = {
					plane2f(vec2f(-1,0),	-wb.min.x),
					plane2f(vec2f(0,-1),	-wb.min.y),
					plane2f(vec2f(1,0),	-wb.max.x),
					plane2f(vec2f(0,1),	-wb.max.y),
				};
			
				hit = Box_TestCollision(
					del,
					box2f(start, start + 0.1f),
					wb,
					-1,
					planes, 4,
					&colNormal,
					dist);
			}

			if (hit) {
				collision_occured = true;
				last_touch_ent = e;
			}
		}
	}
	
	//oh wait, collision normals aren't returned upon traceline

	return last_touch_ent;
}

static Entity *tracelineIgnoreEnt = NULL;
bool tracelineIgnoreFilter(Entity *e) {
	return e != tracelineIgnoreEnt;
}

//entHit, dist = traceline(vec2f start, vec2f delta, [ layer = -1, ignoreEnts = true, ignoreWorld = true, filter = nil])
int PF_traceline(lua_State *L) {
	int argc = lua_gettop(L);
	if (argc < 4) {
		con.print("usage: PF_traceline start delta [layer=0 ignoreEnts=false ignoreWorld=false ignoreEnt=nil]\n");
		return 0;
	}
	
	vec2f start = luaReadVec<2,float>(L,1);
	vec2f delta = luaReadVec<2,float>(L,2);

	float dist = 1;
	int layer = 0;
	int flags = 0;
	tracelineIgnoreEnt = NULL;
	if (argc >= 3) layer = (int)lua_tonumber(L, 3);
	if (argc >= 4 && lua_toboolean(L, 4)) flags |= TR_IGNORE_ENTS;
	if (argc >= 5 && lua_toboolean(L, 5)) flags |= TR_IGNORE_WORLD;
	if (argc >= 6) {
		lua_getfield(L, 6, "_ptr");
		entptr_t ep = {(int)lua_tonumber(L, -1)};
		tracelineIgnoreEnt = entAllocSys.entForPtr(ep);
	}
//	con.print("dist in: %f\n", dist);
	Collidable *col = TraceLine(start, delta, &dist, layer, flags, tracelineIgnoreEnt ? tracelineIgnoreFilter : NULL);	//no filter yet
//	con.print("dist out: %f\n", dist);

	lua_getglobal(L, "_ents");				//[_ents]
	lua_pushinteger(L, entAllocSys.ptrForEnt(col).i);	//[_ents] [colIndex]
	lua_rawget(L, -2);							//[_ents] [_ents[colIndex]]
	lua_remove(L, lua_gettop(L)-1);				//[_ents[colIndex]]
	lua_pushnumber(L, dist);
	return 2;
}

//same as below, but with movement
//how so you ask?  we transform the delta vector and do traceline tests between vertices and bounding boxes . . .
bool OBB_TestCollision(
	const vec2f &world_delta,
	const vec2f &p1, float r1, const box2f &b1,
	const vec2f &p2, float r2, const box2f &b2,
	//other bounds side flags
	char sides,		//sides on b2 to test collision with . . .
	//return
	vec2f *colNormal,
	float *steptime)
{
	bool collided = false;
	
	//move cis2 into cis1's orientation . . .
	const vec2f *p[2] = {&p1, &p2};
	float r[2] = {r1, r2};
	const box2f *b[2] = {&b1, &b2};

	//when i is zero, then we're testing the 2nd object's vertices for collision with the 1st object
	//and we use a transformed negative delta
	//otherwise, when i is one, we're testing the 1st object's vertices for col with the 2nd object
	//and we're using a trasnformed positive delta
	for (int c = 0; c < 2; c++) {
		int d = !c;
		
		vec2f delta = world_delta;
		if (!c) delta = -delta;
		
		//for object i, transform object j into object i's frame

		//first inverse translate
		vec2f pd = *p[d] - *p[c];
		//next inverse rotate
		pd = vec2rot(pd, -r[c]);	//you can use rotv with the cis basis if you'd like
		float rd = r[d] - r[c];
		delta = vec2rot(delta, -r[c]);
		//now you could inverse scale it, but that would lose our qualification as a rectangle . . .
		//so instead do a bounds check on this against each of object j's rectangle's corners . . .
		//actually right now scale doesnt affect collision bounds . . .
		
#if 1	//stingy way: test all vertices
		for (int m = 0; m < 2; m++) {
			for (int n = 0; n < 2; n++) {
#else	//or - just test the vtx closest aligned with the delta vector
		//get the delta vector from object j to object i, except in 'j's space this time
		//i dont think this is picking the right corner maybe . . .
		vec2f delrel = vec2rot(world_delta, -r[d]);
		if (!c) delrel = -delrel;
		
		{
			int m = delrel.x > 0;
			{
				int n = delrel.y > 0;
#endif				
				
				//get the m/n'th corner of box b[j] (local space)
				vec2f v((*b[d])[m].x, (*b[d])[n].y);
				//relative-rotate it into obj i's space
				v = vec2rot(v, rd);
				//relative translate it into i's space
				v += pd;
				//now do a plane test.  v is in object i's space, so just do identity-based tests
				//or do a trace along 'd', and see if 'v' intersects box b[i]
//				if (b[i]->contains(v)) return true;
				float del, mid;
				
				float laststep = *steptime;
				
				for (int i = 0; i < 2; i++) {
					int j = !i;
					
					if (delta(i)) {
						if (delta(i) > 0)	//test + side of player to - side of map
							del = (*b[c]).min[i] - v[i];
						else				//test - side of player to + side of map
							del = (*b[c]).max[i] - v[i];
			
						//divide by player_vx so we get del in range of [0,steptime]
						//whereas dividing by movement_x would be in range [0,1]
				//		float frac = del / delta.x;
				
						//tweak to prevent insertion into surfaces
						float frac;
						if (del < 0) {
							frac = (del + faceStepEpsilon) / delta(i);
						} else {
							frac = (del - faceStepEpsilon) / delta(i);
						}
			
				//		//clamp adjusted value to [0,1]
				//		if (frac < 0.f) frac = 0.f;
				//		if (frac > 1.f) frac = 1.f;
			
						//if we have valid movement
						if (frac >= -collisionEpsilon && frac <= *steptime + collisionEpsilon) {
							//test and see if, after moving 'frac' fraction,
							//if the opposing dimensions align...
							mid = v(j) + delta(j) * frac;
							
							//test <min,max> to <ymin,ymax>
							if (mid < (*b[c]).max[j] && mid > (*b[c]).min[j]) {
								
								//COLLISION!l!@:lk@!#XXOR
								//if c == 1then we're testing the 1st obj's vertices against the 2nd object
								//so we use the 2nd object's faces
								if (i == 0) {
									if (delta(i) < 0 && (!c || sides & bitflag(SIDE_RIGHT))) {
										*steptime = frac;
										*colNormal = vec2rot(vec2f(1,0), r[c]);
										collided = true;
									}
									
									if (delta(i) > 0 && (!c || sides & bitflag(SIDE_LEFT))) {
										*steptime = frac;
										*colNormal = vec2rot(vec2f(-1,0), r[c]);
										collided = true;
									}
								} else {
									if (delta(i) < 0 && (!c || sides & bitflag(SIDE_DOWN))) {
										*steptime = frac;
										*colNormal = vec2rot(vec2f(0,1), r[c]);
										collided = true;
									}
									
									if (delta(i) > 0 && (!c || sides & bitflag(SIDE_UP))) {
										*steptime = frac;
										*colNormal = vec2rot(vec2f(0,-1), r[c]);
										collided = true;
									}
								}
							}
						}
					}
				}
				
				//now we know if we collided
				if (debugDisplayCollision && laststep != *steptime) {
					vec2f v((*b[d])[m].x, (*b[d])[n].y);
					v = vec2rot(v, r[d]);
					v += *p[d];
					glColor3f(0,1,0);
					glBegin(GL_LINE_LOOP);
					glVertex2fv(v + vec2f(-.1f,-.1f));
					glVertex2fv(v + vec2f(.1f,-.1f));
					glVertex2fv(v + vec2f(-.1f,.1f));
					glVertex2fv(v + vec2f(.1f,.1f));
					glEnd();
				}
			}
		}
	}
	
	return collided;
}

//I'm glad You see me as a person
//when being a person
//is such an accomplishment for me

//if a vertex of one box is inside the other then we have a collision
//so take turns computing vertices into the other box's relative space
bool OBB_TestIntersection(
	const vec2f &p1, float r1, const box2f &b1,
	const vec2f &p2, float r2, const box2f &b2)
{
	//create some planes . . .
//	vec2f cis1 = vec2expd(r1);
//	vec2f cis2 = vec2expd(r2);

	//determine one relative to the other's space
	//with the other (the non-relative), calc its planes, then determine
	//if all the vertices of the one (the relative) are in or out of those planes
	//this will make vertex tests easy - just a double double loop
	
	//move cis2 into cis1's orientation . . .
	const vec2f *p[2] = {&p1, &p2};
	float r[2] = {r1, r2};
	const box2f *b[2] = {&b1, &b2};
//	const vec2f *cis[2] = {&cis1, &cis2};
	
	for (int i = 0; i < 2; i++) {
		int j = !i;
		
		//for object i, transform object j into object i's frame

		//first inverse translate
		vec2f pj = *p[j] - *p[i];
		//next inverse rotate
		pj = vec2rot(pj, -r[i]);	//you can use rotv with the cis basis if you'd like
		float rj = r[j] - r[i];
		//now you could inverse scale it, but that would lose our qualification as a rectangle . . .
		//so instead do a bounds check on this against each of object j's rectangle's corners . . .
		//actually right now scale doesnt affect collision bounds . . .
		for (int m = 0; m < 2; m++) {
			for (int n = 0; n < 2; n++) {
				//get the m/n'th corner of box b[j] (local space)
				vec2f v((*b[j])(m).x, (*b[j])(n).y);
				//relative-rotate it into obj i's space
				v = vec2rot(v, rj);
				//relative translate it into i's space
				v += pj;
				//now do a plane test.  v is in object i's space, so just do identity-based tests
				if (b[i]->contains(v)) return true;
			}
		}
	}
	
	return false;
}

static void RunTouchScript(lua_State *L, Entity *self, Entity *other) {
	int top = lua_gettop(L);
	lua_getglobal(L, "_ents");					// [_ents]
	assert (lua_istable(L,-1));
	
	entptr_t ep = entAllocSys.ptrForEnt(self);
	lua_pushinteger(L, ep.i );	// [_ents] [self._ptr]
	lua_rawget(L, -2);			// [_ents] [self]
	if (!lua_istable(L,-1)) {
//this condition will be gone by the time everything is ported into script
	} else {
		lua_getfield(L, -1, "touch");					// [_ents] [self] [self:touch]
		if (lua_isfunction(L, -1)) {
			lua_insert(L, lua_gettop(L)-1);				// [_ents] [self:touch] [self]
			
			lua_pushinteger(L, entAllocSys.ptrForEnt(other).i);	// [_ents] [self:touch] [self] [other._ptr]
			lua_rawget(L, -4);							// [_ents] [self:touch] [self] [other]
			
			if (lua_pcall(L, 2, 0, 0)) {				// [_ents]
				con << "lua_pcall got error rendering:\n " << lua_tostring(L, -1) << "\n " << lua_tostring(L, -2) << "\n";
			}
		} else {
			//non-collidable
		}
	}
	lua_settop(L,top);
}

//int collision_frame_count = 0;

static void Update_Entity_Physics(lua_State *L, Collidable *e) {

	if (debugDisplayCollision) {
		glPushAttrib(GL_ENABLE_BIT);
		glDisable(GL_BLEND);
		glDisable(GL_TEXTURE_2D);
	}

	World *world = worldSys.getWorld();

//reset body forces
//	e->force_x = 0;
	if (e->useGravity && e->massInv) {
//		e->force_y = world_gravity / e->massInv;
		vec2f grav = MapTagType::def.grav;
		
		if (world) {
			short tagKeyID = world->tags->maploc( floor_ftoi(e->getPos().x), floor_ftoi(e->getPos().y));
			map<short,MapTagKey*>::iterator it = world->tagkeys->keys.find(tagKeyID);
			if (it != world->tagkeys->keys.end()) {
				MapTagType *tt = it->second->tagType;
				if (tt) grav = tt->grav;
			}
		}
		e->vel += grav * (game_deltatime / e->massInv);
	}

//world speed limit
	for (int i = 0; i < 2; i++) {
		if (e->vel[i] > maxVelocity)	e->vel[i] = maxVelocity;
		if (e->vel[i] < -maxVelocity)	e->vel.x = -maxVelocity;
	}
	
	//this is somewhat physics, but unrelated to all else, so it goes here:
	e->angle += e->rotation * game_deltatime;

	//save this for post-traceline movement for the friction code
//	bool e_contact_down_prev = e->contact_down;
	//nevermind - this causes jumps to experience friction just before lifting off the ground

	//hack - infinite mass implies no velocity or force
	if (!e->massInv) e->vel = 0.f;
	
	
	if (world) {
		//if we are within water prior to position integration
		//then dampen our velocity (FIXME)
		const MapKey *mk = world->getMap()->safeGet(floor_ftoi(e->getPos().x), floor_ftoi(e->getPos().y), e->layer);
		if (mk->type == MAPTYPE_LIQUID) {
			float viscosity = mk->friction[0];

	//make drag timestep-independant
	//TODO: just run physics on a fixed framerate.
			if (viscosity != 0.f && viscosity != 1.f)
				viscosity = (float)pow(viscosity, game_deltatime * 100.f);

			e->vel *= viscosity;
			e->rotation *= viscosity;
		}
	}

	//zero contact flags
	//e->contact_side = 0;
	
		//collision / position integration
	float time = 0;
	float target_time = game_deltatime;

	box2i worldbox;
	if (world) {
		worldbox = box2i(vec2i(),vec2i((const int*)world->getMapSize())-1);	//inclusive
	}
	
	e->contactNormal = 0.f;
	
	//first do a tracemove up a step
	//then do our tracemove along the velocity vector
	//then do a tracemove down a step . . .
	
	//prevent lockup - esp with restitution > 100% - and cap the iteration count
	int iterations;
	for (iterations = collisionMaxTests; iterations--;) {
		if (time == game_deltatime) break;

		int x,y;

		//the timestep of this instance of the 'while'
		float steptime = target_time - time;
		
		vec2f newpos = e->getPos() + e->vel * steptime;

		//world box of entity 'e'
		box2f wb = e->bbox + e->getPos();
		
		//calc movement bounds
		box2i mb;
		//int xmin, xmax, ymin, ymax;
		if (e->vel.x > 0) {
			mb.min.x = (int)(wb.min.x);
			mb.max.x = (int)(e->bbox.max.x + newpos.x);
		} else {
			mb.min.x = (int)(e->bbox.min.x + newpos.x);
			mb.max.x = (int)(wb.max.x);
		}

		if (e->vel.y > 0) {
			mb.min.y = (int)(wb.min.y);
			mb.max.y = (int)(e->bbox.max.y + newpos.y);
		} else {
			mb.min.y = (int)(e->bbox.min.y + newpos.y);
			mb.max.y = (int)(wb.max.y);
		}

		//if our range is OOB then flip this flag and dont test world
		bool collision_test_map = !e->collIgnoreWorld;
		
		if (!world) {
			collision_test_map = false;
		} else {
			if (!mb.touchesE(worldbox)) {	//clip
				collision_test_map = false;
			} else {	//clamp
				mb.shrink(worldbox);
			}
		}

		vec2f colNormal;
		
		//show tested range
		if (debugDisplayCollision) {
			glColor3f(1,0,1);
			glBegin(GL_LINE_LOOP);
			glVertex2f(mb.min.x + 0.1f, mb.min.y + 0.1f);
			glVertex2f(mb.min.x + 0.1f, mb.max.y + 0.9f);
			glVertex2f(mb.max.x + 0.9f, mb.max.y + 0.9f);
			glVertex2f(mb.max.x + 0.9f, mb.min.y + 0.1f);
			glEnd();

			glColor3f(1,1,0);
			glBegin(GL_LINE_LOOP);
			glVertex2f(wb.min.x, wb.min.y);
			glVertex2f(wb.min.x, wb.max.y);
			glVertex2f(wb.max.x, wb.max.y);
			glVertex2f(wb.max.x, wb.min.y);
			glEnd();
		}

		bool collision_occured = false;
		vec2i map_col;

		box2f colbox, f;
		if (collision_test_map) {
			//cycle thru movement range from floor(min) to ceil(max) and test for collision
			for (y = mb.min.y; y <= mb.max.y; y++) {
				for (x = mb.min.x; x <= mb.max.x; x++) {
					//if the movement of the player rect collides with the box <x,y>x<x+1,y+1> then stop the movement where it should be stopped
					const MapKey *mk = world->getMap()->riskyGet(x,y,e->layer);
					if (mk->type != MAPTYPE_SOLID) continue;

					if (debugDisplayCollision) {
						glColor3f(1,0,0);
						glBegin(GL_LINES);
						glVertex2i(x,y);glVertex2i(x+1,y+1);
						glVertex2i(x+1,y);glVertex2i(x,y+1);
						glEnd();
					}

					f.set( vec2f((float)x, (float)y), vec2f((float)x + 1.f, (float)y + 1.f));

					plane2f planes[4];
					int pi = 0;
					for (int i = 0; i < 4; i++) {
						if (mk->sides & bitflag(i)) {
							planes[pi] = planeTranslate(mk->planes[i], vec2f(x,y));
							pi++;
						}
					}

					//if we are using the std box . . .
					if (Box_TestCollision(
						e->vel,
						wb,
						f,
						mk->sides,
						mk->planes, pi,
						&colNormal, &steptime))
					{
/*
con << "ent-world touchonly collision" << endl;
con << "this box " << wb << endl;
con << "other box " << f << endl;
con << "delta " << e->vel << endl;
con << "steptime " << steptime << endl;
*/
						collision_occured = true;
						colbox = f;
						map_col.set(x,y);		
					}
				}
			}
		}

		Collidable *last_touch_ent = NULL;

		if (!e->collIgnoreEnts) {

			//now cycle thru all entities in the area and test them to collision
			//BSP TREE ME!!!
			for (x = 0; x < entAllocSys.getCount(); x++) {
				Collidable *ex = dynamic_cast<Collidable *>(entAllocSys.entForIndex(x));
				if (!ex ||
					!ex->active ||
					ex->noPhysics ||
					ex->collIgnoreEnts ||
					ex->collTouchOnly ||
					ex == e)
				{
					continue;
				}

				//must...fix...this...
				if (ex == e->collIgnore) continue;
				if (ex->collIgnore == e) continue;

				if ((ex->collDiffClass || e->collDiffClass) && (ex->entClassID == e->entClassID)) continue;
				
				//world box of entity 'ex'
				f = ex->bbox + ex->getPos();
				
				if (e->collTouchOnly) {
				
					bool hit = false;
					//if this or the other is an instance of an orientable bounding box
					//do an obb intersection test (search for separating plane)
#ifdef OOBTEST
					OOBTest *obb1 = dynamic_cast<OOBTest *>(e);
					OOBTest *obb2 = dynamic_cast<OOBTest *>(ex);
					if (obb1 || obb2) {
						float r1 = obb1 ? obb1->angle : 0;
						float r2 = obb2 ? obb2->angle : 0;
						
						hit = OBB_TestIntersection(
							e->pos, r1, e->bbox,
							ex->pos, r2, ex->bbox);

					} else
#endif
					{
						hit = wb.touches(f);	//do a intersection test here
					}
					
					if (hit) {
						collision_occured = true;
						last_touch_ent = ex;
						colbox = f;
					}
				} else {
				
					bool hit = false;
#ifdef OOBTEST
					OOBTest *obb1 = dynamic_cast<OOBTest *>(e);
					OOBTest *obb2 = dynamic_cast<OOBTest *>(ex);
					if (obb1 || obb2) {
						float r1 = obb1 ? obb1->angle : 0;
						float r2 = obb2 ? obb2->angle : 0;

						hit = OBB_TestCollision(
							e->vel,
							e->pos, r1, e->bbox,
							ex->pos, r2, ex->bbox,
							-1, &colNormal, &steptime);
							
						//hack for the time being - disc by 4
//						if (hit) colNormal = vec2expd(90 * (int)((atan2(colNormal.y, colNormal.x) * RADIAN_TO_DEGREE + 45) / 90));
					} else
#endif
					{
						plane2f planes[4] = {
							plane2f(vec2f(-1,0),	-f.min.x),
							plane2f(vec2f(0,-1),	-f.min.y),
							plane2f(vec2f(1,0),		-f.max.x),
							plane2f(vec2f(0,1),		-f.max.y),
						};
					
						hit = Box_TestCollision(
							e->vel,
							wb,
							f,
							-1,
							planes, 4,
							&colNormal, &steptime);
					}
					if (hit) {
#if 0
con << "ent-ent touchonly collision" << endl;
con << "this box " << wb << endl;
con << "other box " << f << endl;
con << "delta " << e->vel << endl;
con << "steptime " << steptime << endl;
#endif
						collision_occured = true;
						last_touch_ent = ex;
						colbox = f;
					}
				}
			}
		}

		//time = start of timeframe
		//steptime + time = intermediate position in timeframe
		//target_time = end of timeframe
		//game_deltatime = end of uber timeframe

		//integrate over partial timestep
		e->setPos(e->getPos() + e->vel * steptime);
		float restitution;
		if (last_touch_ent) {
			restitution = last_touch_ent->restitution;
		} else {
			restitution = world->getMap()->safeGet(VEC2ELEM(map_col), e->layer)->restitution;
		}
		restitution *= e->restitution;

		//touchonly seems like a bad idea
		//if we elim the contact side then we end up not zero-ing the velocity . . .
		//how about making touchonly apply only to entities?
		//like collision_with_world_but_only_execute_touch_function_of_entities ?
		if (e->collTouchOnly || (last_touch_ent && last_touch_ent->collTouchOnly)) {
			colNormal = 0.f;
		}
//		if (e->collTouchOnly) {
//			e->contact_side = 0;		//elim the contact side bitflags =)
//		}

		if (debugDisplayCollision) {
			glColor3f(1,0,1);
			glBegin(GL_LINES);
			glVertex2fv(e->getPos());
			glVertex2fv(e->getPos() + colNormal);
			glEnd();
			glBegin(GL_LINE_LOOP);
			glVertex2fv(e->getPos() + vec2f(-.1f,-.1f));
			glVertex2fv(e->getPos() + vec2f(.1f,-.1f));
			glVertex2fv(e->getPos() + vec2f(-.1f,.1f));
			glVertex2fv(e->getPos() + vec2f(.1f,-.1f));
			glEnd();
		}

		//only push entities if they have a mass inverse (ie if they are integratable)
		if ((colNormal.x != 0 || colNormal.y) != 0 && e->massInv) {
//debug
//collision_frame_count++;

			//set the flag
			e->contactNormal = colNormal;

			//update the info
			e->contactTime = game_thistime;
			e->contactEnt = last_touch_ent;
			e->contactTile = map_col;
			
			newpos = e->getPos();
			
			//assumes normal is normalized
			float v_n = dot(e->vel, colNormal);
			if (v_n < 0)
			{
				//e->vel += colNormal * (v_n * (1.f + restitution));
				v_n *= -restitution;
			} else { //otherwise no collision to begin with?
				//or just zero the velocity . . .
				//e->vel += colNormal * v_n;
				v_n = 0;
			}
			
			vec2f tangent(colNormal.y, -colNormal.x);
			float t_n = dot(e->vel, tangent);

			if (v_n > -minVelocity && v_n < minVelocity) {
				v_n = 0;
			}
			
			//reconstruct the basis . . .
			e->vel = tangent * t_n + colNormal * v_n;
			e->setPos(e->getPos() + colNormal * planePushBack);

		}


		//do a touch function - post physics response
		//add in a pre-velocity-application touch function? separate 'touch' and 'bounce' ?
		if (collision_occured) {
			RunTouchScript(L, e, last_touch_ent);
			//e->touch(last_touch_ent);
		}

		//make sure we're still alive
		//when do we want to skip contact?
		if (!e->active) break;
		if (e->noPhysics || e->collIgnoreEnts) {
			e->contactNormal = 0.f;
			break;
		}

		//test the other guy for collision
		if (last_touch_ent &&
			last_touch_ent->active &&
			!last_touch_ent->noPhysics &&
			!last_touch_ent->collIgnoreEnts)
		{
			RunTouchScript(L, last_touch_ent, e);
			//last_touch_ent->touch(e);
		}

		//make sure we're still alive
		if (!e->active) break;
		
		if (last_touch_ent) {
			if (!last_touch_ent->active || last_touch_ent->noPhysics || last_touch_ent->collIgnoreEnts) {
				e->contactNormal = 0.f;
			}
		}

		//	compute next the timeframe from 'time' to the end of our 'game_deltatime'
		time += steptime;
		target_time = game_deltatime;
	}

	if (!iterations) con << "exceeded physics iteration bound for ent " << e << "\n";

	if (e->contactNormal.x || e->contactNormal.y) {

		//offset velocity if we're resting on an entity
		//should this be tangent normal based?
		if (e->contactEnt && e->contactNormal.y < -0.5) {
			//instead of directly setting the velocity
			//how about adding the change in touch ent velocity to this ent?
			e->vel.x -= e->contactEnt->vel.x * e->massInv;
		}

		vec2f tangent(e->contactNormal.y, -e->contactNormal.x);
		
		//get velocity travelling along tangent
		float v_t = dot(e->vel, tangent);
		
		//subtract it from our current velocity (remove the tangent component)
		e->vel -= tangent * v_t;

		if (v_t) {
			float friction = 0;
			if (e->contactEnt) {
				friction = e->contactEnt->friction;
			} else {
				vec2f n(e->contactNormal);
				vec2f an(
					n.x < 0 ? -n.x : n.x,
					n.y < 0 ? -n.y : n.y);
				int side = SIDE_UP;	//by default
				if (an.x > an.y) {
					if (n.x < 0) {
						side = SIDE_LEFT;
					} else {
						side = SIDE_RIGHT;
					}
				} else {
					if (n.y < 0) {
						side = SIDE_UP;
					} else {
						side = SIDE_DOWN;
					}
				}
				friction = world->getMap()->safeGet(VEC2ELEM(e->contactTile), e->layer)->friction[side];
			}
					
			float force = friction * e->friction * e->massInv * baseFrictionScalar;
			int dir = v_t > 0;
			if (dir) force = -force;
				
			v_t += force * game_deltatime;
			//dont allow friction to go negative
			if ((v_t > 0) != dir) v_t = 0;
		}

		//re-add the tangent component
		if (v_t) {
			e->vel += tangent * v_t;
		}
		
		//re-offset velocity if we're resting on an entity
		if (e->contactEnt && e->contactNormal.y < -0.5) {
			e->vel.x += e->contactEnt->vel.x * e->massInv;
		}
	}

////integrate forces
//	e->vel.x += e->force_x * e->massInv * game_deltatime;
//	e->vel.y += e->force_y * e->massInv * game_deltatime;

	//prevent fall through the world
	if (e->getPos().y > world->getMapSize().y + 100) {
		//delete e;
		e->setPos(vec2f(e->getPos().x, (float)world->getMapSize().y + 100.f));
		Collidable *c = dynamic_cast<Collidable *>(e);
		if (c) c->vel.y = 0;
	}

		//something is causing bombs to do this just before destruction
	if (isnan(e->getPos().x) || isnan(e->getPos().y)) {
		e->setPos(vec2f(0.f));
		Collidable *c = dynamic_cast<Collidable *>(e);
		if (c) c->vel = 0.f;
	}
	
	if (debugDisplayCollision) {
		glPopAttrib();
	}
}

// PF_updateEntityPhysics(Collidable)
int PF_updateEntityPhysics(lua_State *L) {
	int argc = lua_gettop(L);
	if (argc < 1) {
		con << "usage: updateEntityPhysics(Collidable)\n";
		return 0;
	}
	
	lua_rawgetfield(L, 1, "_ptr"); 
	entptr_t ep = { (int)lua_tonumber(L, -1) };
	Entity *e = entAllocSys.entForPtr(ep);
	if (!e) {
		con << "PH_updateEntityPhysics: entity " << ep.i << " does not exist!\n";
		return 0;
	}
	if (!e->active) {
		con << "PH_updateEntityPhysics: entity " << ep.i << " is not active!\n";
		return 0;
	}
	Collidable *col = dynamic_cast<Collidable *>(e);
	if (!col) {
		con << "PH_updateEntityPhysics: entity " << ep.i << " is not of Collidable!\n";
		return 0;
	}
	Update_Entity_Physics(L, col);
	return 0;
}

