/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <string>
#include <sstream>
#include <sstream>
#include <lua.hpp>

#include "system/exception.h"

#include "menu/sys.h"

#include "inspiration/game/main.h"
#include "inspiration/game/program.h"
#include "inspiration/game/programmacros.h"
#include "inspiration/game/physics.h"	//PF_traceline
#include "inspiration/game/spawn.h"		//EntRegLink
#include "inspiration/game/nettables.h"	//PF_engine*MetaTable

#include "inspiration/nettables.h"	//netTablesSys
#include "inspiration/fileSys.h"
#include "inspiration/world.h"		//mapget...
#include "inspiration/animation.h"
#include "inspiration/entalloc.h"
#include "inspiration/entity.h"
#include "inspiration/spawnobj.h"
#include "inspiration/blend.h"
#include "inspiration/shader.h"		//PF_createShader
#include "inspiration/lighting.h"
#include "inspiration/texloader.h"
#include "inspiration/client.h"		//PF_entClientGetInput
#include "inspiration/draw.h"
#include "inspiration/program.h"

using namespace std;


static int PF_spawn(lua_State *L) {
	if (!lua_gettop(L)) {
		con.print("usage: PF_spawn(string spawnObjectName)");
		return 0;
	}
	const char *str = lua_tostring(L, 1);
	entptr_t ep = {-1};
	if (str) {
		for (int i = 0; i < spawnObjSys.count(); i++) {
			SpawnObj *s = spawnObjSys.get(i);
			assert(s);
			string name;
			if (s->rw("name") >> name && name.length()) {
				if (name == str) {
					//get the current existing entity
					ep = s->ent;
					//if its not there then spawn it
					if (!entAllocSys.entForPtr(ep)->active) {
						if (s->createEnt()) {
							ep = s->ent;
						}
					}
				}
			}
		}
	}
	lua_pushnumber(L, ep.i);
	return 1;
}

//entity newEnt([string classname, table params {key=value,...}])  ... eventually ...
static int PF_newEnt(lua_State *L) {
	const char *name = NULL;
	if (lua_gettop(L)) name = lua_tostring(L,1);
	
	entptr_t ep = {-1};
	Entity *e = NULL;
	try {
		//create an entity by name if provided
		e = name ? EntRegLink::createByName(name) : new Entity();
		//new shouldn't return null...
		ep = entAllocSys.ptrForEnt(e);
		
		//complain if it didnt work
		if (ep.i == -1) {
			con << "failed to find index for entity named " << name << endl;
		}
	} catch (Exception &e) {
		con << "failed to allocate entity of type " << name << endl;
	}

	//return it in lua
	lua_pushinteger(L, ep.i);
	return 1;
}

static int PF_entRemove(lua_State *L) {
	if (!lua_gettop(L)) {
		con.print("usage: bool success = PF_remove(entity who)\n");
		return 0;
	}
	entptr_t ep = {(int)lua_tonumber(L,1)};
	Entity *e = ep.i == -1 ? NULL : entAllocSys.entForPtr(ep);
	if (!e || !e->active) {
		lua_pushboolean(L, 0);
	} else {
		delete e;
		lua_pushboolean(L, 1);
	}
	return 1;
}

static int PF_entGetCount(lua_State *L) {
	lua_pushinteger(L, entAllocSys.getCount());
	return 1;
}

static int PF_entGetIndex(lua_State *L) {
	int argc = lua_gettop(L);
	if (argc != 1) {
		con << "usage: PF_entGetIndex(entity)\n";
		return 0;
	}
	int i = (int)lua_tonumber(L,1);
	lua_pushinteger(L, entAllocSys.ptrForEnt(entAllocSys.entForIndex(i)).i);
	return 1;
}


//bool active(entity)
static int PF_entActive(lua_State *L) {
	if (lua_gettop(L) != 1) {
		con.print("usage: bool entActive(entity)\n");
		return 0;
	}
	
	EXPOSE_ENTITY_PARAM_IN_LUA__GRAB_ENT_PTR(entActive, 1, ep);
	
	Entity *e = entAllocSys.entForPtr(ep);
	lua_pushboolean(L, (e && e->active));
	return 1;
}

bool Lua_CreateSpawnObj(lua_State *L, int soIndex);


EXPOSE_ENTITY_SHORT_PARAM_IN_LUA(entEntClassID, Entity, entClassID)

//[int so =] entSpawnObj(entity[, int so]);
EXPOSE_ENTITY_PARAM_IN_LUA_HEADER(entSpawnObj, Entity, 1)
	int so = spawnObjSys.getIndex(e->spawnobj);
	if (argc == 2) e->spawnobj = spawnObjSys.get((int)lua_tonumber(L, 2));
	if (so == -1) {
		lua_pushnil(L);
	} else {
		if (!Lua_CreateSpawnObj(L, so)) {
			con << "PF_entSpawnObj failed to create a new SpawnObj!\n"
				<< lua_tostring(L,-1) << "\n" << lua_tostring(L,-2);	//spawntable
		}
	}
	return 1;
}

EXPOSE_ENTITY_INT_PARAM_IN_LUA(entLayer, Entity, layer)

//this is going to be messy:
//table exposure, x, y, and xy ... all with getters and setters ...
#if 0
EXPOSE_ENTITY_VEC2F_PARAM_IN_LUA( entPos, Entity, pos )
#else

//this could be avoided by writing EXPOSE_ENTITY_VEC2F (and like parameters)
//as macros to use getters and setters
//write those getters/setters so they can be overloaded with a direct write (tempalte metaprogram?)
//then plug in getPos()/setPos() in place of our usual variable access ...

EXPOSE_ENTITY_PARAM_IN_LUA_HEADER( entPos, Entity, 2 )
	vec2f v = e->getPos();
	if (argc == 3) e->setPos(vec2f(lua_tonumber(L, 2), lua_tonumber(L, 3)));
	lua_pushnumber(L, v.x);
	lua_pushnumber(L, v.y);
	return 2;
}
EXPOSE_ENTITY_PARAM_IN_LUA_HEADER( entPosTable, Entity, 1 )
	vec2f v = e->getPos();
	if (argc == 2) {
		if (!lua_istable(L, 2)) {
			con << "PF_entPosTable expected param 2 to be a table\n";
			return 0;
		}
		e->setPos(luaReadVec<2,float>(L, 2));
	}
	entityLuaExposeGetOrMakeTable(L, "_obj_entPos", 1);
	int tabletop = lua_gettop(L);
	luaWriteVec<2,float>(L, tabletop, v);
	return 1;
}

EXPOSE_ENTITY_PARAM_IN_LUA_HEADER( entPosX, Entity, 1 )
	float old = e->getPos().x;
	if (argc > 1) e->setPos(vec2f(lua_tonumber(L, 2), e->getPos().y));
	lua_pushnumber(L, old);
	return 1;
}

EXPOSE_ENTITY_PARAM_IN_LUA_HEADER( entPosY, Entity, 1 )
	float old = e->getPos().y;
	if (argc > 1) e->setPos(vec2f(e->getPos().x, lua_tonumber(L, 2)));
	lua_pushnumber(L, old);
	return 1;
}

#endif

EXPOSE_ENTITY_FLOAT_PARAM_IN_LUA( entNextThink, Entity, nextthink)

//entServerSendMessage(ent, ...)
static int PF_entServerSendMessage(lua_State *L) {
	int argc = lua_gettop(L);
	EXPOSE_ENTITY_PARAM_IN_LUA__GRAB_ENT_PARAM(entServerSendMessage, Entity, 1, ep, e)
	vector<string> params;
	for (int i = 2; i <= argc; i++) {
		//of all the lua data, I'll end up only supporting null-term strings here for now
		//the idea for now is this isn't anything to do continuously
		if (lua_isstring(L, i)) {
			params.push_back(lua_tostring(L, i));
		} else {
			lua_getglobal(L, "tostring");
			lua_pushvalue(L, i);
			lua_pcall(L, 1, 1, 0);
			params.push_back(lua_tostring(L, -1));
			lua_pop(L,1);
		}
	}
	entitySys.sendMessage(ep, params);
	return 0;
}

//to be called clientside while netTablesSys is set on a client table somewhere
static int PF_entClientSendMessage(lua_State *L) {
	int argc = lua_gettop(L);
	vector<string> params;
	entptr_t ep;
	ep.i = lua_tointeger(L,1);
	for (int i = 2; i <= argc; i++) {
		if (lua_isstring(L, i)) {
			params.push_back(lua_tostring(L, i));
		} else {
			lua_getglobal(L, "tostring");
			lua_pushvalue(L, i);
			lua_pcall(L, 1, 1, 0);
			params.push_back(lua_tostring(L, -1));
			lua_pop(L,1);
		}
	}
	clientSys.sendMessage(ep, params);
	return 0;
}

/*
this does two things
er - clientSys.getClientInput does two things:
1) lookup the client index based on the passed entity
2) construct and return that client's input data

menus needs input data
but it doesn't have an ent conn
so we need to either split this in two
or provide an alternative for getMenuClientInput
	- that returns input based on the menu's key mappings
*/
static int PF_entClientGetInput(lua_State *L) {
	int argc = lua_gettop(L);
	if (!argc) luaL_error(L, "usage: cmd, mouse_active, mouse_x, mouse_y = PF_entClientGetInput(entityID)\n");
	entptr_t ep;
	ep.i = lua_tointeger(L,1);
	ClientInput clientInput;
	if (!clientSys.getClientInput(ep, &clientInput)) return 0;
	lua_pushinteger(L, clientInput.cmd);
	lua_pushboolean(L, clientInput.mouseActive);
	lua_pushnumber(L, clientInput.mousePos.x);
	lua_pushnumber(L, clientInput.mousePos.y);
	return 4;
}

static int PF_clientGetMenuInput(lua_State *L) {
	ClientInput clientInput;
	if (!clientSys.getMenuClientInput(&clientInput)) return 0;
	lua_pushinteger(L, clientInput.cmd);
	lua_pushboolean(L, clientInput.mouseActive);
	lua_pushnumber(L, clientInput.mousePos.x);
	lua_pushnumber(L, clientInput.mousePos.y);
	return 4;
}

EXPOSE_ENTITY_BOOL_PARAM_IN_LUA( animLeft, Renderable, left)
EXPOSE_ENTITY_FLOAT_PARAM_IN_LUA( animAngle, Renderable, angle)
EXPOSE_ENTITY_VEC2F_PARAM_IN_LUA( animScale, Renderable, scale)
EXPOSE_ENTITY_UCHAR_PARAM_IN_LUA( animBlend, Renderable, blend)
EXPOSE_ENTITY_UINT_PARAM_IN_LUA( animShader, Renderable, shader)
EXPOSE_ENTITY_COLOR_PARAM_IN_LUA( animColor, Renderable, color )

EXPOSE_ENTITY_INT_PARAM_IN_LUA( animAnimFrame, Renderable, anim.frame)
EXPOSE_ENTITY_FLOAT_PARAM_IN_LUA( animAnimFrameNextTime, Renderable, anim.frameNextTime)
EXPOSE_ENTITY_FLOAT_PARAM_IN_LUA( animAnimFrameLastTime, Renderable, anim.frameLastTime)

EXPOSE_ENTITY_PARAM_IN_LUA_HEADER( animAnimGroup, Renderable, 1)
	string oldgroupname;
	int oldgroupindex = e->anim.group;
	animSys.getGroupName(e->anim.group, oldgroupname);	//should i null the oldgroupname if it fails? just leave it to the func.
	if (argc == 2) e->anim.setGroup(lua_tostring(L, 2));
	lua_pushstring(L, oldgroupname.c_str());
	lua_pushinteger(L, oldgroupindex);
	return 2;
}

EXPOSE_ENTITY_PARAM_IN_LUA_HEADER( animAnimSeq, Renderable, 1)
	string oldseqname;
	int oldseqindex = e->anim.seq;
	animSys.getSeqName(e->anim.group, e->anim.seq, oldseqname);
	if (argc == 2) e->anim.setSeq(lua_tostring(L, 2));
	lua_pushstring(L, oldseqname.c_str());
	lua_pushinteger(L, oldseqindex);
	return 2;
}

EXPOSE_ENTITY_PARAM_IN_LUA_HEADER( animAnimIdleSeq, Renderable, 1)
	string oldidleseqname;
	int oldidleseqindex = e->anim.idleSeq;
	animSys.getSeqName(e->anim.group, e->anim.idleSeq, oldidleseqname);
	if (argc == 2) e->anim.setIdleSeq(lua_tostring(L, 2));
	lua_pushstring(L, oldidleseqname.c_str());
	lua_pushinteger(L, oldidleseqindex);
	return 2;
}

EXPOSE_ENTITY_PARAM_IN_LUA_HEADER( animAnimAbsFrame, Renderable, 1)
	int oldAbsFrame = e->anim.getAbsFrame();
	if (argc == 2) e->anim.setFromAbsFrame((int)lua_tonumber(L, 2));
	lua_pushinteger(L, oldAbsFrame);
	return 1;
}

EXPOSE_ENTITY_BOOL_PARAM_IN_LUA(colNoPhysics, Collidable, noPhysics)
EXPOSE_ENTITY_FLOAT_PARAM_IN_LUA( colRotation, Collidable, rotation)
EXPOSE_ENTITY_VEC2F_PARAM_IN_LUA( colVel, Collidable, vel)
EXPOSE_ENTITY_FLOAT_PARAM_IN_LUA( colMassInv, Collidable, massInv)
EXPOSE_ENTITY_BOOL_PARAM_IN_LUA(colUseGravity, Collidable, useGravity)
EXPOSE_ENTITY_BOX2F_PARAM_IN_LUA( colBBox, Collidable, bbox)
EXPOSE_ENTITY_BOOL_PARAM_IN_LUA(colIgnoreEnts, Collidable, collIgnoreEnts)
EXPOSE_ENTITY_BOOL_PARAM_IN_LUA(colIgnoreWorld, Collidable, collIgnoreWorld)
EXPOSE_ENTITY_BOOL_PARAM_IN_LUA(colTouchOnly, Collidable, collTouchOnly)
EXPOSE_ENTITY_BOOL_PARAM_IN_LUA(colDiffClass, Collidable, collDiffClass)
EXPOSE_ENTITY_FLOAT_PARAM_IN_LUA(colRestitution, Collidable, restitution)
EXPOSE_ENTITY_FLOAT_PARAM_IN_LUA(colFriction, Collidable, friction)
EXPOSE_ENTITY_VEC2F_PARAM_IN_LUA(colContactNormal, Collidable, contactNormal)
EXPOSE_ENTITY_VEC2I_PARAM_IN_LUA(colContactTile, Collidable, contactTile)
EXPOSE_ENTITY_FLOAT_PARAM_IN_LUA(colContactTime, Collidable, contactTime)
EXPOSE_ENTITY_ENTITY_PARAM_IN_LUA(colContactEnt, Collidable, contactEnt, Collidable)
EXPOSE_ENTITY_ENTITY_PARAM_IN_LUA(colIgnoreCollider, Collidable, collIgnore, Collidable)

// addLight params:
// cl			table with "_ptr" field with userdata pointing to our client context
// pos
// color		// defaults to 1 1 1
// tex			//defaults to -1
// range		// defaults to -10,-10, 10, 10
// spot			//defaults to pos
// rot			//defaults to 1,0
static int PF_addLight(lua_State *L) {
	int argc = lua_gettop(L);
	if (argc < 1) {
		con.print("usage: PF_addLight(vec2 pos[, vec3 color = (1,1,1), tex=-1, box2 range =(-10,10), vec2 spot=pos, vec2 rot = (1,0)]");
		return 0;
	}
	//init
	Light *l = lightSys.addLight();
	l->pos = luaReadVec<2,float>(L,1);
	//defaults
	l->color = 1.f;
	l->tex.i = -1;
	l->range.set(-10.f,10.f);
	l->spot = l->pos;
	l->rot.set(1.f,0.f);
	//fill
	if (argc >= 2 && lua_istable(L,2)) {
		l->color = luaReadVec<3,float>(L,2);
		if (argc >= 3) {
			l->tex.i = (int)lua_tonumber(L,3);
			if (argc >= 4 && lua_istable(L,4)) {
				lua_getfield(L, 4, "min");
				l->range.min = luaReadVec2f(L);
				lua_pop(L,1);
				lua_getfield(L, 4, "max");
				l->range.max = luaReadVec2f(L);
				lua_pop(L,1);
				if (argc >= 5 && lua_istable(L,5)) {
					l->spot = luaReadVec<2,float>(L, 5);
					if (argc >= 6 && lua_istable(L,6)) {
						l->rot = luaReadVec<2,float>(L,6);
					}
				}
			}
		}
	}
	l->process();
	return 0;
}

#define CLAMP(a,b,c) ( a < b ? b : (a > c ? c : a ) )

static int PF_drawLine(lua_State *L) {
	int argc = lua_gettop(L);
	if (argc < 8) {
		con.print("usage: PF_drawLine(pos,rot,dist,thick,tex,color,tc,useLighting");
		return 0;
	}

	vec2f pos = luaReadVec<2,float>(L,1);
	vec2f rot = luaReadVec<2,float>(L,2);

	float dist = lua_tonumber(L, 3);
	float thick = lua_tonumber(L, 4);
	TexLoad *tex = texSys.getObj(TexPtr((int)lua_tonumber(L, 5)));
	
	vec4f color = luaReadVec<4,float>(L,6).clamp(0,1) *= 255.f;
	vec4uc rcolor = color;
	
	vec2f tc = luaReadVec<2,float>(L,7);
	
	bool use_lighting = lua_toboolean(L, 8);
	
	int shader = (int)lua_tonumber(L, 9);

	bool postFBO = lua_toboolean(L,10);

	vec2f delta = rot * dist;
	vec2f normal(rot.y * thick, -rot.x * thick);

	EGeometry *g = renderSys.newGeom();
	g->texcoord[0].set(0,tc.y);
	g->texcoord[1].set(0,0);
	g->texcoord[2].set(tc.x,0);
	g->texcoord[3].set(tc);
	g->vertex[0].set(pos + normal);
	g->vertex[1].set(pos - normal);
	g->vertex[2].set(pos - normal + delta);
	g->vertex[3].set(pos + normal + delta);
	g->depth = -5.f;
	g->colormap = tex ? tex->tex.getTextureID() : 0;
	g->bumpmap = tex ? tex->bumpmap.getTextureID() : 0;
	g->color = rcolor;
	g->shader = shaderSys.getIndex(shader);
	g->blend = Blend();
	g->use_alpha_test = true;
	g->use_bump_mapping = true;
	g->use_lighting = use_lighting;
	g->postFBO = postFBO;
	g->dir = -rot;
	g->flip = 1;
	renderSys.geomDone(g);

	return 0;
}

// spawnobj = spawnObjFind(table of constraints: {key=value}, lastspawnobj)
using namespace std;
int PF_spawnObjFind(lua_State *L) {
	ostringstream s;
	try {
		const char *_str = lua_tostring(L, 1);
		if (!_str) throw Exception() << "first arg wasnt a string!";
		string str = _str;

		int index = -1;
		if (lua_istable(L, 2)) {
			lua_pushstring(L, "_spawnobj");
			lua_rawget(L, 2);
			index = (int)lua_tonumber(L, -1);
		}
		SpawnObj *so = spawnObjSys.get(index);
		so = spawnObjFindNext(str.c_str(), so);

		//this function should probably check for a cached copy first
		//thus making the _spawnobjs[] array a standard just like the _ents[] array
		if (!Lua_CreateSpawnObj(L, spawnObjSys.getIndex(so))) throw Exception() << "failed to create a new SpawnObj!\n"
				<< lua_tostring(L,-1) << "\n" << lua_tostring(L,-2);	//spawntable

		return 1;
	} catch (Exception &t) {
		con << "PF_spawnObjFind Error: " << t << "\n";
	}
	return 0;
}

//hack for the spawn object system - returns the number of spawnobjs
int PF_spawnObjGetCount(lua_State *L) {
	lua_pushinteger(L, spawnObjSys.count());
	return 1;
}

int PF_spawnObjSystemUpdateTags(lua_State *L) {
	spawnObjSystemUpdateTags((int)lua_tonumber(L, 1), (int)lua_tonumber(L, 2));
	return 0;
}
int PF_mapGetTag(lua_State *L) {
	World *world = worldSys.getWorld();
	if (!world) return 0;
	lua_pushinteger(L, world->tags->maploc((int)lua_tonumber(L,1),(int)lua_tonumber(L,2)));
	return 1;
}
int PF_mapGetType(lua_State *L) {
	World *world = worldSys.getWorld();
	if (!world) return 0;
	lua_pushinteger(L, world->getMap()->safeGet((int)lua_tonumber(L,1),(int)lua_tonumber(L,2), (int)lua_tonumber(L, 3))->type);
	return 1;
}
int PF_mapGetFriction(lua_State *L) {
	World *world = worldSys.getWorld();
	if (!world) return 0;
	lua_pushnumber(L, world->getMap()->safeGet((int)lua_tonumber(L,1), (int)lua_tonumber(L, 2), (int)lua_tonumber(L, 3))->friction[(int)lua_tonumber(L,4)]);
	return 1;
}
int PF_mapGetWidth(lua_State *L) {
	World *world = worldSys.getWorld();
	if (!world) return 0;
	lua_pushnumber(L, world->getMapSize().x);
	return 1;
}
int PF_mapGetHeight(lua_State *L) {
	World *world = worldSys.getWorld();
	if (!world) return 0;
	lua_pushnumber(L, world->getMapSize().y);
	return 1;
}

int PF_mapType(lua_State *L) {
	World *world = worldSys.getWorld();
	if (!world) return 0;
	if (!lua_gettop(L)) return 0;
	int index = (int)lua_tonumber(L,1);
	if (index < 0 || index >= mapTypes.size()) return 0;
	lua_pushstring(L, mapTypes[index].c_str());
	return 1;
}

int PF_mapTypeCount(lua_State *L) {
	World *world = worldSys.getWorld();
	if (!world) return 0;
	lua_pushinteger(L, mapTypes.size());
	return 1;
}


int PF_viewGetBounds(lua_State *L) {
	box2f viewarea = clientSys.getViewArea();
	lua_pushnumber(L, viewarea.min.x);
	lua_pushnumber(L, viewarea.min.y);
	lua_pushnumber(L, viewarea.max.x);
	lua_pushnumber(L, viewarea.max.y);
	return 4;
}

int PF_getFontCharWidth(lua_State *L) {
	if (!lua_gettop(L)) luaL_error(L, "usage: start, finish = PF_getFontCharWidth(ascii)");
	int ch = lua_tointeger(L, 1);
	ch -= 32;
	pwn::MenuFontTexture *ft = pwn::menuSys.getFont();
	if (ch < 0 || ch >= numberof(ft->widths)) return 0;
	if (!ft) return 0;
	pwn::MenuFontTexture::Width *w = ft->widths + ch;
	lua_pushnumber(L, w->start);
	lua_pushnumber(L, w->end);
	return 2;
}

int PF_drawGeometry(lua_State *L) {

	int luaIndex = 1;	//lua index

	EGeometry *g = renderSys.newGeom();
	//always assume 4 vtxcount -- always drawing quads ...
	float *fv = *g->vertex;
	float *tv = *g->texcoord;
	for (int i = 0; i < 8; i++) {
		//vtx
		lua_pushinteger(L, i+1);
		lua_rawget(L, luaIndex);
		*fv = lua_tonumber(L, -1);
		fv++;
		//texcoord
		lua_pushinteger(L, i+1);
		lua_rawget(L, luaIndex+1);
		*tv = lua_tonumber(L, -1);
		tv++;
	}
	luaIndex += 2;
	
	g->depth = lua_tonumber(L,luaIndex); luaIndex++;
	TexLoad *tex = texSys.getObj(TexPtr((int)lua_tonumber(L,luaIndex))); luaIndex++;
	g->colormap = tex ? tex->tex.getTextureID() : 0;
	g->bumpmap = tex ? tex->bumpmap.getTextureID() : 0;
	g->color = (vec4uc)(luaReadVec<4,float>(L,luaIndex++).clamp(0,1) *= 255.f);
	g->shader = shaderSys.getIndex((int)lua_tonumber(L, luaIndex)); luaIndex++;
	g->blend = getBlend((int)lua_tonumber(L,luaIndex)); luaIndex++;
	g->use_alpha_test = lua_toboolean(L,luaIndex); luaIndex++;		//alpha test
	g->use_bump_mapping = lua_toboolean(L,luaIndex); luaIndex++;	//bumpmap
	g->use_lighting = lua_toboolean(L,luaIndex); luaIndex++;		//lighting
	g->postFBO = lua_toboolean(L,luaIndex); luaIndex++;
	g->dir.set( lua_tonumber(L,luaIndex),	lua_tonumber(L,luaIndex+1) ); luaIndex += 2;
	g->flip = lua_tonumber(L,luaIndex); luaIndex++;					//flip
	
	renderSys.geomDone(g);

	return 0;
}

int PF_clientDrawRenderable(lua_State *L) {
	NetTableInstance *inst = netTablesSys.getCurrentInstance();
	if (!inst) return 0;
	//to cut down on lua/engine calls
	//this renders according to the params in the current nettable
/*
here is our nettable layout:
	active = 'bool',
	entClassID = 'short',
	pos = 'vec2f',
	absFrame = 'int',		-- 4 bytes
	angle = 'angle',		-- 2 bytes .. i bet we could reduce this to 12 bits or something
	left = 'bool',			-- 1 byte .. but it only needs 1 bit ...
	scale = 'vec2f',		-- 8 bytes ... that hardly anyone ever uses
	blend = 'uchar',		-- 1 byte, but not many bits are needed
	color = 'color'			-- 4 bytes
*/
	vec2f *pos = (vec2f*)inst->rawAccess("pos");						if (!pos) return 0;
	int *absFrame = (int*)inst->rawAccess("absFrame");					if (!absFrame) return 0;
	unsigned short *angle = (unsigned short*)inst->rawAccess("angle");	if (!angle) return 0;	//remap from unsigned short range to 360.f
	bool *left = (bool*)inst->rawAccess("left");						if (!left) return 0;
	vec2f *scale = (vec2f*)inst->rawAccess("scale");					if (!scale) return 0;
	unsigned char *blend = (unsigned char *)inst->rawAccess("blend");	if (!blend) return 0;
	vec4uc *color = (vec4uc*)inst->rawAccess("color");					if (!color) return 0;

	animSys.render(
		pos->x, pos->y,
		lua_tonumber(L,1),	//depth
		(float)*angle / 65536.f * 360.f,
		scale->x, scale->y, 
		*color,
		(int)lua_tonumber(L,2),	//shader
		getBlend(*blend),
		lua_toboolean(L,3),	//postFBO
		true,	//use lighting
		*left,
		*absFrame 
	);
	lua_pushboolean(L, true);
	return 1;
}
	
//no one uses this atm
//vec2f pos, float angle, vec2f scale, vec4f color, unsigned int shader, Blend blend, bool left, int absFrame, float depth
int PF_drawAbsFrame(lua_State *L) {
	int luaIndex = 1;
	vec2f pos = luaReadVec<2,float>(L,luaIndex++);
	float angle = lua_tonumber(L, luaIndex++);
	vec2f scale = luaReadVec<2,float>(L,luaIndex++);
	vec4f color = luaReadVec<4,float>(L,luaIndex++).clamp(0,1) *= 255.f;
	vec4uc rcolor = color;
	
	unsigned int shader = (int)lua_tonumber(L, luaIndex++);
	Blend blend = (int)lua_tonumber(L, luaIndex++);
	bool postFBO = lua_toboolean(L, luaIndex++);
	bool useLighting = lua_toboolean(L, luaIndex++);
	bool left = lua_toboolean(L, luaIndex++);
	int absFrame = (int)lua_tonumber(L, luaIndex++);
	float depth = lua_tonumber(L, luaIndex++);
	
	animSys.render(
		VEC2ELEM(pos),
		depth,
		angle,
		VEC2ELEM(scale),
		rcolor,
		shader,
		getBlend(blend),
		postFBO,
		useLighting,
		left,
		absFrame);
	
	return 0;
}

//vec2f pos, float angle, vec2f scale, vec4f color, Blend blend, bool left, string group, string seq, int frame
int PF_drawSprite(lua_State *L) {
	int luaIndex = 1;
	vec2f pos = luaReadVec<2,float>(L,luaIndex++);
	float angle = lua_tonumber(L, luaIndex++);
	vec2f scale = luaReadVec<2,float>(L,luaIndex++);
	vec4uc rcolor = (vec4uc)(luaReadVec<4,float>(L,luaIndex++).clamp(0,1) *= 255.f);
	unsigned int shader = (int)lua_tonumber(L, luaIndex++);	
	Blend blend = (int)lua_tonumber(L, luaIndex); luaIndex++;
	bool postFBO = lua_toboolean(L, luaIndex); luaIndex++;
	bool useLighting = lua_toboolean(L, luaIndex); luaIndex++;
	bool left = lua_toboolean(L, luaIndex); luaIndex++;
	const char *group = lua_tostring(L, luaIndex); luaIndex++;
	const char *seq = lua_tostring(L, luaIndex); luaIndex++;
	int frame = (int)lua_tonumber(L, luaIndex); luaIndex++;
	float depth = lua_tonumber(L, luaIndex); luaIndex++;
	
	AnimInfo anim;
	anim.setGroup(group);
	if (seq) anim.setSeq(seq);
	anim.frame = frame;
	int absFrame = anim.getAbsFrame();

	animSys.render(
		VEC2ELEM(pos),
		depth,
		angle, 
		VEC2ELEM(scale), 
		rcolor, 
		shader, 
		getBlend(blend), 
		postFBO, 
		useLighting, 
		left, 
		absFrame);
	
	return 0;
}

#define LUA_GLUE_PAIR(x)	{ #x, PF_##x }
#define LUA_GLUE_VEC2(x)	LUA_GLUE_PAIR(x), { #x"X", PF_##x##X }, {#x"Y", PF_##x##Y }, {#x"Table", PF_##x##Table }
#define LUA_GLUE_BOX2(x)	LUA_GLUE_PAIR(x), { #x"Table", PF_##x##Table }, LUA_GLUE_VEC2(x##Min), LUA_GLUE_VEC2(x##Max)
#define LUA_GLUE_COLOR(x)	LUA_GLUE_PAIR(x), { #x"X", PF_##x##X }, {#x"Y", PF_##x##Y }, {#x"Z", PF_##x##Z }, {#x"W", PF_##x##W }, {#x"Table", PF_##x##Table }

static const luaL_reg modGlueSessionLevelFuncs[] = {
	//entity class meta
	LUA_GLUE_PAIR(engineEntityMetaTable),
	LUA_GLUE_PAIR(engineRenderableMetaTable),
	LUA_GLUE_PAIR(engineCollidableMetaTable),

	//spawn system
	LUA_GLUE_PAIR(spawn),							//entity spawn(string spawnObjName)
	LUA_GLUE_PAIR(newEnt),							//entity newEnt([string classname?])
	//entity member methods
	LUA_GLUE_PAIR(entRemove),						//bool entRemove(entity who)
	LUA_GLUE_PAIR(entGetIndex),						//entity entGetIndex(int cycle index)
	LUA_GLUE_PAIR(entGetCount),						//int entGetCount()
	LUA_GLUE_PAIR(entServerSendMessage),			//void entServerSendMessage(entity who, ...)
	LUA_GLUE_PAIR(entClientSendMessage),			//clientside: void entClientSendMessage(int whoIndex, ...)
	LUA_GLUE_PAIR(entClientGetInput),				//clientside: input_cmd, mouse_active, mousepos_x, mousepos_y = entClientGetInput(entity)
	LUA_GLUE_PAIR(entActive),						//bool entActive(entity)
	LUA_GLUE_PAIR(entEntClassID),
	LUA_GLUE_PAIR(entNextThink),					// float exposure
	LUA_GLUE_PAIR(entLayer),
	LUA_GLUE_VEC2(entPos),							//vec2f pos
	LUA_GLUE_PAIR(entSpawnObj),						//[int so =] entSpawnObj(entity[, int so]);
	//renderable
	LUA_GLUE_PAIR(animAnimGroup),					//string animGroup(entity who, [string newGroup])
	LUA_GLUE_PAIR(animAnimSeq),						//string animSeq(entity who, [string newSeq])
	LUA_GLUE_PAIR(animAnimIdleSeq),					//string animIdleSeq(entity who, [string newIdleSeq])
	LUA_GLUE_PAIR(animAnimFrame),					//int animFrame(entity who, int newFrame)
	LUA_GLUE_PAIR(animAnimAbsFrame),				//int animAbsFrame(entity who, int newAbsFrame)
	LUA_GLUE_PAIR(animAnimFrameNextTime),
	LUA_GLUE_PAIR(animAnimFrameLastTime),
	LUA_GLUE_PAIR(animLeft),						//Renderable bool left
	LUA_GLUE_PAIR(animAngle),
	LUA_GLUE_VEC2(animScale),
	LUA_GLUE_PAIR(animBlend),
	LUA_GLUE_PAIR(animShader),
	LUA_GLUE_COLOR(animColor),
	//collidable
	LUA_GLUE_BOX2(colBBox),						//[{min={x,y},max={x,y}} =] colBBox(entity who [, {min={x=minx,y=miny},max={x=maxx,y=maxy}}])
	LUA_GLUE_PAIR(colRotation),							// float rotation
	LUA_GLUE_PAIR(colIgnoreEnts),					// bool ent param
	LUA_GLUE_PAIR(colIgnoreWorld),
	LUA_GLUE_PAIR(colTouchOnly),
	LUA_GLUE_PAIR(colDiffClass),
	LUA_GLUE_PAIR(colNoPhysics),
	LUA_GLUE_PAIR(colUseGravity),					//[bool useGrav =] colUseGravity(entity who[, bool newUseGrav])
	LUA_GLUE_VEC2(colVel),							//[... = ] colVel(entity, float, float)
	LUA_GLUE_VEC2(colContactNormal),
	LUA_GLUE_VEC2(colContactTile),
	LUA_GLUE_PAIR(colContactTime),
	LUA_GLUE_PAIR(colContactEnt),
	LUA_GLUE_PAIR(colIgnoreCollider),
	LUA_GLUE_PAIR(colMassInv),						//[... = ] colMassInv(entity, float)
	LUA_GLUE_PAIR(colFriction),
	LUA_GLUE_PAIR(colRestitution),
	//physics / util
	LUA_GLUE_PAIR(traceline),					//entHit, dist = traceline(vec2f start, vec2f delta, [ layer = -1, ignoreEnts = true, ignoreWorld = true, filter = nil])
	LUA_GLUE_PAIR(updateEntityPhysics),				//updateEntityPhysics(Collidable who)
	//net tables
	LUA_GLUE_PAIR(entUpdateNetInfo),
	LUA_GLUE_PAIR(animUpdateNetInfo),
	LUA_GLUE_PAIR(colUpdateNetInfo),
	//misc
	LUA_GLUE_PAIR(addLight),
	LUA_GLUE_PAIR(drawLine),
	LUA_GLUE_PAIR(spawnObjGetCount),
	LUA_GLUE_PAIR(spawnObjFind),
	LUA_GLUE_PAIR(spawnObjSystemUpdateTags),
	LUA_GLUE_PAIR(mapGetTag),
	LUA_GLUE_PAIR(mapGetType),
	LUA_GLUE_PAIR(mapGetFriction),
	LUA_GLUE_PAIR(mapGetWidth),
	LUA_GLUE_PAIR(mapGetHeight),
	LUA_GLUE_PAIR(mapType),
	LUA_GLUE_PAIR(mapTypeCount),
	LUA_GLUE_PAIR(drawAbsFrame),
	LUA_GLUE_PAIR(clientDrawRenderable),
	LUA_GLUE_PAIR(drawSprite),
	LUA_GLUE_PAIR(createShader),
	{NULL, NULL}
};

static const luaL_reg modGlueGameLevelFuncs[] = {
	LUA_GLUE_PAIR(getClientKeyMap),					//{keyname=keyvalue; ...} = getClientKeyMap()
	LUA_GLUE_PAIR(clientGetMenuInput),
	LUA_GLUE_PAIR(viewGetBounds),
	LUA_GLUE_PAIR(drawGeometry),
	LUA_GLUE_PAIR(getFontCharWidth),
	{NULL, NULL}
};


void GameAPI::progAddSessionLevelPredefined(lua_State *L) {
	for (const luaL_reg *reg = modGlueSessionLevelFuncs; reg->func; reg++) {
		lua_register(L, reg->name, reg->func);
	}
}

void GameAPI::progAddGameLevelPredefined(lua_State *L) {
	for (const luaL_reg *reg = modGlueGameLevelFuncs; reg->func; reg++) {
		lua_register(L, reg->name, reg->func);
	}
}


/*
we have too many engine-bound script functions all in the global namespace
we need to (a) organize them into tables and (b) make them context-based (rather than globals)
that means listing out all functions...

///////////////// PF_animLoad - now the only lua animation interface ///////////////// 

server side - animation system
	PF_animLoad			anim.load

server side - entity
	PF_entUpdateNetInfo			Entity:updateNetInfo
	PF_animUpdateNetInfo		Renderable:updateNetInfo
	PF_colUpdateNetInfo			Collidable:updateNetInfo
	PF_updateEntityPhysics		Entity:updatePhysics

server side - server	TODO - rename these to 'menusys' and give the member array to this ??
	PF_menu_new					menuSys.newMenu
	PF_menu_sysSize				menuSys.size
	PF_menu_sysSizeX			menuSys.sizeX
	PF_menu_sysSizeY			menuSys.sizeY
	PF_menu_root				menuSys.root
	
server side - menu		TODO - make this its own metatable and give it an __index member
	PF_menu_delete				menu:delete
	PF_menu_open				menu:open
	PF_menu_addChild			menu:addChild
	PF_menu_child				menu:child
	PF_menu_numChilds			menu:numChilds
	PF_menu_parent				menu:parent
	PF_menu_unlink				menu:unlink

server side - entity system and world system
	PF_traceline				entSys.traceline

server side - entity system
	PF_newEnt					Entity:new	<- or '.new' ?

server side - spawnobj system
	PF_spawn					Spawn:new ?

server side - entity
	PF_entRemove				Entity:delete
	PF_entActive				Entity:isActive
	
server side - entity system
	PF_entGetCount				entSys.getCount
	PF_entGetIndex				entSys.entForIndex
	PF_entServerSendMessage		Entity:serverSendMessage

server side - animation system
	PF_spriteLoad				anim.spriteLoad
	
server side - spawn obj system
	PF_spawnObjFind				spawnObjSys:find
	PF_spawnObjGetCount			spawnObjSys:getCount
	PF_spawnObjSystemUpdateTags	spawnobjSys:updateTags

server side - world
	PF_mapGetTag
	PF_mapGetType
	PF_mapGetFriction
	PF_mapGetWidth
	PF_mapGetHeight
	PF_mapType
	PF_mapTypeCount

server side - net table system
	PF_registerEntClass			entclass
	PF_netTableCreate			entclass:createNetTable
	
server side - net table instance access
	PF_netTable					net:

server side - threads
	PF_fork						process.new
	PF_wait						process.wait
	PF_sleep					process.sleep
	
server side - game timer
	PF_getTime
	PF_getDeltaTime
	PF_getLastTime
	PF_gameTimeScalar
server side - timeout
	PF_setTimeout
	PF_removeTimeout
	
server side - world system
	PF_loadWorld
	
server side - session var system
	PF_sessionVar
	
server side
	PF_cvarExec
	PF_createShader

server side - spawn object system / entity system
	PF_entForSpawnName

server side - spawn object system
	PF_spawnType
	PF_spawnObjMetaNewIndex
	PF_spawnObjMetaIndex

server side - entity
	EXPOSE_ENTITY_PARAM_IN_LUA_HEADER -- and all macros that use it ...

client side - client
	PF_getClientKeyMap

client side - entity system
	PF_entClientSendMessage
	PF_entClientGetInput

client side - render system
	PF_addLight
	PF_drawLine

client side - view
	PF_viewGetBounds

client side - render system
	PF_drawGeometry
	PF_clientDrawRenderable
	PF_drawAbsFrame
	PF_drawSprite
	
client - view system
	PF_viewTileSize
	
both
PF_getFontCharWidth

both
	PF_include	not used?
	
both - bit library
	PF_bitand			bit.band
	PF_bitor			bit.bor
	PF_bitxor			bit.bxor
	PF_bitnot			bit.bnot
	PF_bitshiftleft		bit.bshl
	PF_bitshiftright	bit.bshr
	
both - executable timer
	PF_getSysTime
	PF_getSysDeltaTime(lua_State*L)
	PF_getSysLastTime
	
both - console print
	PF_print
	

*/