/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include "system/gl.h"
#include <math.h>

#include "base/base.h"

#include "lighting.h"
#include "client.h"		//camera_range

//called after 'addlight' to finalize lighting info
void Light::process() {
	//todo - dynamic spacing vectors (for matrix col/row extraction and box corner extraction)

	//rotate xmin,ymin
	coord[0] = pos + vec2rotv(range.min, rot);
	//rotate xmax,ymin
	coord[1] = pos + vec2rotv(vec2f(range.max.x, range.min.y), rot);
	//rotate xmax,ymax
	coord[2] = pos + vec2rotv(range.max, rot);
	//rotate xmin,ymax
	coord[3] = pos + vec2rotv(vec2f(range.min.x, range.max.y), rot);

	//calc light bounds based upon which quadrant the rectangle was rotated by
	if (rot.y > 0) {
		if (rot.x > 0) {	//quadrant 1
			bbox.set(
				coord[3].x,	//xmin coord 2
				coord[0].y,	//ymin coord 0
				coord[1].x,	//xmax coord 1
				coord[2].y);//ymax coord 3
		} else {				//quadrant II
			bbox.set(
				coord[2].x,
				coord[3].y,
				coord[0].x,
				coord[1].y);
		}
	} else {					//quadrant III
		if (rot.x <= 0) {
			bbox.set(
				coord[1].x,
				coord[2].y,
				coord[3].x,
				coord[0].y);
		} else {				//quadrant IV
			bbox.set(
				coord[0].x,
				coord[1].y,
				coord[2].x,
				coord[3].y);
		}
	}
}


class LightSysImpl {
public:
	vec3f ambient;
	Light lights[256];
	unsigned int count;
	Light *base;
	int displayCount;

	LightSysImpl();
	bool sysInit();
	Light *addLight();
	void setAmbientLevel(const vec3f &ambient);
	vec3f getAmbientLevel();
	Light *getBase();
	unsigned int getCount();
};

LightSysImpl::LightSysImpl()
:	ambient(.5),
	count(0),
	displayCount(0)
{
	base = lights;
}

bool LightSysImpl::sysInit() { 
	if (displayCount) con.print("light count: %d\n", count);
	count = 0;
	return true;
}

Light *LightSysImpl::addLight() {
	if (count >= numberof(lights)) {
		//con << "AddLight() we maxed out on lights" << endl;
		count = numberof(lights) - 1;
	}

	Light *l = &lights[count];
	count++;
	
	*l = Light();

	return l;
}

Light *LightSysImpl::getBase() { return base; }

unsigned int LightSysImpl::getCount() { return count; }

void LightSysImpl::setAmbientLevel(const vec3f &ambient_) { ambient = ambient_; }
vec3f LightSysImpl::getAmbientLevel() { return ambient; }


LightSys::LightSys() : impl(new LightSysImpl()) {}
LightSys::~LightSys() { delete impl; }
bool LightSys::sysInit() { return impl->sysInit(); }
Light *LightSys::addLight() { return impl->addLight(); }
void LightSys::setAmbientLevel(const vec3f &ambient) { impl->setAmbientLevel(ambient); }
vec3f LightSys::getAmbientLevel() { return impl->getAmbientLevel(); }
Light *LightSys::getBase() { return impl->getBase(); }
unsigned int LightSys::getCount() { return impl->getCount(); }

LightSys lightSys;

static CVarw *local_cvars[] = {
	new CVarw(&lightSys->displayCount, "displayLightCount"),
};
