/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <algorithm>	//for sort
#include <map>		//windows needs these included for binary_function

#include <stdio.h>
#include <stdlib.h>
#include "system/exception.h"
#include "base/base.h"
#include "game/main.h"
#include "main.h"
#include "maptex.h"
#include "filesys.h"
#include "world.h"	//mapKeySys

using namespace std;

// *******************
// MAP TEX ARRAY
// *******************

const TexPtr MapTex::getCurrentImage() const {
	//test remap bounds
	if (curImage < 0 || curImage >= owner->texs.size()) return TexPtr();
	//return
	return owner->texs[curImage]->image;
}

struct ci_maptex : binary_function<MapTex*, MapTex*, bool> {
	bool operator() (const MapTex *a, const MapTex *b) { return a->texName < b->texName; }
};


void MapTexSet::reset() {
	texSys.removeAll(TEXTURE_WORLD);
	for (int i = 0; i < texs.size(); i++) delete texs[i];
	texs.resize(0);
}

bool MapTexSet::isValid(int index) {
	return index >= 0 && index < texs.size();
}

int MapTexSet::getIndex(MapTex *mt) {
	for (int i = 0; i < texs.size(); i++) {
		if (texs[i] == mt) return i;
	}
	return -1;
}

int MapTexSet::indexForName(const char *name) {
	for (int i = 0; i < texs.size(); i++) {
		if (!strcmp(name, texs[i]->texName.c_str())) return i;
	}
	return -1;
}

MapTex *MapTexSet::safeTexForName(const char *name) {
	int index = indexForName(name);
	if (index == -1) return NULL;
	return texs[index];
}

int MapTexSet::create() {
	texs.push_back(new MapTex(this));
	return texs.size()-1;
}

bool MapTexSet::destroy(int index) {
	assert(isValid(index));

	world->keys->removeTexIndex(index);	//linked from world
	
	//delete the entry
	delete texs[index];

	//and copy backwards
	for (; index < texs.size()-1; index++) {
		texs[index] = texs[index+1];
	}
	//decrement
	if (texs.size()) texs.resize(texs.size()-1);
	return true;
}

bool MapTexSet::set2D(int index, const char *filename) {
	MapTex *mt = texs[index];
	try {
		if (!filename || !filename[0]) throw Exception() << "got an invalid filename";

		//found in GameAPI
		string result;
		if (!fileSys.search(filename, game.getTexturePaths(), result)) throw Exception() << "failed to find texture";

		TexPtr texload = texSys.load(result.c_str(), TEXTURE_WORLD);
		if (!texload.isValid()) throw Exception() << "failed to load texture";
		mt->sources[0] = filename;
		mt->image = texload;
		mt->isCube = false;
		
		//set up some defaults for the cycling
		mt->curImage = index;
	} catch (Exception &t) {
		con << "MapTex::set2D(" << filename << ") failed with error: " << t << "\n";
		return false;
	}
	
	return true;
}

//parallels set2D so closely that something should be done about it
bool MapTexSet::set2DSub(int index, const char *filename, int x, int y, int w, int h) {
	MapTex *mt = texs[index];
	try {
		if (!filename || !filename[0]) throw Exception() << "got an invalid filename";

		//found in GameAPI
		string result;
		if (!fileSys.search(filename, game.getTexturePaths(), result)) throw Exception() << "failed to find texture";

		TexPtr texload = texSys.loadSub(result.c_str(), x,y,w,h, TEXTURE_WORLD);
		if (!texload.isValid()) throw Exception() << "failed to load texture";
		mt->sources[0] = filename;
		mt->image = texload;
		mt->isCube = false;
		
		//set up some defaults for the cycling
		mt->curImage = index;
	} catch (Exception &t) {
		con << "MapTex::set2DSub(" << filename << ", " << x << ", " << y << ", " << w << ", " << h << ") failed with error: " << t << "\n";
		return false;
	}

	return true;
}

//only absolute paths with this - i dont wanna think how to do the six dif tex lookups all before texcubeload ...
bool MapTexSet::setCube(int index, const char *filenames[7]) {
	MapTex *mt = texs[index];
	int i;
	for (i = 0; i < numberof(mt->sources); i++) {
		if (!filenames[i][0]) return false;
	}
	TexPtr texload = texSys.loadCube(filenames, TEXTURE_WORLD);
	if (!texload.isValid()) return false;
	for (i = 0; i < numberof(mt->sources); i++) {
		mt->sources[i] = filenames[i];
	}
	mt->image = texload;
	mt->isCube = true;
	
	//set up some defaults for the cycling
	mt->curImage = index;

	return true;
}

//load TexLoad's for each texture
void MapTexSet::refresh() {
	int i;

	//out with the old
	texSys.removeAll(TEXTURE_WORLD);

	//in with the new
	for (i = 0; i < texs.size(); i++) {
		MapTex *t = texs[i];
		if (!t->sources[0].length()) {
			t->image = TexPtr();
		} else if (t->subtexSize.x > 0 && t->subtexSize.y > 0) {
			set2DSub(i, t->sources[0].c_str(), VEC2ELEM(t->subtexOffset), VEC2ELEM(t->subtexSize));
		} else {
			set2D(i, t->sources[0].c_str());
		}

		t->nextIndex = indexForName(t->nextName.c_str());

		//todo - last - create a backup - dynamic copy - for texture cycling
		//include the image and the duration
		t->curImage = i;
		t->curDuration = t->duration;
	}
}

void MapTexSet::update() {
	for (vector<MapTex*>::iterator it = texs.begin(); it != texs.end() ;it++) {
		MapTex *t = *it;
		//get the current image's maptexarray index - this if condition should never be met!
		if (t->curImage < 0 || t->curImage >= texs.size()) continue;
		//only cycle if we have a tex to jump to
		MapTex *c = texs[t->curImage];
		if (c->nextIndex < 0 || c->nextIndex >= texs.size()) continue;

		t->curDuration -= game_deltatime;
		if (t->curDuration < 0) {
			t->curImage = c->nextIndex;
			t->curDuration = texs[c->nextIndex]->duration;
		}
	}
}

void MapTexSet::sort() {
	::sort<vector<MapTex*>::iterator, ci_maptex>(texs.begin(), texs.end(), ci_maptex());
}

MapTexSet::MapTexSet(World *world_) : world(world_) {}