/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <lua.hpp>

#ifdef WIN32
#include <direct.h>
#include "winsystem/dirent.h"
#define getcwd _getcwd
#define chdir _chdir
#define strcasecmp _stricmp
#else
#include <dirent.h>
#endif

#include "system/exception.h"
#include "system/singletons.h"
#include "base/base.h"
#include "menu/sys.h"

#include "game/client.h"	//lua_getInstFunc
#include "menu.h"
#include "main.h"
#include "world.h"
#include "client.h"
#include "program.h"

//#define BYPASS_MENU_STATES

using namespace std;

class MainMenuSystemImpl {
protected:
	string initMenuFile;
	lua_State *L;
public:
	vector<string> cmd;
	bool capturingKey;
	int capturedKey;

	MainMenuSystemImpl() : L(NULL), capturingKey(false), capturedKey(-1) {}
	void update();

	bool gameInit();
	void gameKill();
	
	void setInitMenuFile(const string &initMenuFile_) { initMenuFile = initMenuFile_; }
};

/*
how to do this...
scriptable...
with the old menu bindings...
or skip that

some things should be done at the engine level rather than the script level
like our basic menu...

	new game	-> hide splash screen, start a new game 
	load game	-> prompt w/list from 'savegames'
	save game	-> grey if no game going, saves state (pluto?)
	config		-> prompt with list of inputs
	settings		-> resolution, etc
	switch mod	-> list of mods

menus should be extensible beyond this by each mod


how should they be handled
there's an internal menu system cl/sv handled at the moment
its all script driven.  display is fed thru engine renderer and input thru cl/sv input
we could use that 
but, with the editor, we're already using the full-on menu system
so we might as well use that
that'd mean exposing it to the script, so we'd have two sets of menus that the script could see
one routed through game controls, the other otherwise
but this would mean doubling up the game<->input handling, which is already implemented in the cl/sv menu system
so why not just use that...

this would mean ... pushing everything back into the connection model (cl/sv)

but the main menu is different than the game interface -- it uses a preset key/mouse interface
so it should probably use the menu system that the editor uses

bleh to that
that'd mean two sets of lua menu bindings
and it'd mean two methods of lua menu inputs
so in its place, have a clientside object for empty times


hmm, we want a lua state ... with all the proper menu-connected code ...
but that means the game state, which is created on sessionInit ...
sessions are reserved for in-game experiences
which means we'll either be (a) creating a new lua state here for menu handling
or (b) piggybacking the init.lua state
both of which mean distinct lua states from the session/game lua state

though if a person wants to enter into the menu during a session/game ...
then we will want to use the game lua state ...
so we're talking about lua code that is gonna be read/launched from either the system-level init.lua/separate state
or run from teh session-level game state
either or ...

makes me think it just deserves its own
(which would mean a lot of repetative lua code in memory? all of the menu code at least?
which makes me lean towards menu bindings once again ...)
(cuz we all know if the engine is phone-ported, we're favoring the c++ menu sys over the lua menu sys)

maybe doubling up menu in memory code is fine,
as long as its only around when the user opens the menu???
that means we'll want a lua_State that is optionally close-able (and null-able)
...that gets re-created when the user enters into the menu
...and re-destroyed when the user exits it
...and can only do stuff via engine bindings (since it doesn't share state with the game)

this also means no forks ...
	which is a big problem for menu's serverside (which we wont use)
	
	
more thoughts ...
using lua's menus
separated out the menus from all ent-based code
so the menu lua state can use it separate
long story short: lua menus use client keys
client keys are/will be configurable by the main menu
so there will be a problem of menu key config potentially messing itself up?
*/

static int panic(lua_State *L) {
	con << "Lua Paniced!" << endl
		<< lua_tostring(L, -1) << endl
		<< lua_tostring(L, -2) << endl;
	return 0;
}

bool MainMenuSystemImpl::gameInit() {
	if (L) lua_close(L);
	L = NULL;
#ifdef BYPASS_MENU_STATES
	gameState = GS_RUNNING;
	mainSys.gameReset();
	clientSys.connect(1,0);
#endif
	return true;
}

void MainMenuSystemImpl::gameKill() {
	if (L) lua_close(L);
	L = NULL;
}

static int PF_getModList(lua_State *L) {
	//search the cwd for all dirs
	//if it isn't base...
	//look in them for an init.lua
	//if it's there...
	//add it to our list
	
	int modcount = 0;
	
	DIR *dirp = opendir(".");
	if (!dirp) {
		con << "mod find failed" << endl;
	} else {
		struct dirent *dp;
		while (dp = readdir(dirp)) {
			if (!strcasecmp(dp->d_name, "base")) continue;	//no loading the base
			if (!strcasecmp(dp->d_name, ".")) continue;
			if (!strcasecmp(dp->d_name, "..")) continue;
			FILE *f = fopen((string(dp->d_name) + "/init.lua").c_str(), "r");
			if (!f) continue;
			fclose(f);
			lua_pushstring(L, dp->d_name);
			modcount++;
		}
		closedir(dirp);
	}
	
	return modcount;
}

static int PF_mainMenuCmd(lua_State *L) {
	int argc = lua_gettop(L);
	if (argc < 1) luaL_error(L, "expected strings, didnt get anything");
	mainMenuSys->cmd.clear();
	for (int i = 1; i <= argc; i++) {
		const char *str = lua_tostring(L, i);
		if (!str) luaL_error(L, "expected a string, got another type");
		mainMenuSys->cmd.push_back(str);
	}
	return 0;
}

static int PF_getMenuCaptureKey(lua_State *L) {
	if (mainMenuSys->capturedKey < 0 || mainMenuSys->capturedKey >= Key::COUNT) return 0;
	const char *keyname = Key::name[mainMenuSys->capturedKey];
	if (!keyname) return 0;
	lua_pushstring(L, keyname);
	//clear it if you read a good value <=> you get one chance to interpret
	mainMenuSys->capturedKey = -1;
	return 1;
}

static int PF_getLocalClientCount(lua_State *L) {
	lua_pushnumber(L, MAX_LOCAL_CLIENTS);
	return 1;
}

static const luaL_reg mainMenuGlueFuncs[] = {
	{"getModList",			PF_getModList},
	{"mainMenuCmd",			PF_mainMenuCmd},
	{"getMenuCaptureKey",	PF_getMenuCaptureKey},
	{"getLocalClientCount",	PF_getLocalClientCount},
	{NULL, NULL}
};

void MainMenuSystemImpl::update() {
#ifndef BYPASS_MENU_STATES
	PROFILE();
	const char *mainMenu = "MainMenu";

	if (gameState == GS_MENU) {
		PROFILE();
		if (capturingKey) {
			PROFILE();
			//...then clear the clientsys input cmd here, before it might get accessed in the subsequent lua call
			clientSys.clearMenuKeys();
			//and grab anything in the buffer?
			if (kbd.getKeyDownQueueSize()) {
				capturedKey = kbd.getKeyDownQueue(0);
				capturingKey = false;
			}
		}
	
		if (!L) {
			PROFILE();
			try {
				PROFILE();
				if (!initMenuFile.length()) throw Exception() << "no initmenu provided";

				L = lua_open();
				lua_atpanic(L, panic);
				luaL_openlibs(L);
				//do any other binding here too
				luaStateAddFuncs(L, frameworkGlueFuncs);
				luaStateAddFuncs(L, engineGlueFuncs);
				luaStateAddFuncs(L, mainMenuGlueFuncs);
				game.progAddGameLevelPredefined(L);
				
				//then compile yer menu base code here ... probably mod/menu.lua
				//	but we don't remember mod dirs, so probably a var specified in init.lua (init'd during GameAPI::gameInit)
				ostringstream ss;
				ss << "require '" << initMenuFile << "'\n";
				string sstr = ss.str();
				const char *buffer = sstr.c_str();

				if (luaL_loadbuffer(L, buffer, strlen(buffer), "menuscript")) {
					throw Exception() << "loading script code:\n" << buffer << "\n\n"
						<< " we got some errors:\n"
						<< lua_tostring(L, -1) << "\n"
						<< lua_tostring(L, -2) << "\n";
				}

				{
					PROFILE();
					int base = lua_gettop(L);
					lua_pushliteral(L, "_TRACEBACK");
					lua_rawget(L, LUA_GLOBALSINDEX);
					lua_insert(L, base);
					if (lua_pcall(L, 0, 0, 0)) throw Exception() << "running base got lua error:\n" << lua_tostring(L, -1);
					lua_remove(L, base);
				}
				
				lua_getglobal(L, mainMenu);
				if (!lua_istable(L, -1)) throw Exception() << "expected " << mainMenu << " to be a table";
				const char *mainMenuInit = "init";
				lua_getInstFunc(L, mainMenuInit);
				if (!lua_isfunction(L, -2)) throw Exception() << mainMenuInit << " expected to be a function but found " << lua_typename(L, lua_type(L, -2));
				if (lua_pcall(L, 1, 0, 0)) throw Exception() << "lua_pcall failed: " << lua_tostring(L, -1);
				
			} catch (Exception &e) {
				con << "MainMenuSystemImpl::init failed to initialize menu: " << e << endl << "...proceeding with new game..." << endl;
				worldSys.loadInitWorld();
				clientSys.connect(1);
				gameState = GS_RUNNING;
				mainSys.gameInit();
				return;
			}
		} else {
			PROFILE();
			try {
				PROFILE();
				lua_getglobal(L, mainMenu);
				if (!lua_istable(L, -1)) throw Exception() << "expected " << mainMenu << " to be a table";
				const char *mainMenuUpdate = "update";
				lua_getInstFunc(L, mainMenuUpdate);
				if (!lua_isfunction(L, -2)) throw Exception() << mainMenuUpdate << " expected to be a function but found " << lua_typename(L, lua_type(L, -2));
				if (lua_pcall(L, 1, 0, 0)) throw Exception() << "lua_pcall failed: " << lua_tostring(L, -1);
			} catch (Exception &e) {
				con << "MainMenuSystemImpl::update failed to initialize menu: " << e << endl << "...proceeding with new game..." << endl;
				worldSys.loadInitWorld();
				clientSys.connect(1);
				gameState = GS_RUNNING;
				mainSys.gameInit();
				return;
			}			
		}
	} else {
		PROFILE();
		if (L) lua_close(L);
		L = NULL;
	}
	
	if (cmd.size()) {
		PROFILE();
		if (cmd[0] == "newgame") {
			PROFILE();
			gameState = GS_RUNNING;
			mainSys.gameReset();
			clientSys.connect(1,0);
		} else if (cmd[0] == "resetmod") {
			PROFILE();
			gameState = GS_MENU;
			mainSys.gameReset();
		} else if (cmd[0] == "quit") {
			PROFILE();
			sys.requestExit();
		} else if (cmd[0] == "capturekey") {
			PROFILE();
			capturingKey = true;
		}
		cmd.clear();
	}
#endif
}

MainMenuSystem::MainMenuSystem() : impl(new MainMenuSystemImpl()) {}
MainMenuSystem::~MainMenuSystem() { delete impl; }
void MainMenuSystem::update() { impl->update(); }
void MainMenuSystem::setInitMenuFile(const string &initMenuFile_) { impl->setInitMenuFile(initMenuFile_); }
bool MainMenuSystem::gameInit() { return impl->gameInit(); }
void MainMenuSystem::gameKill() { impl->gameKill(); }
MainMenuSystem mainMenuSys;
