/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <assert.h>
#include <stdio.h>
#include <math.h>

#include <vector>
#include <sstream>

#include "base/base.h"
#include "system/memory.h"
#include "system/exception.h"

#include "game/program.h"

//for spawn object functions:
#include "entalloc.h"
#include "client.h"
#include "main.h"
#include "spawnobj.h"
#include "program.h"
#include "strcasecmp.h"

using namespace std;

// *******************************************************************
//							SPAWN OBJECT TYPES
// *******************************************************************

//generic spawntype class
class SpawnType {
public:

	//ctor

		//content
	SpawnType(const char *_name, const char *_desc, const char *_animGroup, const box2f &_bbox)
	: name(_name), desc(_desc), animGroup(_animGroup), bbox(_bbox) {}
	
	//methods

	virtual const char *getName() { return name.c_str(); }
	virtual const char *getAnimGroup() { return animGroup.c_str(); }
	virtual box2f		getBBox() { return bbox; }
	virtual const char *getDesc() { return desc.c_str(); }
	virtual entptr_t	spawn(SpawnObj *) = 0;

	//member vars
	
	string		name;
	string		desc;
	string		animGroup;
	box2f		bbox;
};

static const char *spawnTypeTableName = "_spawntypes";
//the lua spawntype table will parallel the C one ... sorta
//here's how it'll work:
//_spawntypes = {
//	<typename> = {
//		group = "<groupname>",			-- animation group
//		desc = "<desc>",				-- entity description
//		spawnfunc = function () ... end	-- entity = function spawnfunc(table params) ... end
//		bbox = {xmin,xmax,ymin,ymax}	-- bounding box (consider lib'ing and revamping)
//	}
//}


/*
spawnobject structure:
integer _spawnobj	- spawnobject index
metatable --
	by default grabs the key specified with some exceptions ...
	_ent	- the entity associated with the spawnobject
*/
bool Lua_CreateSpawnObj(lua_State *L, int soIndex) {
	try {
		lua_getglobal(L, "SpawnObj");	//[spawnobj]
		if (lua_isnil(L, -1) || !lua_istable(L, -1)) throw Exception() << "failed to find the SpawnObj class";
		lua_getfield(L, -1, "new");		//[spawnobj] [spawnobj.new]
		if (lua_isnil(L, -1) || !lua_isfunction(L, -1)) throw Exception() << "failed to find SpawnObj.new";
		lua_insert(L, lua_gettop(L)-1);	//[spawnobj.new] [spawnobj]
		lua_pushinteger(L, soIndex);	//[SpawnObj.new] [SpawnObj] [index]
		if (lua_pcall(L, 2, 1, 0)) throw Exception() << "lua call SpawnObj:new() failed with error: " << lua_tostring(L, -1) << "\n" << lua_tostring(L, -2);
	} catch (Exception &t) {
		con << "Lua_CreateSpawnObj(" << soIndex << ") failed: " << t << "\n";
		return false;
	}
	return true;
}

//spawntype instance for Script callbacks
class SpawnTypeScript : public SpawnType {
	typedef SpawnType super;
public:
	SpawnTypeScript(const char *_name, const char *_desc, const char *_group, const box2f &_bbox)
	: super(_name, _desc, _group, _bbox) {}
	
	virtual entptr_t spawn(SpawnObj *so) {
		entptr_t ent = {-1};
		//see if we can spawn the string through Lua's main state ...
		lua_State *L = progSys.getState();
		assert(L);
		assert(so);
		
		//check the "_spawntype" table for a key with our type name:
		int tableTop = lua_gettop(L);	//uniracers!
		ostringstream s;
		try {
			lua_getglobal(L, spawnTypeTableName);	//[main table]
			if (!lua_istable(L, -1)) throw 0;		//spawn type table existence is optional
			lua_rawgetfield(L, -1, name.c_str());	//[main	table] [type.value == spawntype table]
			if (!lua_istable(L, -1)) throw Exception() << "expected all entries in the spawntype table to map to tables!";
			lua_rawgetfield(L, -1, "spawnfunc");	//[main table] [type table] [spawnfunc]
			if (!lua_isfunction(L, -1)) throw Exception() << "expected a \"spawnfunc\" associated with spawntype table key " << name;
			//push anything on top?  not at the moment...
			//TODO - build and push a table with all key/val spawnobj relations...

			if (!Lua_CreateSpawnObj(L, spawnObjSys.getIndex(so))) throw Exception() << "failed to create a new SpawnObj!\n"
				<< lua_tostring(L,-1) << "\n" << lua_tostring(L,-2);	//spawntable

			//for the current spawnobj
			if (lua_pcall(L, 1, 1, 0)) throw Exception() << "error while executing \"spawnfunc\" of type " << name << "\n"
				<< lua_tostring(L,-1) << "\n" << lua_tostring(L,-2);
			//expect one table on the stack: our new entity
			lua_rawgetfield(L, -1, "_ptr");
			ent.i = (int)lua_tonumber(L, -1);
			if (!entAllocSys.entForPtr(ent)) throw Exception() << "spawnfunc of type " << name << " returned an invalid entity!";
		} catch (Exception &t) {
			con << "SpawnTypeScript::spawn failed: " << t << "\n";
		} catch (int) {
		}
		lua_settop(L, tableTop);	//ha ha, uniracers roxxors
		return ent;
	}
};

// ************** spawn type system

class SpawnTypeSys : public ESpawnTypeSys {
	vector<SpawnType *> spawntypes;
public:

	//engine code
	
	virtual SpawnType *findByName(const char *name) {
		for (int i = 0; i < spawntypes.size(); i++) {
			SpawnType *st = spawntypes[i];
			assert(st);
			if (!strcmp(st->name.c_str(), name)) return st;
		}
//		con.print("spawnTypeFindByName(): couldn't find type: %s\n", name);
		return NULL;
	}
	
	virtual const char *getDescByName(const char *name) {
		SpawnType *st = findByName(name);
		return st ? st->getDesc() : "<unknown type>";
	}

	virtual int getIndexByName(const char *name) {
		SpawnType *st = findByName(name);
		if (!st) return -1;
		return getIndex(st);
	}
	
	virtual const char *getNameForIndex(int index) {
		if (index < 0 || index >= spawntypes.size()) return "";
		SpawnType *st = spawntypes[index];
		if (!st) return "";
		if (!st->name.length()) return "";
		return st->name.c_str();
	}
	
	virtual int getGroupIndex(int index) {
		if (index < 0 || index >= spawntypes.size()) return 0;
		SpawnType *st = spawntypes[index];
		if (!st) return 0;
		string group = st->getAnimGroup();
		AnimInfo anim;
		memset(&anim, 0, sizeof(anim));
		anim.setGroup(group);
		return anim.group;
	}

	virtual SpawnType *get(int index) {
		if (index < 0 || index >= spawntypes.size()) return NULL;
		return spawntypes[index];
	}
	
	virtual int getIndex(SpawnType *type) {
		for (int i = 0; i < spawntypes.size(); i++) {
			if (spawntypes[i] == type) return i;
		}
		return -1;
	}

	virtual int count() { return spawntypes.size(); }

	//mod code

	virtual void clear() {
		VECTOR_DELETE_ALL(spawntypes);
	}

	
	//add via Lua callback function name
	virtual bool addScript(
		const char *name,
		const char *desc,
		const char *group_name,
		const box2f &bbox)
	{
		SpawnType *s = new SpawnTypeScript(name, desc, group_name, bbox);
		if (s) spawntypes.push_back(s);
		return !!s;
	}
};
static SpawnTypeSys g_spawnTypeSys;
ESpawnTypeSys &spawnTypeSys = g_spawnTypeSys;

// *******************************************************************
//							SPAWN OBJECTS
// *******************************************************************

//all refs for the SpawnObj->createEnt code

//TODO - replace the spawnObjMap with simply accessing the spawn object's ent field
//this is dangerous!  it allows the script to rewrite the spawn object's entity
//that could lead to mistakes when performing mod tasks such as repopulating tagged areas
//SOLUTION: make some script variables read-only!
//other issues: in mid-editing a game, spawn objects are moved around, thus displacing variable
//references. this is fixed when we use a separate map - they are not touched until the game is restarted
//
//1-1 map between the elements of this array and the elements
//of the spawn_obj array:
//spawn_obj[i].ent == entity_list[spawnObjMap[i]]

static int spawnObjMap[1024];

bool SpawnObj::createEnt() {

	ent.i = -1;

//		con.print(" -- spawn object type name: %s\n", spawnObjTypeTable[type.i]->name);
	string classname;
	if (!(rw("classname") >> classname)) return false;
	SpawnType *st = g_spawnTypeSys.findByName(classname.c_str());
	if (st) ent = st->spawn(this);
	
	//spawn failed
	if (ent.i == -1) return false;

	//should this be here, or in the init function?
	//move it to the init function once we abstract entities from the engine?
	entAllocSys.entForPtr(ent)->spawnobj = this;

	if (!entAllocSys.isActive(ent)) {
		con.print("Could not create entity from spawn object\n");
		return false;
	}
	//change the spawn object with matching name to point to the right entity
	int index = spawnObjSys.getIndex(this);
	assert(index != -1);
	spawnObjMap[index] = ent.i;
	
	//rather than just the map ...
	//in Lua ...
	// just write to global!
	string name;
	if (rw("name") >> name && name.length()) {
		lua_State *L = progSys.getState();
		lua_getglobal(L, "_ents");		//[_ents]
		lua_pushinteger(L, ent.i);		//[_ents] [_ptr]
		lua_rawget(L, -2);			//[_ents] [ent]
		lua_setglobal(L, name.c_str());	//[_ents]	;	name = _ents[_ptr]
		lua_pop(L,1);					//clear!
	}

	return true;
}

const char *SpawnObj::getTypeGroup() {
	static const char *nowhere = "null";
	string classname;
	if (!(rw("classname") >> classname)) return nowhere;
	SpawnType *st = g_spawnTypeSys.findByName(classname.c_str());
	if (!st) return nowhere;
	return st->getAnimGroup();
}

box2f SpawnObj::getTypeBBox() {
	string classname;
	if (!(rw("classname") >> classname)) return box2f(-.5f,-.5f,.5f,.5f);
	SpawnType *st = g_spawnTypeSys.findByName(classname.c_str());
	if (!st) return box2f(-.5f, -.5f, .5f, .5f);
	return st->getBBox();
}

// ************** the fatty spawn object array

static int PF_entForSpawnName(lua_State *L);

//todo
//add spawnObjCreateEntity
//and spawnObjArrayUnlinkEntity
//to the system
//then private-ize the spawnObjMap
class SpawnObjSys : public ISpawnObjSys {
	vector<SpawnObj *> spawnobjs;
public:

	virtual void sysKill() {
		VECTOR_DELETE_ALL(spawnobjs);
	}
	
	virtual void sessionKill() {
		spawnTypeSys.clear();
	}

	// *******************************************************************
	//new game function:
	
	virtual void newSystem() {
		VECTOR_DELETE_ALL(spawnobjs);	//clear all our spawn objects
	}
	
	// *******************************************************************
	//script
	virtual void addScriptGlobals() {
		lua_register(progSys.getState(), "entNamed", PF_entForSpawnName );
	}

	//array access
	virtual SpawnObj *get(int index) {
		if (index < 0 || index >= spawnobjs.size()) return NULL;
		return spawnobjs[index];
	}
	
	virtual SpawnObj *createByName(const char *name) {
		//constrain
		if (spawnobjs.size() == numberof(spawnObjMap)) {
			return NULL;
		}
		//grab
		SpawnObj *s = new SpawnObj();//&spawn_obj[spawn_obj_count];	//clear	memset(s, 0, sizeof(SpawnObj));
		if (!s) return NULL;
		spawnobjs.push_back(s);
		//init
		s->rw("classname") << name;
		s->ent.i = -1;
		//and return
		return s;
	}


	virtual int removeIndex(int index) {	//seems best to still do this by index
		//check validity
		if (!get(index)) return spawnobjs.size()-1;
	
		//unattach all other objects pointing to the removed object *HERE*
		int i;
		for (i = 0; i < spawnobjs.size(); i++) {
			SpawnObj *s = spawnobjs[i];
			assert(s);
			if (i == index) continue;
			
			//UT OH! WE NO LONGER HAVE THE ABILITY TO DECREMENT SPAWNOBJ INDICES!
			//solution? DO IT ALL IN LUA
		}

		//make sure what we're deleting is actually there
		assert(index >= 0 && index < spawnobjs.size());
		assert(spawnobjs[index]);
	
		//do the removal
		//spawn_obj_count--;
		//for (i = index.i; i < spawn_obj_count; i++) {
		//	memcpy(&spawn_obj[i], &spawn_obj[i+1], sizeof(SpawnObj));
		//}
		delete spawnobjs[index];
		for (i = index; i < spawnobjs.size()-1; i++) {
			spawnobjs[i] = spawnobjs[i+1];
		}
		spawnobjs.resize(spawnobjs.size()-1);	//size us down one
	
		//return the next valid, if any are left
		if (index == spawnobjs.size()) index--;
	
		return index;
	}

	//funny thing, i think this is exposed mod-side through ISpawnObj::getIndex
	//oh well. its only exposed engine-side here.
	virtual int getIndex(SpawnObj *o) {
		for (int i = 0; i < spawnobjs.size(); i++) {
			SpawnObj *s = spawnobjs[i];
			assert(s);
			if (s == o) return i;
		}
		return -1;
	}

	virtual void unlinkEnt(entptr_t e) {
		for (int i = 0; i < spawnobjs.size(); i++) {
			SpawnObj *s = spawnobjs[i];
			assert(s);
			if (s->ent.i == e.i) {
				s->ent.i = -1;
				//rewrite the spawn object map entry as well
				spawnObjMap[i] = -1;
			}
		}
	}

	//mod-side:
	
	virtual int count() { return spawnobjs.size();	}
};
static SpawnObjSys g_spawnObjSys;
ISpawnObjSys &spawnObjSys = g_spawnObjSys;

//int spawnobj(string name) //returns the entity index associated with the spawn object
//i believe i was considering making such a function called "spawn" before ?
//but one that spawned the object if it wasnt yet created
//
//if only i knew how to add in object access directly to lua . . . insert these objects into it and what not . . .
static int PF_entForSpawnName(lua_State *L) {
	if (!lua_gettop(L)) {
		con.print("usage: entity ent = PF_entForSpawnName(string spawnObjName)\n");
		return 0;
	}
	const char *findname = lua_tostring(L, 1);
	if (!findname) {
		con.print("PF_entForSpawnName expected a name!\n");
		return 0;
	}

	for (int i = 0; i < spawnObjSys.count();i++) {
		SpawnObj *s = spawnObjSys[i];
		assert(s);
		string name;
		if (s->rw("name") >> name && name == findname) {
			lua_pushinteger(L, s->ent.i);
			return 1;
		}
	}
	return 0;
}


//internal spawntype registration
int PF_spawnType(lua_State *L) {
	//expect a table
	if (lua_gettop(L) < 2 || !lua_isstring(L, 1) || !lua_istable(L, 2)) {
		con << "PF_spawnType expected a table param\n";
		return 0;
	}
	const char *str;
	str = lua_tostring(L,1);
	assert(str);
	string name(str);
//con << "PF_spawnType got name to be " << name << "\n";
	//swap orders right off the bat
	lua_settop(L,2);	//[newname] [newtable]
	lua_insert(L,1);	//[newtable] [newname]
	//make sure our spawntype table is there
//con << "looking for the spawn type table ... \n";
	lua_getglobal(L, spawnTypeTableName);
	if (!lua_istable(L, -1)) {
//con << "not there -- creating \n";
		lua_pop(L,1);
		lua_newtable(L);
		lua_setglobal(L, spawnTypeTableName);
		lua_getglobal(L, spawnTypeTableName);
//con << "created\n";
	}	//[newtable] [newname] [maintable]
	//try and add in our new entry
//con << "adding new entry to it\n";
	lua_insert(L, 2);	//[newtable] [maintable] [newname]
	lua_pop(L,1);		//[newtable] [maintable]
	lua_newtable(L);	//[newtable] [maintable] [emptytable]
	lua_rawsetfield(L, 2, name.c_str());	//[newtable] [maintable={newname=emptytable}]
	lua_rawgetfield(L, 2, name.c_str());	//[newtable] [maintable] [emptytable]
	assert(lua_istable(L,-1));
	lua_insert(L, 2);	//[newtable] [emptytable] [maintable]
	lua_pop(L,1);		//[newtable] [emptytable]
	assert(lua_istable(L,1));
	assert(lua_istable(L,2));
	//but we want that table back on top
	//now that its there, and on top of the stack, lets start referencing stuff from it
	//TODO - any typechecking?  any registering with the usual spawnType array?  subclass it between C and Lua spawntypes? for the editor's sake
	//(maybe abstrating the spawntype system to strings wasn't a necessary move after all ... 
//con << "copying fields over..\n";
	lua_rawgetfield(L, 1, "group");
	string group = (str = lua_tostring(L,-1)) ? str : "unknown";
	lua_rawsetfield(L, 2, "group");
	lua_rawgetfield(L, 1, "desc");
	string desc = (str = lua_tostring(L,-1)) ? str : "";
	lua_rawsetfield(L, 2, "desc");
	lua_rawgetfield(L, 1, "spawnfunc");
	//dont get the spawnfunc ... leave that in the table -- we'll reference it using the spawntype name
	lua_rawsetfield(L, 2, "spawnfunc");
	lua_rawgetfield(L, 1, "bbox");
	str = lua_tostring(L,-1);
	box2f bbox(-0.5, -0.5, 0.5, 0.5);
	if (str) {
		string bboxstr = str;
		sscanf(bboxstr.c_str(), "%f %f %f %f", &bbox.min.x, &bbox.min.y, &bbox.max.x, &bbox.max.y);
	}
	lua_rawsetfield(L, 2, "bbox");
//con << "registering spawntype\n";
	g_spawnTypeSys.addScript(name.c_str(), desc.c_str(), group.c_str(), bbox);
//con << "PF_spawnType end\n";
	return 0;
}


//other than updating the spawnobj ent ptrs
//why don't we just make a lua copy of the spawnobj and feed it to the spawn functions?


//void __newindex(table,key,value)
//don't forget to use rawget rather than getfield in here ... or else
int PF_spawnObjMetaNewIndex(lua_State *L) {
	string key = lua_tostring(L,2);
	lua_rawgetfield(L, 1, "_spawnobj");
	int soi = (int)lua_tonumber(L, -1);
	SpawnObj *so = spawnObjSys[soi];
	if (!so) {
		con << "spawnObjMetaNewIndex had bad spawnobj #: " << soi << "\n";
		return 0;
	}

	if (!strcmp(key.c_str(), "_ent")) {
		entptr_t ep = { -1 };
		if (lua_istable(L,3)) {
			lua_pushstring(L, "_ptr");
			lua_rawget(L,3);
			ep.i = (int)lua_tonumber(L,-1);
		}
		so->ent = ep;
	}	//else...?
	return 0;
}

//value __index(table, key)
int PF_spawnObjMetaIndex(lua_State *L) {
	string key = lua_tostring(L, 2);
	lua_pushstring(L,"_spawnobj");
	lua_rawget(L, 1);
	int soi = (int)lua_tonumber(L, -1);
	SpawnObj *so = spawnObjSys[soi];
	if (!so) {
		con << "tried to index the key " << key << " from a bad spawnobj #: " << soi << "\n";
		return 0;
	}
	
	if (!strcmp(key.c_str(), "_ent")) {
		int i = so->ent.i;
		if (i == -1) lua_pushnil(L);
		else lua_pushinteger(L, i);
		return 1;		
	}
	
	map<string,string>::iterator it = so->m.find(key);
	if (it != so->m.end()) {
		lua_pushstring(L, it->second.c_str());
		return 1;
	}
				
	return 0;
}

