/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <stdio.h>
#include <math.h>

#include "system/gl.h"
#include "base/base.h"
#include "system/exception.h"
#include "system/memory.h"
#include "texloader.h"
#include "strcasecmp.h"

using namespace std;

class TexSys : public ITexSys {
public:
	vector<TexLoad*> texture;

	TexSys() {}

	virtual bool sysInit() {
		reset();
		return true;
	}
	virtual void sysKill() {
		reset();
	}
	
	//internal
	void reset() {
		for (vector<TexLoad*>::iterator i = texture.begin(); i != texture.end(); ++i) {
			if (*i) delete *i;
		}
		texture.resize(0);
	}
	
	virtual void removeAll(int key) {
		for (vector<TexLoad*>::iterator i = texture.begin(); i != texture.end(); ++i) {
			if (*i && (*i)->tex_id == key) {
				delete *i;
				*i = NULL;
			}
		}
	}
	
	virtual TexPtr load(const char *filename, int key, int flags) {
		if (!filename) {
			con << "TexSys::load was given a null filename" << endl;
			return TexPtr();
		}

		//return a previously loaded texture
		int i;
		for (i = 0; i < texture.size(); i++) {
			if (texture[i] &&
				!strcasecmp(texture[i]->title.c_str(), filename) &&		//if the textures are the same
				texture[i]->tex_id == key)						//and the ids are the same
			{
				return TexPtr(i);
			}
		}

		//search for an empty slot to fill
		for (i = 0; i < texture.size(); i++) {
			if (!texture[i]) {
				texture[i] = new TexLoad();
				break;
			}
		}

		if (i >= texture.size()) {
			i = texture.size();
			texture.push_back(new TexLoad());
		}

		//todo - cache here - not doing so will proly cause errors
		texture[i]->title = filename;
		texture[i]->tex_id = key;
		bool active = texture[i]->tex.load2D(filename);
		if (!active) {
			delete texture[i];
			texture[i] = NULL;
			con << "TexSys::load was unable to load filename " << filename << endl;
			return TexPtr();
		}

	//	con.print("texture ID: %d\n", texture[i]->tex.texture_id);

		if ((flags & TEXLOAD_NO_BUMPMAP) == 0) {
			generateBumpMap(texture[i]);
		}

		//and return our goods
		return TexPtr(i);
	}

	virtual TexPtr loadSub(const char *filename, int x, int y, int w, int h, int id) {
		if (!filename) return TexPtr();

		char newtexname[256];
		sprintf(newtexname, "%s@%.4d,%.4d,%.4d,%.4d", filename, x,y,w,h);

	//	con.print("loading subtex %s\n", newtexname);

		//if our subtex is previously loaded then return it
		int i;
		for (i = 0; i < texture.size(); i++) {
			if (texture[i] &&
				!strcasecmp(texture[i]->title.c_str(), newtexname) &&		//if the textures are the same
				texture[i]->tex_id == id)						//and the ids are the same
			{
				//con.print("found previously cached subtex! returning\n");
				return TexPtr(i);
			}
		}

		Texture *parentTex = NULL;

		//return a previously loaded texture
		for (i = 0; i < texture.size(); i++) {
			if (texture[i] &&
				!strcasecmp(texture[i]->title.c_str(), filename))		//if the textures are the same
			{
				//con.print("found parent tex: %s\n", texture[i]->title);
				parentTex = &texture[i]->tex;
				break;
			}
		}

		if (!parentTex) {
			//con.print("caching parent tex from file: %s\n", filename);

			//search for an empty slot to fill
			for (i = 0; i < texture.size(); i++) {
				if (!texture[i]) {
					texture[i] = new TexLoad();
					break;
				}
			}

			if (i >= texture.size()) {
				i = texture.size();
				texture.push_back(new TexLoad());
			}

			//con.print("attempting to load texture file %s\n", filename);

			texture[i]->title = filename;
			texture[i]->tex_id = id;
			bool active = texture[i]->tex.load2D(filename);
			if (!active) {
				delete texture[i];
				texture[i] = NULL;
				con.print("failed to create parent texture.  returning\n");
				return TexPtr();
			}

			//con.print("texture file %s loaded successfully\n", filename);

			parentTex = &texture[i]->tex;
		}

	//	con.print("returning parent tex [DEBUG]\n");
	//	return parentTex;

	//	con.print("caching subtex...\n");

		//now to create the actual subtex
		//search for an empty slot to fill
		for (i = 0; i < texture.size(); i++) {
			if (!texture[i]) {
				texture[i] = new TexLoad();
				break;
			}
		}

		//unable to comply - didnt find an empty slot
		if (i >= texture.size()) {
			i = texture.size();
			texture.push_back(new TexLoad());
		}

		//grab the texture width and height
		glBindTexture(GL_TEXTURE_2D, parentTex->getTextureID());	//parentTex->bind();
		
		GLint parent_width = 0;
		GLint parent_height = 0;
		glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &parent_width);
		glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT, &parent_height);

		if (!parent_width || !parent_height) {
			delete texture[i];
			texture[i] = NULL;
			con << "unable to read parent tex dim.  returning" << endl;
			return TexPtr();
		}

	//	con.print("read parent tex dim as %d x %d\n", parent_width, parent_height);

		unsigned char *data = NULL;
		data = (unsigned char *)MEMALLOCNAMED(4 * parent_width * parent_height, "texture data");
		if (!data) {
			delete texture[i];
			texture[i] = NULL;
			con.print("failed to alloc temp buffer\n");
			return TexPtr();
		}

	//	con.print("parent dim: %d %d subtex dim: %d %d  subtex ofs: %d %d\n",
	//		parent_width, parent_height, w, h, x, y);

		glBindTexture(GL_TEXTURE_2D, parentTex->getTextureID());	//parentTex->bind();
		glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);

		texture[i]->title = newtexname;
		texture[i]->tex_id = id;
		bool active = texture[i]->tex.createEmpty(w,h,GL_RGBA, GL_UNSIGNED_BYTE);
		active |= texture[i]->tex.build2DMipmaps(w, h, GL_RGBA, GL_UNSIGNED_BYTE, &data[(x + y * parent_width) * 4], 4, parent_width * 4);

		memFree(data);

		if (!active) {
			delete texture[i];
			texture[i] = NULL;
			con.print("failed to create subtex.  returning\n");
			return TexPtr();
		}

		//con.print("subtex ID: %d\n", texture[i].tex.texture_id);
		//con.print("everything worked well.  returning\n\n");

		generateBumpMap(texture[i]);

		return TexPtr(i);
	}

	//this is halfass
	//todo	- up the 'title' array size and store all six names within it
	//		- load all the parent textures
	//		- throw in 'link count' to all textures
	virtual TexPtr loadCube(const char *filename[7], int id) {
		if (!filename) return TexPtr();

		//return a previously loaded texture
		int i;
		for (i = 0; i < texture.size(); i++) {
			if (texture[i] &&
				!strcasecmp(texture[i]->title.c_str(), filename[6]) &&		//if the textures are the same
				texture[i]->tex_id == id)						//and the ids are the same
			{
				return TexPtr(i);
			}
		}

		//search for an empty slot to fill
		for (i = 0; i < texture.size(); i++) {
			if (!texture[i]) {
				texture[i] = new TexLoad();
				break;
			}
		}

		//unable to comply - didnt find an empty slot
		if (i >= texture.size()) {
			i = texture.size();
			texture.push_back(new TexLoad());
		}

		//todo - cache here - not doing so will proly cause errors
		texture[i]->title = filename[6];
		texture[i]->tex_id = id;
		bool active = texture[i]->tex.loadCubeMap(
			filename[0], filename[1], filename[2],
			filename[3], filename[4], filename[5]);

		//con.print("texture ID: %d\n", texture[i].tex.texture_id);

		//and return our goods
		return TexPtr(i);
	}

	virtual TexLoad *getObj(TexPtr tp) {
		if (tp.i < 0 || tp.i >= texture.size()) return NULL;
		return texture[tp.i];
	}
	
	//aux function:

	void generateBumpMap(TexLoad *tex) {
		unsigned char *data = NULL;
		try {
			int x, y;
			unsigned char *p;
			
			//grab the texture width and height
			glBindTexture(GL_TEXTURE_2D, tex->tex.getTextureID());//	tex->tex.bind();

			_vec<2,GLint> size;
			glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &size.x);
			glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT, &size.y);
			if (!size) throw Exception() << "unable to read parent tex dim";

			data = (unsigned char *)MEMALLOCNAMED(4 * size.x * size.y, "bumpmap texture data");
			if (!data) throw Exception() << "failed to alloc temp buffer";
			
			glBindTexture(GL_TEXTURE_2D, tex->tex.getTextureID());//	tex->tex.bind();
			glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);

			//first - greyscale
			//while we're at it - calculate the transparent flags
			//0.3r + 0.5g + 0.2b => b

			tex->has_transparent = false;
			tex->has_semi_transparent = false;

			vec3d grey(.3, .6, .1);
			p = data;
			y = size.x * size.y;
			for (x = 0; x < y; x++, p += 4) {
				if (p[3] < 128) {	//if we have transparency
					p[2] = 0;
				} else {
					p[2] = (unsigned char)( dot((vec3d)*(vec3uc*)p, grey) );
				}

				//todo - tolerance?
				if (p[3] < 0xFF) {
					tex->has_transparent = true;
					if (p[3] > 0) {
						tex->has_semi_transparent = true;
					}
				}
			}

	#define BUMP_DEFAULT_EXTRUSION_SQUARED		(double)(1<<8)	//16^2

			//b[+1,0]/2 + 0.5 - b[-1,0]/2 => r, b[0,+1]/2 + 0.5 - b[0,-1]/2 => g, 0.5 => b
			p = data;
			for (y = 0; y < size.y; y++) {
				for (x = 0; x < size.x; x++) {
					double dx = (double)data[((((x + 1) % size.x) + y * size.x)<<2)+2]
							  - (double)data[((((x + size.x - 1) % size.x) + y * size.x)<<2)+2];
					double dy = (double)data[((x + ((y + 1) % size.y) * size.x)<<2)+2]
							  - (double)data[((x + ((y + size.y - 1) % size.y) * size.x)<<2)+2];
					double ilen = 0.5 / sqrt(dx*dx + dy*dy + BUMP_DEFAULT_EXTRUSION_SQUARED);

					dx = dx * ilen + 0.5;
					dy = dy * ilen + 0.5;

					p[0] = (unsigned char)(dx * 255.0);
					p[1] = (unsigned char)(dy * 255.0);

					p += 4;
				}
			}

			//normalize for 'b'

			p = data;
			y = size.x * size.y;
			for (x = 0; x < y; x++) {
				double dx = (double)(p[0] / 255.0 - 0.5) * 2.0;
				double dy = (double)(p[1] / 255.0 - 0.5) * 2.0;

				double dz = 1.0 - dx*dx - dy*dy;
				if (dz < 0.0) dz = 0.0;
				if (dz > 1.0) dz = 1.0;
				dz = -sqrt(dz);										//negative thanks to our messed up ONB
				p[2] = (unsigned char)((dz * 0.5 + 0.5) * 255.0);

				p += 4;
			}

		//upload and cache as bumpmap

			tex->bumpmap_active = tex->bumpmap.createEmpty(VEC2ELEM(size), GL_RGBA, GL_UNSIGNED_BYTE);
			tex->bumpmap_active |= tex->bumpmap.build2DMipmaps(VEC2ELEM(size), GL_RGBA, GL_UNSIGNED_BYTE, data);
		} catch (Exception &t) {
			con << "generateBumpMap failed: " << t << "\n";
		}
		if (data) memFree(data);
	}

};
static TexSys g_texSys;
ITexSys &texSys = g_texSys;
