/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
/*

new desired structure ...

replace the tag grid with an arbitrary-rectangle structure (arbitrary polygon would be tough with camera tracking)

multiple layer / multiple types
	grid layers
	vector layers
	
add to each layer
	blend options
	or should these be added to each key?
	how about per-tile coloring, like the landscape engine allows?  this would really inflate our file sizes ...
	might as well just put a texture on it all, and paint to it or something ...
	how about a paintable texture layer?  almost seems out of our scope ...
	
	how about an option for a layer to be a lightmap layer?  then whatever textures are added to it could
	contribute dierctly to the ambient lighting in a level?
	and from there, what about a radiosity scattering algorithm or something like that
	ooh, the many possibilities ...
	
interpolation between tiles:
	- give a tile a "blendtex"	- which is only used for its greyscale [->alpha] value
	- this tex will be divided into a 3x3.  the middle will assume to be opaque and discarded
	- then any time a tex is bordering another tex (without its own blendtex - precedence will need to be established - most likely by index (make this sortable in editor))
		we will overlay some borders ontop of the other tex by blending this tex with the blendtex (scale the alphas of our tex and of the blendtex)
*/

#include <sstream>
#include <fstream>
#include <string>
#include <vector>
#include <algorithm>

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "system/gl.h"
#include "system/system.h"
#include "system/memory.h"

#include "base/filebuffer.h"
#include "base/base.h"

#include "inspiration/game/program.h"
#include "inspiration/game/main.h"

#include "math/vec.h"

#include "inspiration/world.h"
#include "inspiration/entity.h"
#include "inspiration/texloader.h"
#include "inspiration/main.h"
#include "inspiration/client.h"
#include "inspiration/draw.h"
#include "inspiration/spawnobj.h"
#include "inspiration/shader.h"
#include "inspiration/maptex.h"
#include "inspiration/animation.h"	//for drawing sprites
#include "inspiration/filesys.h"
#include "inspiration/strcasecmp.h"
#include "inspiration/lua.h"

#include "editor/editor.h"

using namespace std;

//this is actually per-mod...
std::vector<std::string> mapTypes;

// ****************************************** MAP LOADER

struct ci_tagtype : binary_function<MapTagType*, MapTagType*, bool> {
	bool operator() (const MapTagType *a, const MapTagType *b) { return string(a->name) < string(b->name); }
};

/*
struct ci_tagkey : binary_function<MapTagKey, MapTagKey, bool> {
	bool operator() (const MapTagKey &a, const MapTagKey &b) { return a.id < b.id; }
};
*/


class LuaStringIterCharCallback {
public:
	virtual bool callback(unsigned char i) = 0;
};

//big big hack for reading data arrays from strings
class LuaStringIterChar {
public:
	LuaStringIterChar(lua_State *L, const char *field, LuaStringIterCharCallback &callback) {
		int tableIndex = lua_gettop(L);
		lua_getfield(L, tableIndex, field);
		if (!lua_isstring(L, -1)) {
			lua_pop(L,1);
			throw Exception() << "failed to find field " << field;
		}
		const char *str = lua_tostring(L, -1);
		int len = strlen(str);
		for (const char *p = str; p < (str + len); p++) {
			if (!callback.callback((unsigned char)*p)) break;
		}
		lua_pop(L, 1);
	}
};

static int panic(lua_State *L) {
	con.print("Lua Paniced!\n");
	con.print("%s\n", lua_tostring(L, -1));
	con.print("%s\n", lua_tostring(L, -2));
	return 0;
}

static void LoadWorld_Map(lua_State *L, World *world) {
	int top = lua_gettop(L);
	try {
		LuaTable map;
		LuaTable(L, LUA_GLOBALSINDEX).req("map") >> map;

		vec3i newMapSize;
		map.req("layers") >> newMapSize.z;
		map.req("size") >> *(vec2i*)&newMapSize;

		unsigned char *newMapDataIndices = new unsigned char[newMapSize.x * newMapSize.y * newMapSize.z];
		memset(newMapDataIndices, 0, sizeof(newMapDataIndices[0]) * newMapSize.x * newMapSize.y * newMapSize.z);
		
		class ReadDataCallback : public LuaStringIterCharCallback {
		public:
			int index;
			unsigned char *dst;
			int size;
			ReadDataCallback(unsigned char *_dst, int _size) : dst(_dst), size(_size), index(0) {}
			
			virtual bool callback(unsigned char ch) {
				if (ch < ' ') return true;	//true means keep going
				ch -= ' ';
				assert(ch != '[' && ch != ']');		//we're doing away with these.
				dst[index] = ch;
				index++;
				return index < size;
			}
		};
		ReadDataCallback _r(newMapDataIndices, newMapSize.x * newMapSize.y * newMapSize.z);
		LuaStringIterChar _l(L, "data", _r);

		//how could we read tables?
		//LuaTable keyTable;
		//map.req("keys") >> keyTable
		//then how about that iterator ...
		
		LuaTable keys;
		map.req("keys") >> keys;
		for (keys.startNext(); keys.next();) {

			//-2 is the key is a single-char string for our mapkey index
			//-1 is the value - a table of mapkey contents
			if (lua_type(L, -2) != LUA_TSTRING) throw Exception() << "expected the key to be a string";
			const char *key = lua_tostring(L, -2);
			if (!key || strlen(key) != 1) throw Exception() << "found a key that wasn't one character long: " << key;
			int key_id = (unsigned char)key[0] - ' ';

			MapKey *k = (*world->keys)[key_id];
			if (!k) {
				k = (*world->keys)[key_id] = new MapKey();
				if (!k) throw Exception() << "failed to allocate map key: " << key;
			}

			LuaTable mapkey(L);
			mapkey.rw("tex") >> k->texname;
			mapkey.rw("shader") >> k->shaderName;
			mapkey.rw("postFBO") >> k->postFBO;
			mapkey.rw("bbox") >> *(vec4f*)&k->bbox;

			string type;
			if (mapkey.rw("type") >> type) {
				if (type[0] == '#') {
					k->type = atoi(type.c_str() + 1);
				} else {
					k->type = -1;
					for (int i = 0; i < mapTypes.size(); i++) {
						if (!strcasecmp(mapTypes[i].c_str(), type.c_str())) {
							k->type = i;
							break;
						}
					}
					if (k->type == -1) {
						con << "got an unknown maptype: " << type << endl;
						k->type = 0;
					}
				}
			}

			mapkey.rw("z") >> k->z;
			mapkey.rw("friction") >> *(vec4f*)k->friction;
			mapkey.rw("bounce") >> k->restitution;

			_vec<6,float> texgen;
			texgen[0] = texgen[4] = 1;
			mapkey.rw("texgen") >> texgen;
			k->texu.set( (float*)texgen );
			k->texv.set( (float*)texgen + 3);
			
			LuaTable sides;
			if (mapkey.rw("sides") >> sides) {
				k->sides = 0;
				for (int i = 1; i <= 4; i++) {
					char ss[1];
					fixedstring<sizeof(ss)> _f(ss);
					
					if (!(sides.rw(i) >> _f)) break;
					
					char ch = ss[0];
					if (ch == 'u' || ch == 'U') k->sides |= bitflag(SIDE_UP);
					if (ch == 'd' || ch == 'D') k->sides |= bitflag(SIDE_DOWN);
					if (ch == 'l' || ch == 'L') k->sides |= bitflag(SIDE_LEFT);
					if (ch == 'r' || ch == 'R') k->sides |= bitflag(SIDE_RIGHT);
				}
			}

			LuaTable sprite;
			if (mapkey.rw("sprite") >> sprite) {
				try {
					sprite.req(1) >> k->groupname;
					sprite.req(2) >> k->seqname;
				} catch (Exception &t) { throw Exception() << "sprite: " << t; }
			}

			LuaTable planes;
			if (mapkey.rw("planes") >> planes) {
				try {
					for (int i = 0; i < numberof(k->planes); i++) {
						planes.req(i+1) >> k->planes[i];
					}
				} catch (Exception &t) { throw Exception() << "planes: " << t; }
			}

			mapkey.rw("pull") >> k->pull;
			mapkey.rw("borderTex") >> k->borderTexName;
			mapkey.rw("borderThickness") >> k->borderThickness;
			mapkey.rw("borderUVThickness") >> k->borderUVThickness;
		}
		
		//now here
		//store the mapData as keys
		int newMapVolume = newMapSize.x * newMapSize.y * newMapSize.z;
		constMapKeyPtr *newMapData = new constMapKeyPtr[newMapVolume];
		for (int i = 0; i < newMapVolume; i++) {
			newMapData[i] = world->keys->safeGet(newMapDataIndices[i]);
		}
		world->resizeMap(newMapSize, newMapData);
		
		delete[] newMapDataIndices;
		delete[] newMapData;
		
	} catch (Exception &t) {
		throw Exception() << "\nLoadWorld_Map() got error: " << t;
	}
	assert(top == lua_gettop(L));
}

static void LoadWorld_Tags(lua_State *L, World *world) {
	int top = lua_gettop(L);
	try {
		LuaTable tags;
		LuaTable(L, LUA_GLOBALSINDEX).req("tags") >> tags;

		vec2i tagTileSize;
		tags.req("size") >> tagTileSize;

		//round up.  this depends on newMapSize.xy, which has already been grabbed in the method called before this one.
		vec2i tagSize;
		tagSize.x = world->getMapSize().x / tagTileSize.x  + !!(world->getMapSize().x % tagTileSize.x);
		tagSize.y = world->getMapSize().y / tagTileSize.y + !!(world->getMapSize().y % tagTileSize.y);

		short *tagData;
		int tagDataSize = sizeof(*tagData) * tagSize.x * tagSize.y;
		tagData = (short *)MEMALLOCNAMED(tagDataSize, "tag data");
		if (!tagData) throw Exception() << "failed to allocate tag data";
		memset(tagData, 0, tagDataSize);

		//unlike the last one, there is no condition and no offset.  just a direct copy.
		class ReadTagDataCallback : public LuaStringIterCharCallback {
		public:
			int index;
			unsigned char *dst;
			int size;
			ReadTagDataCallback(unsigned char *_dst, int _size) : dst(_dst), size(_size), index(0) {}
			virtual bool callback(unsigned char sh) {
				if (sh == '\r' || sh == '\n') return true;	//and skip a line or something?
				dst[index] = sh;
				index++;
				return index < size;
			}
		};
		ReadTagDataCallback _r((unsigned char *)tagData, tagDataSize * sizeof(short));
		try {
			LuaStringIterChar(L, "data", _r);
		} catch (Exception &t) {
			con << "read tag data failed: " << t << "\n";
		}
		
		if (world->tags->data) memFree(world->tags->data);
		world->tags->data = tagData;
		world->tags->size = tagSize;
		world->tags->tagsize = tagTileSize;
		
		//now iterate through the keys
		LuaTable keys;
		tags.req("keys") >> keys;
		for (keys.startNext(); keys.next();) {
			//-2 is key, -1 is tagname
			MapTagKey k;
			if (lua_type(L, -2) != LUA_TSTRING) throw Exception() << "expected key to be a string";
			const char *key = lua_tostring(L, -2);
			if (!key || strlen(key) < 2) throw Exception() << "got a tag key too small: " << key;
			int id = *(short *)key;
			
			const char *name = lua_tostring(L, -1);
			if (!name) throw Exception() << "got a bad value for tag key: " << key;
			k.tagTypeName = name;
			world->tagkeys->keys[id] = new MapTagKey(k);
		}
		
		//sort mapTagKeySys.keys by its key
		//(make it a map)
		//sort<vector<MapTagKey>::iterator, ci_tagkey>(mapTagKeySys.keys.begin(), mapTagKeySys.keys.end(), ci_tagkey());

		//now iterate through the types
		
		LuaTable types;
		tags.req("types") >> types;
		for (types.startNext(); types.next();) {
			if (lua_type(L, -2) != LUA_TSTRING) throw Exception() << "expected key to be a string";
			const char *name = lua_tostring(L, -2);
			if (!name) throw Exception() << "types got a non-string key";

			MapTagType *tt = new MapTagType();
			tt->name = name;

			LuaTable tagType(L);
			tagType.rw("tex") >> tt->texname;
			tagType.rw("scroll") >> tt->scroll;
			tagType.rw("scale") >> tt->scale;
			tagType.rw("grav") >> tt->grav;
			tagType.rw("light") >> tt->light;

			world->tagtypes->tags.push_back(tt);	//TODO - MAKE ME A MAP

		}
		
		//now sort the tagtypes ... by name
		sort<vector<MapTagType*>::iterator, ci_tagtype>(world->tagtypes->tags.begin(), world->tagtypes->tags.end(), ci_tagtype());

	} catch (Exception &t) {
		throw Exception() << "\nLoadWorld_Tags() got error: " << t;
	}
	assert(top == lua_gettop(L));
}

static void LoadWorld_Texs(lua_State *L, World *world) {
	//this wasnt reset before.  reset also clears textures.  will this fix a bug for us?

	int top = lua_gettop(L);
	try {
		LuaTable texs;
		LuaTable(L, LUA_GLOBALSINDEX).req("texs") >> texs;
		
		for (texs.startNext(); texs.next();) {
			int it = world->texs->create();
			if (it == -1) throw Exception() << "ran out of room for map textures";
			MapTex *t = world->texs->texs[it];	//maptexarray + it;	//moving back

			if (lua_type(L, -2) != LUA_TSTRING) throw Exception() << "expected the key to be a string";
			const char *name = lua_tostring(L, -2);	//key
			if (!name) throw Exception() << "got a null key";
			t->texName = name;

			LuaTable tex(L);
			tex.rw("file") >> t->sources[0];

			vec4i area;
			if (tex.rw("area") >> area) {
				t->subtexOffset = *(vec2i*)&area[0];
				t->subtexSize = *(vec2i*)&area[2];
			}
			
			//here's an oddly placed one.  next has index [1] a string and [2] a number
			LuaTable next;
			if (tex.rw("next") >> next) {	//if we find the next table then we'd better expect its contents
				try {
					next.req(1) >> t->nextName;
					next.req(2) >> t->duration;
				} catch (Exception &t) { throw Exception() << "next: " << t; }
			}
		}
	} catch (Exception &t) {
		throw Exception() << "\nLoadWorld_Texs() got error: " << t;
	}
	assert(top == lua_gettop(L));
	
	//sort textures by name ... ?
	//thats still safe to do now -- nothing has attached itself to their indices
	world->texs->sort();
}

static void LoadWorld_Scripts(lua_State *L, World *world) {
	int top = lua_gettop(L);
	try {
		LuaTable(L, LUA_GLOBALSINDEX).rw("script") >> world->scriptList;
	} catch (Exception &t) {
		throw Exception() << "\nLoadWorld_Script() got error: " << t;
	}
	assert(top == lua_gettop(L));
}

static void LoadWorld_Ents(lua_State *L) {
	int top = lua_gettop(L);
	try {
		LuaTable ents;
		LuaTable(L, LUA_GLOBALSINDEX).req("ents") >> ents;
		
		for (ents.startNext(); ents.next();) {
			//expect a table for each value
			int subTableIndex = lua_gettop(L);

			if (!lua_istable(L, -1)) throw Exception() << "found a non-table in the ents table";
			
			//why don't we just store everything as map entries?
			//and why don't we store the classname as just that? 'classname'
			//and name as 'name'
			//then just read those back when we go to instanciate the class

			//get the spawn type
			string type;
			LuaTable ent(L);
			
			SpawnObj *spawn_id = NULL;
			if (!(spawn_id = spawnObjSys.createByName(type.c_str()))) throw Exception() << "couldn't create spawn object";

			//get all else
			bool gotClassName = false;
			for (ent.startNext(); ent.next();) {
				if (!lua_isstring(L, -2)) throw Exception() << "found a non-string key in our parameters!";
				if (!lua_isstring(L, -1)) throw Exception() << "found a non-string value in our parameters with key " << lua_tostring(L, -2);
		
				const char *keyname = lua_tostring(L, -2);
				const char *value = lua_tostring(L, -1);
				spawn_id->m[keyname] = value;
				//
				if (strcmp(keyname, "classname")) gotClassName = true;
			}
			
			if (!gotClassName) throw Exception() << "failed to find classname!";
		}
	} catch (Exception &t) {
		throw Exception() << "\nLoadWorld_Ents() got error: " << t;
	}
	assert(top == lua_gettop(L));
}

static bool ReadMapFile(const char *filename, World *world) {
	bool result = true;	//innocent until proven guilty

	FILE *file = NULL;	//gah, what kind of standard is there on stl reading
	lua_State *L = NULL;

	try {		
		file = fopen(filename, "r");
		if (!file) throw Exception() << "failed to open file";
		
		L = lua_open();
		if (!L) throw Exception() << "failed to open lua state";
		lua_atpanic(L, panic);

		luaL_openlibs(L);

		FileBuffer fb;
		if (!fb.load(filename)) throw Exception() << "failed to read the map file";
		
		const char *buffer = fb.getData();
		size_t bufferSize = fb.getSize();
		
		int err = 0;
		if (err = luaL_loadbuffer(L, buffer, bufferSize, "world")) {
			throw Exception() << "for loading world:\n" << buffer << "\n\n"
				<< " we got some errors:\n"
				<< lua_tostring(L, -1) << "\n"
				<< lua_tostring(L, -2) << "\n";
		}

		int base = lua_gettop(L);
		lua_pushliteral(L, "_TRACEBACK");
		lua_rawget(L, LUA_GLOBALSINDEX);
		lua_insert(L, base);
		if (lua_pcall(L, 0, 0, 0)) throw Exception() << "readmapfile got lua error\n" << lua_tostring(L, -1);
		lua_remove(L, base);

		//now that we have the state lets grab the vars from global, etc
		LoadWorld_Map(L, world);
		LoadWorld_Tags(L, world);
		LoadWorld_Texs(L, world);
		LoadWorld_Scripts(L, world);
		LoadWorld_Ents(L);	//already done for ya =) just about
	} catch (Exception &t) {
		con << "ReadMapFile(" << filename << ") got error: " << t << "\n";
		void lua_printbacktrace(lua_State *L);	//program.cpp
		lua_printbacktrace(L);
		result = false;
	}

	//traceback is still on the stack
	if (L) lua_close(L);
	if (file) fclose(file);
	return result;
}

static World *LoadWorld(const char *filename) {
	World *world = new World();
	try {
		if (!world) throw Exception() << "failed to allocate World object";
		
		//grab the filename if its there
		//found in GameAPI
		string result;
		if (!fileSys.search(filename, game.getMapPaths(), result)) throw Exception() << "failed to find the file";

		//reset everything in preparation
		//if we're resetting all things (like the spawn objects) then the old world data is janked
		//so why the need to preserve it?
#pragma warning TODO - put spawnObjSys in World
		spawnObjSys.newSystem();	//reset the spawn object system

		if (!ReadMapFile(result.c_str(), world)) {
			throw Exception() << "failed to read the world";
		}

		world->texs->refresh();		//refresh the maplist texturs
		world->keys->refreshTexIndices();		//align the mapkeys to the maplist textures
		world->tagtypes->refresh();			//map the texture names to the texture types
		world->tagkeys->refresh();			//go through tag keys, map all rooms to room indexes

		world->filename = result;		//store the last loaded map's name...

		//just return here unless you need any shutdown stuff
		return world;

	} catch (Exception &t) {
		con << "LoadWorld(" << filename << ") failed: " << t << "\n";
		if (world) delete world;
		return NULL;
	}

}

// ****************************************** SAVE THE WORLD

void World::saveData(ofstream &file) {
	file << "map = {\n";
	file << "\tlayers = " << backupMap->size.z << ";\n";
	vec2i v(VEC2ELEM(backupMap->size));
	file << "\tsize = " << arrayToLua(v) << ";\n";
	file << "\tdata = [[\n";

	constMapKeyPtr *c = backupMap->data;
	for (int k = 0; k < backupMap->size.z; k++) {
		for (int j = 0; j < backupMap->size.y; j++) {
			for (int i = 0; i < backupMap->size.x; i++, c++) {
				int key = 0;
				assert(*c);
				if (*c == &MapKey::def) {
					key = 0;
				} else {
					key = keys->getIndex(*c);
					assert(keys->isValid(key));
				}
				file << (char)(key+' ');
			}
			file << "\n";
		}
	}
	file << "]];\n";	//close off the data
	file << "\n";
	file << "\tkeys = {\n";
	
	MapKey defaultKey;

	for (int i = 0; i < keys->size(); i++) {
		if (!keys->isValid(i)) continue;
		MapKey *m = (*keys)[i];
		if (!m) continue;

		//add in a backslash if its a single quote
		char k[2] = {i+' ','\0'};
		file << "\t\t" << strToLuaKey(k) << " = { ";	//vec2c is my quick alloc 2-char string =)

		//texname
		//TODO: fix this!
		if (m->tex != defaultKey.tex) {
			MapTex *mt = m->tex;
			if (mt->texName.length()) file << "tex = " << strToLua(mt->texName) << "; ";
		}
		if (m->shader != defaultKey.shader) {
			const ShaderObj *so = shaderSys.getIndex(m->shader);
			if (so) {
				const string &name = so->getName();
				if (name.length()) file << "shader = " << strToLua(name) << "; ";
			}
		}
		if (m->postFBO != defaultKey.postFBO) {
			file << "postFBO = " << (m->postFBO ? "true" : "false") << "; ";
		}
		if (m->bbox != defaultKey.bbox) file << "bbox = " << arrayToLua(*(vec4f*)&m->bbox) << "; ";
		//z bias
		if (m->z != defaultKey.z) file << "z = " << m->z << "; ";

		//type
		if (m->type != defaultKey.type) {
			file << "type = ";
			if (m->type < 0 || m->type >= mapTypes.size()) {
				ostringstream s;
				s << "#" << m->type;
				file << strToLua(s.str()) << "; ";
			} else {
				file << strToLua(mapTypes[m->type]) << "; ";
			}
		}
		//friction
		//vector compare will load up all 4 elems of the array then compare all to 1.f at once.  if any differ it'll do the conditional code
		if (vec4f(m->friction) != vec4f(defaultKey.friction)) file << "friction = " << arrayToLua(m->friction, 4) << "; ";
		//restitution
		if (m->restitution != defaultKey.restitution) file << "bounce = " << m->restitution << "; ";
		//texgen
		if (m->texu != defaultKey.texu || m->texv != defaultKey.texv) {
			_vec<6,float> texgen;
			*(vec3f*)&texgen[0] = m->texu;
			*(vec3f*)&texgen[3] = m->texv;
			file << "texgen = " << arrayToLua(texgen) << "; ";
		}
		//sides
		if (m->sides != defaultKey.sides) {
			vector<string> sides;	//i know, i know, very inefficient
			if (m->sides & bitflag(SIDE_UP)) sides.push_back("u");
			if (m->sides & bitflag(SIDE_DOWN)) sides.push_back("d");
			if (m->sides & bitflag(SIDE_LEFT)) sides.push_back("l");
			if (m->sides & bitflag(SIDE_RIGHT)) sides.push_back("r");
			file << "sides = " << arrayToLua(sides) << "; ";
		}
		//texname
		if (m->anim.group != defaultKey.anim.group) {
			string groupname;
			string seqname;
			if (animSys.getGroupName(m->anim.group, groupname) &&
				animSys.getSeqName(m->anim.group, m->anim.seq, seqname))
			{
				_vec<2,string> v(strToLua(groupname), strToLua(seqname));
				file << "sprite = " << arrayToLua(v) << "; ";
			}
		}
		//planes
		if ((_vec<12,float>&)m->planes != (_vec<12,float>&)defaultKey.planes) {
			file << "planes = {";
			const char *comma = "";
			for (int i = 0; i < numberof(m->planes); i++) {
				file << comma << arrayToLua(m->planes[i]);
				comma = ", ";
			}
			file << "}; ";
		}
		if (m->pull != defaultKey.pull) file << "pull = " << m->pull << "; ";
		if (m->borderTex != defaultKey.borderTex) {
			MapTex *mt = m->borderTex;
			if (mt->texName.length()) file << "borderTex = " << strToLua(mt->texName) << "; ";
		}
		if (m->borderThickness != defaultKey.borderThickness) file << "borderThickness = " << m->borderThickness << "; ";
		if (m->borderUVThickness != defaultKey.borderUVThickness) file << "borderUVThickness = " << m->borderUVThickness << "; ";
		file << "};\n";
	}
	file << "\t};\n";
	file << "}\n";
	file << "\n";
}

//[tags]
//[tagkeys]
//[tagtypes]
static void SaveWorld_Tags(World *world, ofstream &file) {
	file << "tags = {\n";
	file << "\tsize = " << arrayToLua(world->tags->tagsize) << ";\n";
	file << "\tdata = [[\n";

	char *c = (char *)world->tags->data;
	for (int j = 0; j < world->tags->size.y; j++) {
		for (int i = 0; i < world->tags->size.x; i++,  c += 2) {
			file << c[0] << c[1];
		}
		file << "\n";
	}
	file << "]];\n";
	file << "\tkeys = {\n";
	
	for (map<short, MapTagKey*>::iterator it = world->tagkeys->keys.begin(); it != world->tagkeys->keys.end(); it++) {
		short id = it->first;
		MapTagType *tt = it->second->tagType;
		if (tt) {
			file << "\t\t" << strToLuaKey(_vec<3,char>(((char *)&id)[0], ((char *)&id)[1], 0))
				<< " = " << strToLua(tt->name) << ";\n";
		}
	}
	file << "\t};\n";

	//default object for comparing (delta encode or however you want to look at it)
	MapTagType dt;

	file << "\ttypes = {\n";
	for (int i = 0; i < world->tagtypes->tags.size(); i++) {
		MapTagType *tt = world->tagtypes->tags[i];
		file << "\t\t" << strToLuaKey(tt->name) << " = { ";
		if (tt->tex != dt.tex) file << "tex = " << strToLua(tt->tex->texName) << "; ";
		if (tt->scroll != dt.scroll) file << "scroll = " << arrayToLua(tt->scroll) << "; ";
		if (tt->scale != dt.scale) file << "scale = " << arrayToLua(tt->scale) << "; ";
		if (tt->grav != dt.grav) file << "grav = " << arrayToLua(tt->grav) << "; ";
		if (tt->light != dt.light) file << "light = " << arrayToLua(tt->light) << "; ";
		file << "};\n";
	}
	file << "\t};\n";
	file << "}\n";
	file << "\n";
}

static void SaveWorld_Texs(World *world, ofstream &file) {
	file << "texs = {\n";

	for (int i = 0; i < world->texs->texs.size(); i++) {
		MapTex *m = world->texs->texs[i];	//maptexarray;	//moving home
		file << "\t" << strToLuaKey(m->texName) << " = { ";
		file << "file = " << strToLua(m->sources[0]) << "; ";
		if (m->subtexSize.x && m->subtexSize.y) {
			vec4i v(VEC2ELEM(m->subtexOffset), VEC2ELEM(m->subtexSize));
			file << "area = " << arrayToLua(v) << "; ";
		}

//setup isnt working
//but
//it loads correctly ...
		if (world->texs->isValid(m->nextIndex)) {
			MapTex *next = world->texs->texs[m->nextIndex];
			if (next) {
				file << "next = {" << strToLua(next->texName) << ", " << m->duration << "}; ";
			}
		}

		//TODO - add cube support too... later...
		//and animation too...
		file << "};\n";
	}
	file << "}\n";
	file << "\n";
}

//write out the script filename
static void SaveWorld_Script(World *world, ofstream &file) {
	int size = 0;

	file << "script = {\n";
	for (int i = 0; i < world->scriptList.size(); i++) {
		file << "\t" << strToLua(world->scriptList[i]) << ";\n";
	}
	file << "}\n";
	file << "\n";
}


static void SaveWorld_Ents(World *world, ofstream &file) {
	file << "ents = {\n";

	char buffer[256];

	for (int i = 0; i < spawnObjSys.count(); i++) {
		SpawnObj *s = spawnObjSys[i];
		if (!s) continue;
		
		file << "\t{\n";

		//write it out backwards so we can read it in forwards =) makes sense, eh?  it really doesn't... must be a lua order thing?
		for (map<string,string>::reverse_iterator iter = s->m.rbegin(); iter != s->m.rend(); iter++) {
			//these days you might as well allow direct access to the string buffer ...
			//its not like you need to encode/decode them anymore...
			file << "\t\t" << strToLuaKey(iter->first) << " = " << strToLua(iter->second) << ";\n";
		}
		file << "\t};\n";
	}
	file << "}\n";
	file << "\n";
}


// ****************************************** END MAP LOADER

void CFunc_LoadWorld(int argc, char **argv) {
	if (argc < 2) {
		con.print("usage: load <filename>\n");
		return;
	}

	char *filename = argv[1];

	//load the bitmap
	if (!worldSys.loadWorld(filename)) {
		con.print("failed to load world %s\n", filename);
		return;
	}
}

void CFunc_LoadInitWorld(int argc, char **argv) {
	if (!worldSys.loadInitWorld()) {
		con << "failed to load init world " << worldSys.getInitWorld() << endl;
		return;
	}
}

void CFunc_SaveWorld(int argc, char **argv) {
	if (!worldSys.saveWorld()) {
		World *world = worldSys.getWorld();
		con << "CFunc_SaveWorld: failed to save world " << (world ? world->filename : "when the world is NULL") << endl;
		return;
	}
}

void CFunc_SaveWorldAs(int argc, char **argv) {
	if (argc < 2) {
		con.print("usage: saveas <filename>\n");
		return;
	}

	char *filename = argv[1];

	if (!worldSys.saveWorldAs(filename)) {
		con.print("failed to save world %s\n", filename);
		return;
	}

	//reset the world bsp tree *HERE*
}

void CFunc_LastWorld(int argc, char **argv) {
	World *world = worldSys.getWorld();
	if (world && world->filename.length()) {
		con << "last loaded map: " << world->filename << "\n";
	} else {
		con << "last loaded map has no filename!\n";
	}
}

void CFunc_SaveLastWorld(int argc, char **argv) {
	World *world = worldSys.getWorld();
	con << "last loaded map: " << world->filename << "\n";
	if (!worldSys.saveWorld()) {
		con << "failed to save world " << world->filename << "\n";
	}
	return;
}

static int drawBackground = 1;
static int drawForeground = 1;
static float backgroundDepth = -1.1f;

static float worldSpriteZScale = -1.5f;
static float worldSpriteZBias = -1.f;
static float worldSpriteZLayerScale = -1.f;

static float worldTexZBias = -1.6f;	//so its just deeper than the sprites
static float worldTexZScale = -1.f;
static float worldTexZLayerScale = -1.f;

static CVarw *local_cvars[] = {
	new CVarw(0,					"load",					CFunc_LoadWorld),
	new CVarw(0,					"loadinit",				CFunc_LoadInitWorld),
	new CVarw(0,					"save",					CFunc_SaveWorld),
	new CVarw(0,					"saveas",				CFunc_SaveWorldAs),
	new CVarw(0,					"lastmap",				CFunc_LastWorld),
	new CVarw(0,					"savelast",				CFunc_SaveLastWorld),
	new CVarw(&drawBackground,		"drawBackground"),
	new CVarw(&drawForeground,		"drawForeground"),
	new CVarw(&backgroundDepth,		"backgroundDepth"),
	
	//for sprites bound to map types
	new CVarw(&worldSpriteZScale,	"worldSpriteZScale"),
	new CVarw(&worldSpriteZBias,	"worldSpriteZBias"),
	new CVarw(&worldSpriteZLayerScale, "worldSpriteZLayerScale"),
	
	//for textures associated with map types
	new CVarw(&worldTexZBias,		"worldTexZBias"),
	new CVarw(&worldTexZScale,		"worldTexZScale"),
	new CVarw(&worldTexZLayerScale,	"worldTexZLayerScale"),
};


//////////// map tag keys:

MapTagKey MapTagKey::def;

void MapTagKeySet::refresh() {

	//TODO here - for all tags that reference uncreaed tagkeys, create a new one
	int i;
	for (i = 0; i < world->tags->size.x * world->tags->size.y; i++) {
		//make sure we all have printable data:
		if (world->tags->data[i] < MapTagKey::minRange) {
			world->tags->data[i] = MapTagKey::minRange;
		}
		//make sure we have an associated tagkey
		map<short, MapTagKey*>::iterator it = keys.find(world->tags->data[i]);
		//if we didn't find one then create a new default one
		if (it == keys.end()) {
			MapTagKey tk;
			tk.tagType = NULL;
			keys[world->tags->data[i]] = new MapTagKey(tk);
		}
	}
	
	for (map<short,MapTagKey*>::iterator it = keys.begin(); it != keys.end(); it++) {
		MapTagKey *tk = it->second;
		tk->tagType = NULL;
		int tagType = world->tagtypes->findByName(tk->tagTypeName.c_str());
		if (tagType >= 0 && tagType < world->tagtypes->tags.size()) {
			tk->tagType = world->tagtypes->tags[tagType];
		}
	}
}

int MapTagKeySet::create() {
	//find the first key that's not used
	for (int i = MapTagKey::minRange; i < MapTagKey::maxRange; i++) {
		map<short,MapTagKey*>::iterator it = keys.find(i);
		if (it == keys.end()) {
			MapTagKey tk;
			tk.tagType = NULL;
			keys[i] = new MapTagKey(tk);
			return i;
		}
	}
	return -1;	//failed
}

int MapTagKeySet::remove(int tagKeyID) {
	//eliminate it from the tag_data (instead of stepping down [which could be painful] just reset the value)
	for (int k = 0; k < world->tags->size.x * world->tags->size.y; k++) {
		if (world->tags->data[k] == tagKeyID) world->tags->data[k] = MapTagKey::minRange;
	}
	//first find the tagkey that matches
	map<short,MapTagKey*>::iterator it = keys.find(tagKeyID);
	if (it == keys.end()) return -1;
	
	int returnKeyID = -1;
	//hunt forward one
	map<short,MapTagKey*>::iterator next = it;	next++;
	if (next != keys.end()) {
		returnKeyID = next->first;
	//if that doesn't work then hunt back one
	} else if (it != keys.begin()) {	//make sure we can trace back one
		next = it;	next--;
		returnKeyID = next->first;
	}
	
	//now we wanna (1) get the next of the iterator (2) erase the current and (3) return the next
	delete it->second;
	keys.erase(it);
	
	return returnKeyID;
}

//////////// map tag types:

void MapTagTypeSet::reset() {
	for (int i = 0; i < tags.size(); i++) delete tags[i];
	tags.resize(0);
}

//map texIndex from MapTagType->texname
void MapTagTypeSet::refresh() {
	for (int i = 0; i < tags.size(); i++) {
		tags[i]->tex = world->texs->safeTexForName(tags[i]->texname.c_str());
	}
}

int MapTagTypeSet::findByName(const char *name) {
	for (int i = 0; i < tags.size(); i++) {
		if (!strcmp(name, tags[i]->name.c_str())) return i;
	}
	return -1;
}

int MapTagTypeSet::getIndex(MapTagType *tt) {
	for (int i = 0; i < tags.size(); i++) {
		if (tags[i] == tt) return i;
	}
	return -1;
}

short MapTagTypeSet::create() {
	MapTagType *tt = new MapTagType();
	tt->name = "tagtype";
	tags.push_back(tt);
	return tags.size()-1;
}

short MapTagTypeSet::remove(short tagtype) {
	delete tags[tagtype];
	for (int j = tagtype+1; j < tags.size(); j++) {
		tags[j-1] = tags[j];
	}
	tags.resize(tags.size()-1);
	if (tagtype >= tags.size()) tagtype--;
	return tagtype;
}

MapTagType MapTagType::def;


//////////// mapkeys:

MapKey MapKey::def;

void MapKeySet::reset() {
	memset(mapkey, 0, sizeof(mapkey));
}

void MapKeySet::shutdown() {
	//delete all the old map keys before reading new ones
	for (int i = 0; i < numberof(mapkey); i++) {
		if (mapkey[i]) delete mapkey[i];
	}
	reset();
}

//safeget is only used for writing keys
const MapKey *MapKeySet::safeGet(int index) {
	if (!isValid(index)) return &MapKey::def;
	if (!mapkey[index]) return &MapKey::def;
	return mapkey[index];
}


int MapKeySet::getIndex(const MapKey *k) {
	for (int i = 1; i < numberof(mapkey); i++) {
		if (!isValid(i)) continue;	//skip the []'s
		if (mapkey[i] == k) return i;
	}
	return 0;
}


bool MapKeySet::isValid(int key) {
	return
		key >= 0 &&
		key < numberof(mapkey) &&
		key != '[' - ' ' &&
		key != ']' - ' ';
}

int MapKeySet::create() {
	//dont create key zero - thats the empty key (background)
	for (int i = 1; i < numberof(mapkey); i++) {
		if (i == '['-' ' || i == ']'-' ') continue;	//skip certain keys that don't encode well
		if (!mapkey[i]) {
			mapkey[i] = new MapKey();
			if (mapkey[i]) return i;
		}
	}
	return 0;
}

int MapKeySet::remove(int key) {
	if (key < 0 || key >= numberof(mapkey)) return 0;
	if (!mapkey[key]) return 0;

	//remove it from the map:
	int i;
	for (i = 0; i < world->map->size.x * world->map->size.y * world->map->size.z; i++) {
		if (world->map->data[i] == mapkey[key]) world->map->data[i] = &MapKey::def;
	}
	//get the next key
	int nextkey = -1;
	for (i = key+1; i != key; i = (i + 1) % numberof(mapkey)) {
		if (!isValid(key)) continue;
		if (mapkey[i]) {
			nextkey = i;
			break;
		}
	}
	
	delete mapkey[key];
	mapkey[key] = NULL;

	return nextkey;
}

MapKey *&MapKeySet::get(int index) {
	assert(isValid(index));
	return mapkey[index];
}

//align to specific worldtexlist_t's texture image
void MapKeySet::refreshTexIndices() {
	int i;
	MapKey **m;
	
	//worldtexlist_t *t;
//	MapTex *t;
	
	//init tile textures
	for (i = 0, m = mapkey; i < numberof(mapkey); i++, m++) {
		if (!(*m)) continue;

		//align texture
		(*m)->tex = world->texs->safeTexForName( (*m)->texname.c_str() );
		(*m)->borderTex = world->texs->safeTexForName( (*m)->borderTexName.c_str() );
	}
}

void MapKeySet::sessionInit() {
	int i;
	MapKey **_m;
	for (i = 0, _m = mapkey; i < numberof(mapkey); i++, _m++) {
		MapKey *m = *_m;
		if (!m) continue;
		
		if (m->groupname.length()) {
//con << "refreshing group " << m->groupname << "\n";
			m->anim.setGroup(m->groupname.c_str());
			m->anim.setSeq(m->seqname.c_str());
			m->anim.setIdleSeq(m->seqname.c_str());
//con << " refreshed group " << m->groupname << "\n";
//con << " got back group " << m->anim.group << "\n";
		}
		
		if (m->shaderName.length()) {
			m->shader = shaderSys.getIndexForName(m->shaderName.c_str());
con << "setting shader " << m->shaderName << " to index " << m->shader << "\n";
		}
	}
}

void MapKeySet::updateAnim() {
	int i;
	MapKey **m;
	
	//init tile textures
	for (i = 0, m = mapkey; i < numberof(mapkey); i++, m++) {
		if (!(*m)) continue;
		if ( (*m)->anim.group > 0) {
			//cycle the map animation  . . . 
			if ((*m)->anim.frameNextTime < game_thistime) {
				(*m)->anim.incFrame();
			}
		}
	}
}

void MapKeySet::removeTexIndex(int index) {
	assert(world->texs->isValid(index));
	MapTex *mt = world->texs->texs[index];
	
	//get our replacement key ready
	MapTex *replaceMT = NULL;
	if (index > 0) {
		replaceMT = world->texs->texs[index-1];
	} else {	//index == 0, make sure another exists
		if (world->texs->texs.size() > 1) {
			replaceMT = world->texs->texs[1];
		}
	}
	
	//for all keys
	//if its tex is the maptexarray_count and >= 0
	//then push it down one.
	for (int i = 0; i < numberof(mapkey); i++) {
		MapKey *mk = mapkey[i];
		if (!mk) continue;
		if (mk->tex == mt) mk->tex = replaceMT;
	}
}

//for in-game updating (editor)
void MapKey::setGroup(const char *group) {
	groupname = group;
	anim.setGroup(groupname.c_str());
	anim.setSeq(seqname.c_str());
	anim.setIdleSeq(seqname.c_str());
}

void MapKey::setSeq(const char *seq) {
	seqname = seq;
	anim.setGroup(groupname.c_str());
	anim.setSeq(seqname.c_str());
	anim.setIdleSeq(seqname.c_str());
}

void MapKey::setShader(const char *_shader) {
	shaderName = _shader;
	//and the index
	shader = shaderSys.getIndexForName(shaderName.c_str());
}


//////////// map tags:

MapTagSet::~MapTagSet() {
	if (data) memFree(data);
}

//depends upon map_width, map_height
bool MapTagSet::setTagSize(vec2i newTagTileSize) {
	if (newTagTileSize.x < 1) newTagTileSize.x = 1;
	if (newTagTileSize.y < 1) newTagTileSize.y = 1;

	//maybe i should make per-element operators ...
	vec2i newTagSize(
		world->map->size.x / newTagTileSize.x + !!(world->map->size.x % newTagTileSize.x),
		world->map->size.y / newTagTileSize.y + !!(world->map->size.y % newTagTileSize.y));

	short *newTagData = (short *)MEMALLOCNAMED(newTagSize.x * newTagSize.y * sizeof(newTagData[0]), "tag data");
	if (!newTagData) return false;
	int defaultTagData = (int)MapTagKey::minRange | ((int)MapTagKey::minRange << 16);	//two shorts of the default ID
	memset(newTagData, defaultTagData , newTagSize.x * newTagSize.y * sizeof(newTagData[0]));

	//copy over all old data
	for (int j = 0; j < size.y; j++) {
		if (j >= newTagSize.y) continue;
		for (int i = 0; i < size.x; i++) {
			if (i >= newTagSize.x) continue;
			newTagData[i + j * newTagSize.x] = (*world->tags)(i,j);
		}
	}

	if (data) memFree(data);
	data = newTagData;
	size = newTagSize;
	tagsize = newTagTileSize;
	return true;
}

short MapTagSet::maploc(int x, int y) const {
	//convert from map space to tag space
	vec2i v(x / tagsize.x, y / tagsize.y);
	if (x < 0) v.x--;
	if (y < 0) v.y--;

	//bounds test
	if (v.x < 0 || v.x >= size.x || v.y < 0 || v.y >= size.y) return 0;	//just oob
		
	//return the tag element
	return data[v.x + v.y * size.x];
}

//////////// map

Map::Map(World *world_) : world(world_), data(NULL) {}
Map::~Map() { delete[] data; }

//dont do this if we have a zero tag size ... i.e. off the bat
bool Map::resize(vec3i newSize, constMapKeyPtr *srcData) {
//	if (width == map_width && height == map_height) return true;

	int dataSize = newSize.x * newSize.y * newSize.z;
	constMapKeyPtr *newMapData = new constMapKeyPtr[dataSize];
	for (int i = 0; i < dataSize; i++) newMapData[i] = &MapKey::def;

	//copy over all old data
	if (srcData) {
		memcpy(newMapData, srcData, sizeof(newMapData[0]) * newSize.x * newSize.y * newSize.z);
	} else {
		for (int k = 0; k < size.z; k++) {
			if (k >= newSize.z) continue;
			for (int j = 0; j < size.y; j++) {
				if (j >= newSize.y) continue;
				for (int i = 0; i < size.x; i++) {
					if (i >= newSize.x) continue;
					newMapData[i + newSize.x * (j + newSize.y * k)] = data[i + size.x * (j + size.y * k)];
				}
			}
		}
	}

	delete[] data;
	data = newMapData;
	size = newSize;
	
	//try to resize the tag size ...
	world->tags->setTagSize(world->tags->tagsize);
	
	return true;
}


void Map::displayLayer(int layer, const box2f &viewarea) {	
	if (!data) return;
	
	box2i bbox((vec2i)viewarea.min, (vec2i)viewarea.max);
	bbox.min -= 1;
	bbox.max += 1;

	//we have a problem: when drawing sprites on the map's surface, they disappear too early
	//so im gonna put a hack here for the time being
	//to more accurately fix this, it would take a lot of work
	//to quickly fix this, just make sure the tilesize by which im stretching this is as big as the biggest image
	//in theory what should be done is the sprites on tiles should be rendered in the same loop that the entity sprites are
	//this would fix the culling issue and allow for better z-sorting
	bbox.min -= 3;
	bbox.max += 3;
	
	box2i worldBox(vec2i(), vec2i(size.x-1,size.y-1));	//inclusive
	if (!worldBox.touchesE(bbox)) return;
	worldBox.clampInPlace(bbox);

	int rowStep = size.x - (bbox.max.x - bbox.min.x + 1);
	
	constMapKeyPtr *c = data + bbox.min.x + size.x * (bbox.min.y + size.y * layer);
	for (int y = bbox.min.y; y <= bbox.max.y; y++, c += rowStep) {
		for (int x = bbox.min.x; x <= bbox.max.x; x++, c++) {
			const MapKey *k = *c;
			if (k == &MapKey::def) continue;
			
			//here - apply any sprites that we're supposed to . . .
			if (k->anim.group > 0) {
				float vx, vy;
				
				//todo - scale x+[0,1] by y+[0,1] - then translate it - then floor() it
				//then apply sprites at all points
				
				vx = x + 0.5f;
				vy = y + 0.5f;
				float vz = k->z + worldSpriteZBias + worldSpriteZScale * vy + worldSpriteZLayerScale * layer;

				animSys.render(
					vx, vy, vz,
					0, 1, 1,
					vec4uc(255,255,255,255),
					0,
					getBlend(BLEND_DECAL),
					false,		//post-fbo
					true,		//use lighting
					false,		//flip
					k->anim.getAbsFrame());
			}
			
			const TexLoad *tex = k->tex ? texSys.getObj(k->tex->getCurrentImage()) : 0;
			const TexLoad *borderTex = k->borderTex ? texSys.getObj(k->borderTex->getCurrentImage()) : 0;
			const ShaderObj *so = shaderSys.getIndex(k->shader);


/*
now for the blend texture ...
if any neighboring tile has a blend texture
(and its index is lower than ours - for ordering sake - so overlaps don't go both ways)
then draw a blended version of it over this geometry
(this can be done simply now by adding a geom at the same pos with the same depth
	but with a different texture and transparency blending set up ... or alpha blending ...
the kicker is that the foreign tile needs to also be drawn
	multitexture blended with the blendtex
	with the right 2nd unit coordinates corresponding with what corners are touching
this brings to concern greater questions of which render pathway to use
	and how to extensify it for shader use in other areas
		(i.e. fbo effects ...
			the guy with the predator invisibility ...
			the guy with the watercolor effect ...
		)
		
what sorts of things will we be doing with our renderer?

[ original bump mapping algorihm ]

- clear color buffer to the ambient light level
- calc sprite masks
	- turn off color masks, turn on alpha testing >.5, turn on depth mask
	- for each sprite
		- render to the depth buffer an alpha sillhouette of our charaacter.  don't forget to use glcolor so the alpha test works.

- accum lighting
	- for all visible lights
		- set up the texenv unit1 to render our light tangent space from the light's position
		- set up depth test equal, using the same depth index as the sprites used for sillhouettes
			this ensures only the sprite's light info goes to it
		- for all touching sprites
			- render the dot3 of the sprite tangent space and the light tangent space
				scale it by the light's color and additive blend it into the color buffer
- do render
	- for all sprites
		- render the sprite dest-mul with what light info is present in the color buffer

[ new proposed system]

- calc lightmap
	- clear lightmap to ambient
	- if we're doing bump mapping then first we'll want to apply a lightmap scalar
		- for all objects
			[alpha test the sprite regions where it will influence the light buffer]
			- draw the object (alpha tested) into the depth buffer / stencil buffer
				while zeroing what goes into the color buffer
				thus freshly painting this object's light info into the light fbo
			- then equality test the depth/stencil buffer so only drawing where the entity is goes through
			- for all lights touching this entity
				- additive blend light into the lightmap buffer, scaled by the dot3 of the light and the entity's tangent space
	- otherwise, no bump mapping matters:
		- for all lights
			- additive blend them into the lightmap buffer
- blend to sprites
	- draw all sprites, back to front
		- setup blending / alpha testing
		- apply your shader
			- the default will multiply blend with the lightmap while dest-blending with the background
				(shader can handle unit0 colormap and unit1 lightmap, GL_BLEND will handle the blend method after the shader is done)
			- custom ones can have their reign of the renderer
		- draw the sprite

gah, the flaw there is our bumpmap will influence the lightmap lighting objects behind us ... not realistic ...

the best way to go would be to make a lightmap the size of the largest sprite to be rendered
(it might as well be the size of the screen)
then 
- for each sprite
	- for all lights touching it
		- additive blend them (with tangent space if so desired) into the light fbo
	- then draw the sprite blended with the light fbo
this overall will require lots of read/writes to the fbo in one frame though ...
	i wonder if it would be faster to have more textures set aside for fbo rendering and swap them in & out ...
	or to do like 3d engines do with dyn lighting and cache them ... per-entity?
this would be as generalized as it gets, with a shader for blending into the light buffer, and a shader for rendering
so how about this one?

and to avoid the light fbo, we can just burn out the tex units with as many light texs as we can hold ...

now for fbo effects (i.e. watercolor) ...

*/

			if (tex || so || borderTex) {
			
				EGeometry *g = renderSys.newGeom();
				vec2f pos(x,y);
				box2f ub = box2f(pos, pos + 1.f);	//unit, untransformed, for texgen
				box2f b(k->bbox.min + pos, k->bbox.max + pos);
				//bbox and texgen are tricky
				//i'm gonna texgen by the original bbox [0,1]^2
				//instead of the newly transformed
				//this way the bbox param is specificly for overhangs
				for (int j = 0; j < 4; j++) {
					//(b.min.x, b.min.y);	10
					//(b.max.x, b.min.y);	00
					//(b.max.x, b.max.y);	01
					//(b.min.x, b.max.y);	11
					int yofs = (j&2)>>1;
					int xofs = yofs == (j&1);
					vec2f ofs(xofs,yofs);
					g->vertex[j].set(b[xofs].x, b[yofs].y);	//gen vtx by overhang box
					
					//TODO - texcoord gen before or after pull?
					vec3f tcgen(ub[xofs].x, ub[yofs].y, 1.f);	//gen tc by unit bbox
					g->texcoord[j].set(dot(tcgen, k->texu), dot(tcgen, k->texv));
				}

				//hardcore smoothing
				int neighbors = 0;
				if (k->pull || k->borderTex) {
					int neighborIndex = 0;
					for (int v = 0; v < 3; v++) {
						int ly = y + v - 1;
						for (int u = 0; u < 3; u++, neighborIndex++) {
							int lx = x + u - 1;
							if (lx < 0 ||
								lx >= size.x ||
								ly < 0 ||
								ly >= size.y)
							{
							}
							else if (k == *(data + lx + ly * size.x))
								//&MapKey::def != *(data + lx + ly * size.x)
							{
								neighbors |= 1 << neighborIndex;
							}
						}
					}
					if (k->pull && neighbors) {
						//TODO - pull back as well?
						//bits 0, 1, 2 == 1, 0, 0
						//up
							//left
						if ((neighbors & 0007) == 0001) {
							g->vertex[1].y -= k->pull;
						}
						//bits 0, 1, 2 == 0, 0, 1
							//right
						if ((neighbors & 0007) == 0004) {
							g->vertex[0].y -= k->pull;
						}
						//down
							//left
						if ((neighbors & 0700) == 0100) {
							g->vertex[2].y += k->pull;
						}
							//right
						if ((neighbors & 0700) == 0400) {
							g->vertex[3].y += k->pull;
						}
						//left
							//up
						if ((neighbors & 0111) == 0001) {
							g->vertex[1].x -= k->pull;
						}
							//down
						if ((neighbors & 0111) == 0100) {
							g->vertex[2].x -= k->pull;
						}
						//right
							//up
						if ((neighbors & 0444) == 0004) {
							g->vertex[0].x += k->pull;
						}
						if ((neighbors & 0444) == 0400) {
							g->vertex[3].x += k->pull;
						}
					}
				}

				//the tiles are half a unit off so they don't interfere with the sprites, correct?
				//that seems like it should be a z bias thing ...
				g->depth = k->z + worldTexZBias + worldTexZScale * y + worldTexZLayerScale * layer;

				g->colormap = tex ? tex->tex.getTextureID() : 0;
				g->bumpmap = tex ? tex->bumpmap.getTextureID() : 0;
				g->color.set(255,255,255,255);
				g->shader = so;
				g->use_alpha_test = tex ? (tex->has_transparent && !tex->has_semi_transparent) : false;	//use_alpha_test if we only have empty alpha values
				g->blend = tex ? (tex->has_semi_transparent ? getBlend(BLEND_DECAL) : Blend()) : Blend();
				g->use_bump_mapping = true;
				g->use_lighting = true;
				g->postFBO = k->postFBO;
				g->dir.set(1,0);
				g->flip = 1;
				
				renderSys.geomDone(g);

				if (k->borderTex) {
					float divs[4] = {0, k->borderThickness, 1.f - k->borderThickness, 1.f};
					float texdivs[4] = {0, k->borderUVThickness, 1.f - k->borderUVThickness, 1.f};
										
					for (int i = 0; i < numberof(divs)-1; i++) {
						for (int j = 0; j < numberof(divs)-1; j++) {
							//box2f box(vec2f(divs[i], divs[j]), vec2f(divs[i+1], divs[j+1]));
							//box2f tcbox(vec2f(texdivs[i], texdivs[j]), vec2f(texdivs[i+1], texdivs[j+1]));

							//bilinear interpolation from the 4 pts to 
							//add extra geometry...
							EGeometry *edge = renderSys.newGeom();
							for (int u = 0; u < 4; u++) {
								int uyofs = (u&2)>>1;
								int uxofs = uyofs == (u&1);
								vec2f vtxCoeffs(divs[i+uxofs], divs[j+uyofs]);
								vec2f tcCoeffs(texdivs[i+uxofs], texdivs[j+uyofs]);
								edge->texcoord[u] = tcCoeffs; // no remapping for texcoords?

								edge->vertex[u] = 0.f;
								for (int v = 0; v < 4; v++) {
									int vyofs = (v&2)>>1;
									int vxofs = vyofs == (v&1);
									vec2f coeff(
										vxofs ? vtxCoeffs.x : 1.f - vtxCoeffs.x,
										vyofs ? vtxCoeffs.y : 1.f - vtxCoeffs.y);
									edge->vertex[u] += g->vertex[v] * coeff.x * coeff.y;
								}
							}
							edge->depth = g->depth;
							edge->colormap = borderTex ? borderTex->tex.getTextureID() : 0;
							edge->bumpmap = borderTex ? borderTex->bumpmap.getTextureID() : 0;
							edge->color.set(255,255,255,255);
							edge->shader = g->shader;
							edge->use_alpha_test = g->use_alpha_test;
							edge->blend = g->blend;
							edge->use_bump_mapping = g->use_bump_mapping;
							edge->postFBO = g->postFBO;
							edge->dir = g->dir;
							edge->flip = g->flip;
							renderSys.geomDone(edge);
						}
					}
				}
			}
		}
	}
}

//////////// world

World::World()
:	map(new Map(this)),
	backupMap(new Map(this)),
	keys(new MapKeySet(this)),
	tags(new MapTagSet(this)),
	tagkeys(new MapTagKeySet(this)),
	tagtypes(new MapTagTypeSet(this)),
	texs(new MapTexSet(this))
{
	//exceptions in constructors ... I hear these are potentials for memory leaks
	if (!tags->setTagSize(vec2i(30,20))) throw Exception() << "tags->setTagSize(30,20) failed";
	if (!map->resize(vec3i(1,1,1))) throw Exception() << "map->resize(1,1) failed";
	refreshBackup();
	
	/*
	EntitySystemImpl::sendChangeLevel will complain/fail unless the map has an associated filename
	so I could ... give them temp filenames until they get legit ones ...
	this might take a bit more hacking into "save" to not work 'til it gets a legit (and not tmp) filename
	but it could double as backups for maps not saved too...
	*/
}

World::~World() { 
	delete map;
	delete backupMap;
	delete tags;
	delete keys;
	delete tagkeys;
	delete tagtypes;
	delete texs;
}


bool World::saveAs(const char *saveFileName) {
	assert(saveFileName);
	ofstream file(saveFileName);
	bool result = false;
	try {
		if (!file.is_open()) throw Exception() << "unable to open the file for writing";
		saveData(file);
		SaveWorld_Tags(this, file);
		SaveWorld_Texs(this, file);
		SaveWorld_Script(this, file);
		SaveWorld_Ents(this, file);		//do ents last -- they are arbitrary sized and dont fit in with the usual loading scheme
		
		filename = saveFileName;

		result = true;	//made it!
	} catch(Exception &t) {
		con << "World::saveAs filename=" << filename << " got an error: " << t << "\n";
	}
	if (file.is_open()) file.close();	//is the condition required?
	return result;
}

bool World::save() {
	if (!filename.length()) {
		con << "World::save had no prior filename.  You need to saveAs instead" << endl;
		return false;
	}
	saveAs(filename.c_str());
	return true;
}

bool World::resizeMap(const vec3i &newSize, constMapKeyPtr *newMapData) {
	if (!map->resize(newSize, newMapData)) return false;
	refreshBackup();
	return true;
}

void World::refreshBackup() {
	backupMap->resize(map->size, map->data);
}

void World::displayLayer(int layer, const box2f &viewarea) {	
	if (!map) return;
	map->displayLayer(layer, viewarea);
}

//////////// world sys

class WorldSystemImpl {
protected:
	World *world;
	string initWorld;
public:
	WorldSystemImpl();
	virtual ~WorldSystemImpl();
	
	bool sysInit();
	void sysKill();
	bool gameInit();
	void gameKill();
	bool sessionInit();
	void sessionKill();

	void update();

	void displayBackground(MapTagType *tt, const box2f &viewarea);
	void displayLayer(int index, const box2f &viewarea);

	bool getScriptFileName(int index, char *buffer, int len);

	World *getWorld();

	bool newWorld();
	bool loadWorld(const char *filename);
	bool saveWorld();
	bool saveWorldAs(const char *filename);

	//why am I doing this? i hate redundant (one-line) getters and setters...
	void setInitWorld(const std::string &initWorld_);
	const std::string &getInitWorld();
	
	bool loadInitWorld();
};

WorldSystemImpl::WorldSystemImpl() : world(NULL) {}
WorldSystemImpl::~WorldSystemImpl() { if (world) delete world; }

bool WorldSystemImpl::sysInit() {
	world = new World();
	if (!world) return false;
	editor.setWorld(world);
	return true;
}

void WorldSystemImpl::sysKill() {
	if (world) delete world;
	world = NULL;
	editor.setWorld(NULL);
}

bool WorldSystemImpl::gameInit() {
	//reset cvars. TODO - make them map/game params
	worldSpriteZScale = -1.5f;
	worldTexZScale = -1.f;

	return loadInitWorld();
}

void WorldSystemImpl::gameKill() {
	newWorld();
}

bool WorldSystemImpl::sessionInit() {
	try {
		world->keys->sessionInit();
		return true;
	} catch (Exception &t) {
		con << "WorldSystemImpl::sessionInit(): " << t << "\n";
		delete world;
	}
	return false;
}

void WorldSystemImpl::sessionKill() {
}
	
//for animations
void WorldSystemImpl::update() {
	world->texs->update();
	world->keys->updateAnim();	//update the animations of the keys . . .
}
	
void WorldSystemImpl::displayBackground(MapTagType *tt, const box2f &viewarea) {
	if (!drawBackground) return;

	if (!tt->tex) return;
	const TexLoad *tex = texSys.getObj(tt->tex->getCurrentImage());
	if (!tex) return;

	//so 0.5f is our scroll factor
	//and tt->scroll is our scale factor
	vec2f camsize = viewarea.size();
	box2f texbox;
	texbox.min.x = viewarea.min.x * (tt->scroll.x * tt->scale.x);
	texbox.min.y = viewarea.min.y * (tt->scroll.y * tt->scale.y);
	texbox.max.x = texbox.min.x + camsize.x * tt->scale.x;
	texbox.max.y = texbox.min.y + camsize.y * tt->scale.y;

	EGeometry *g = renderSys.newGeom();
	g->vertex[0].set(viewarea.min.x, viewarea.min.y);
	g->vertex[1].set(viewarea.min.x, viewarea.max.y);
	g->vertex[2].set(viewarea.max.x, viewarea.max.y);
	g->vertex[3].set(viewarea.max.x, viewarea.min.y);
	g->texcoord[0].set(texbox.min.x, texbox.min.y);
	g->texcoord[1].set(texbox.min.x, texbox.max.y);
	g->texcoord[2].set(texbox.max.x, texbox.max.y);
	g->texcoord[3].set(texbox.max.x, texbox.min.y);
	g->depth = backgroundDepth;
	g->colormap = tex->tex.getTextureID();
	g->bumpmap = tex->bumpmap.getTextureID();
	g->color.set(255,255,255,255);
	g->shader = 0;						//FIXMEPLZ
	g->blend = Blend();
	g->use_alpha_test = false;
	g->use_bump_mapping = true;
	g->use_lighting = true;
	g->postFBO = false;
	g->dir.set(1,0);
	g->flip = 1;
	renderSys.geomDone(g);
}

void WorldSystemImpl::displayLayer(int layer, const box2f &viewarea) {
	if (!drawForeground) return;
	if (!world) return;
	world->displayLayer(layer, viewarea);
}

	//this is a mess.  clean it up.
	//the editor no longer gets to hold the script.  give it to the program.
	//actually lets just have world.cpp keep track of the script filename
	//and then have the loader read it & compile it upon . . . level change? session init?
	//i need to work on separating compiling & executing . . .

bool WorldSystemImpl::getScriptFileName(int index, char *buffer, int len) {
	if (!world) return false;
	if (index < 0 || index >= world->scriptList.size()) return false;
	buffer[0] = '\0';	//why wouldnt' strncpy do this with a null-term'd string?
	strncpy(buffer, world->scriptList[index].c_str(), len);
	return true;
}

bool WorldSystemImpl::newWorld() {
	World *newworld = new World();
	if (!newworld) return false;
	
	//this line usually goes on after finding the file but before reading its spawn info in
	//that means if the world load fails halfway through then the spawnobj vector of the old world is lost
	//also it being here and in LoadWorld is redundant
	//another reason why we need to merge the spawn system into the world
	spawnObjSys.newSystem();
	
	if (world) delete world;
	world = newworld;
	editor.setWorld(world);
	return true;
}

bool WorldSystemImpl::loadWorld(const char *filename) {
	World *newworld = LoadWorld(filename);
	if (!newworld) return false;
	if (world) delete world;
	world = newworld;
	editor.setWorld(world);
	return true;
}

bool WorldSystemImpl::saveWorld() { 
	assert(world); 
	return world->save();
}

bool WorldSystemImpl::saveWorldAs(const char *filename) { 
	assert(world); 
	return world->saveAs(filename);
}

World *WorldSystemImpl::getWorld() { return world; }

void WorldSystemImpl::setInitWorld(const std::string &initWorld_) { initWorld = initWorld_; }
const std::string &WorldSystemImpl::getInitWorld() { return initWorld; }

bool WorldSystemImpl::loadInitWorld() {
	return loadWorld(initWorld.c_str());
}


WorldSystem::WorldSystem() : impl(new WorldSystemImpl()) {}
WorldSystem::~WorldSystem() { delete impl; }
bool WorldSystem::sysInit() { return impl->sysInit(); }
void WorldSystem::sysKill() { impl->sysKill(); }
bool WorldSystem::gameInit() { return impl->gameInit(); }
void WorldSystem::gameKill() { impl->gameKill(); }
bool WorldSystem::sessionInit() { return impl->sessionInit(); }
void WorldSystem::sessionKill() { impl->sessionKill(); }
void WorldSystem::update() { impl->update(); }
void WorldSystem::displayBackground(MapTagType *tt, const box2f &viewarea) { impl->displayBackground(tt, viewarea); }
void WorldSystem::displayLayer(int index, const box2f &viewarea) { impl->displayLayer(index, viewarea); }
bool WorldSystem::getScriptFileName(int index, char *buffer, int len) { return impl->getScriptFileName(index, buffer, len); }
bool WorldSystem::newWorld() { return impl->newWorld(); }
bool WorldSystem::loadWorld(const char *filename) { return impl->loadWorld(filename); }
bool WorldSystem::saveWorld() { return impl->saveWorld(); }
bool WorldSystem::saveWorldAs(const char *filename) { return impl->saveWorldAs(filename); }
void WorldSystem::setInitWorld(const std::string &initWorld) { impl->setInitWorld(initWorld); }
const std::string &WorldSystem::getInitWorld() { return impl->getInitWorld(); }
bool WorldSystem::loadInitWorld() { return impl->loadInitWorld(); }
World *WorldSystem::getWorld() { return impl->getWorld(); }
WorldSystem worldSys;


