require 'lolo.scripts.pushable'
require 'lolo.scripts.puff'	-- for when eggs or eggables die

animLoad('egg', {
	sprites = {
		{file='egg.png'};
		{file='egg_crack.png'};
		{file='egg_sink.png'};
	};
	seqs = {
		idle = {{joint = 1; time = 0; sprite = 1; }};
		crack = {{joint = 1; time = 0; sprite = 2; }};
		sink = {{joint = 1; time = 0; sprite = 3; }};
	};
})

entclass('Egg', Pushable)

Egg.animGroup = 'egg'
Egg.speed = 10	-- Player.speed

-- states by default consist of 'enter', 'think' and 'leave'
-- if omitted it's cool
class('State')

-- this is a shortcut for making states to call callbacks after a certain elapsed time
class('DelayState', State)
function DelayState:ctor(obj)
	union(self,obj)	-- copy the time and fn params over
	-- now if there is an enter/think state then it's gonna override our own delaystate ...
	for i,v in ipairs{'enter','think'} do
		local original = obj[v]
		if original then
			self[v] = function(self)
				DelayState.enter(self)
				original(self)
			end
		end
	end
end
function DelayState:enter()
	-- i hope this 'self' is for self:enter's function and not for DelayState:ctor's ...
	self.stateNextThink = getTime() + self.state.time		-- assuming self.state is the current state
end;
function DelayState:think()
	if getTime() > self.stateNextThink then
		self.state.fn(self)
	end
end;

Egg.STATE = {
	CRACKING = DelayState:new{
		time = 3,
		fn = function(self)
			self:setAnimSeq('crack')
			self:setState(self.STATE.HATCHING)
		end
	};
	HATCHING = DelayState:new{
		time = 2,
		fn = function(self)
			self.owner:turnUnEgg()
			self:delete()
		end
	};
	GONE = DelayState:new{
		enter = function(self)
			-- reposition the eggee to a hidden location
			self:setPos{-10,-10}
			-- and hide him too? seems a little repetative ...
			self:setAnimGroup('null')
		end;
		
		time = 5,
		fn = function(self)
			-- restore the owner to its original position
			-- since then owner is gonna read from us, set our position where they should get it from
			self:setPos(self.owner.startPos)
			self.owner:turnUnEgg()
			self:delete()
		end
	};
	SINKING = DelayState:new{
		-- mind you DelayState.enter was just called when we get here
		-- so we know the self.stateNextThink is already set to getTime() plus that time below
		-- i'm going to make an override possible from our owning entity
		enter = function(self)
			if self.owner.eggSinkTime then
				self.stateNextThink = getTime() + self.owner.eggSinkTime
			end
			self.isBlocking = false	--usually true, but for the time that we're in the water lets not
			self.isBlockingPushers = false
		end;
		leave = function(self)
			self.isBlocking = true	--reset
			self.isBlockingPushers = true
		end;

		time = 2;
		fn = function(self)
			-- oh if the player is standing on us then kill them
			-- maybe do that in their think fn

			-- and don't forget to respawn the eggee in a few short moments
			self:setState(self.STATE.GONE)
		end;
	};
}

function Egg:ctor(owner)
	Egg.super.ctor(self)
	
	-- link to the owner
	self.owner = owner
	self:setPos(owner:getPos())

	-- my state machine
	self:setState(self.STATE.CRACKING)

	-- link back to ourselves
	self.owner:turnEgg(self)
end

function Egg:setState(newState)

	-- if a state calls setState on another then
	-- its hard to set stateStartTime without it getting messed up by the enter/leave functions
	-- one solution is to set stateStartTime first so all the state funcs will always happen after it
	self.stateStartTime = getTime()

	if self.state and self.state.leave then self.state.leave(self) end
	self.state = newState
	if self.state and self.state.enter then self.state.enter(self) end
end

function Egg:think()
	Egg.super.think(self)
	if not self:isActive() then return end
	
	-- if we're in the cracking or hatching state (both pushable)
	if self.state == self.STATE.CRACKING or self.state == self.STATE.HATCHING then
		-- now that the movement has run
		-- see if we're standing still
		if not self.moveFrac then
			-- if we are then see if we're standing on water
			-- on all fours
			local typeUL = mapGetType(self:getPosX() - .25, self:getPosY() - .25, 0)
			local typeUR = mapGetType(self:getPosX() + .25, self:getPosY() - .25, 0)
			local typeLL = mapGetType(self:getPosX() - .25, self:getPosY() + .25, 0)
			local typeLR = mapGetType(self:getPosX() + .25, self:getPosY() + .25, 0)

			-- the egg-under-me hack ...
			for e in allents() do
				if e:isa(Egg) and e.state == e.STATE.SINKING and e ~= self then
					-- if any of our corners are really standing on an egg rather than water then treat it like its empty
					if typeUL == MAPTYPE.Water and (e:getPos() - (self:getPos() + vec2:new(-.25,-.25))):lenLInf() < 0.5 then typeUL = MAPTYPE.Empty end
					if typeUR == MAPTYPE.Water and (e:getPos() - (self:getPos() + vec2:new( .25,-.25))):lenLInf() < 0.5 then typeUR = MAPTYPE.Empty end
					if typeLL == MAPTYPE.Water and (e:getPos() - (self:getPos() + vec2:new(-.25, .25))):lenLInf() < 0.5 then typeLL = MAPTYPE.Empty end
					if typeLR == MAPTYPE.Water and (e:getPos() - (self:getPos() + vec2:new( .25, .25))):lenLInf() < 0.5 then typeLR = MAPTYPE.Empty end
				end
			end
					
			if typeUL == MAPTYPE.Water and
				typeUR == MAPTYPE.Water and
				typeLL == MAPTYPE.Water and
				typeLR == MAPTYPE.Water
			then
				-- todo here - see if there's another egg under us =)
				-- and if there is, only then, then we don't sink... egads ...

				self:setAnimSeq('sink')
				self.isBlocking = false
				self:setState(self.STATE.SINKING)
				
				-- so now in the player's movement
				-- override the normal 'can i move' code
				-- to check and see if there's water OR a sinking egg under us
			end
		end
	end
	
	if self.state then self.state.think(self) end
end

-- static function me?
function Egg:cannotPassThru(maptype)
	return maptype ~= MAPTYPE.Empty and maptype ~= MAPTYPE.Water
end

function Egg:hitByShot(shot, shotPos, shotDir)
	-- now if an egg is hit by a shot
	-- then we go
	-- and our egg'd owner reappears at their original location after a while ...
	-- somewhat messy ...
	
	-- just relocate off-map somewhere
	self.isBlocking = false
	
	-- before we reposition lets send out some smoke
	Puff(self:getPos())

	-- if an egg is in the middle of movement
	-- and it gets hit by a shot then its think will try to tweak its position

	self:setState(self.STATE.GONE)

	-- remove the shot and stop the collision
	shot:delete()
	return self.HIT_RESPONSE.STOP
end

function Egg:onChestOpen()
	Puff(self:getPos())
	self:delete()
end

-- this is the parent class to all eggable objects
-- basically an interface but relatively solid so we'll use it in the class hierarchy


entclass('Eggable', LoloBase)

function Eggable:ctor(so)
	Eggable.super.ctor(self,so)
	if so.sinktime then self.eggSinkTime = castnumber(so.sinktime) end
end

function Eggable:think()
	-- if we have an attached egg then skip our think
	-- that's right, no collision, no acting, no nothing
	if self.eggableEgg then return end

	Eggable.super.think(self)
end

function Eggable:turnEgg(egg)
	self.eggableEgg = egg
	-- store push
	self.eggablePos = vec2:new(self:getPos())
	self.eggableBlocked = self.isBlocked
	-- set hidden state
	self:setPos{-10,-10}
	self.isBlocked = false
end

function Eggable:turnUnEgg()
	-- restore
	self:setPos(self.eggableEgg:getPos())
	self.isBlocked = self.eggableBlocked
	-- clear push
	self.eggablePos = nil
	self.eggableBlocked = nil
	-- detach
	self.eggableEgg = nil
end

function Eggable:hitByShot(shot, shotPos, shotDir)
	-- turn ourselves into an egg...
	Egg:new(self)
	-- remove the shot and stop the collision
	shot:delete()
	return self.HIT_RESPONSE.STOP
end

function Eggable:onChestOpen()
	Puff(self:getPos())
	self:delete()
end
