require 'lolo.scripts.base'

-- for the entities that actually do the moving
-- circular def: lots of base functions are dependant upon this class
entclass('LoloMove', LoloBase)

-- default speed anyone? one tile per second
LoloMove.speed = 1

-- responses to doMove
-- NO_MOVE means the command was bad
-- WAS_BLOCKED means something hit us
-- DID_MOVE means go
LoloMove.MOVE_RESPONSE = enum({'NO_MOVE', 'WAS_BLOCKED', 'DID_MOVE'})

-- in the base movable hitWorld have a flag that delegates whether this entity can walk on eggs in water
-- by default lets have it on?
-- heck, just have it for everyone
function LoloMove:hitWorld(pos, typeUL, typeUR, typeLL, typeLR)
	-- test each corner
	-- if theres a sinking egg within .5 (l-inf) of a water corner then count it as empty

	for e in allents() do
		if e:isa(Egg) and e.state == e.STATE.SINKING and e ~= self then
			-- if any of our corners are really standing on an egg rather than water then treat it like its empty
			if typeUL == MAPTYPE.Water and (e:getPos() - (pos + vec2:new(-.25,-.25))):lenLInf() < 0.5 then typeUL = MAPTYPE.Empty end
			if typeUR == MAPTYPE.Water and (e:getPos() - (pos + vec2:new( .25,-.25))):lenLInf() < 0.5 then typeUR = MAPTYPE.Empty end
			if typeLL == MAPTYPE.Water and (e:getPos() - (pos + vec2:new(-.25, .25))):lenLInf() < 0.5 then typeLL = MAPTYPE.Empty end
			if typeLR == MAPTYPE.Water and (e:getPos() - (pos + vec2:new( .25, .25))):lenLInf() < 0.5 then typeLR = MAPTYPE.Empty end
		end
	end
	
	return LoloMove.super.hitWorld(self,pos,typeUL,typeUR,typeLL,typeLR)
end

-- returns false if we have no valid move cmd
-- true if we got a vlid move cmd (regardless of whether we moved)
-- possible returns:
--  no move
--  was blocked
--  did move
function LoloMove:doMove(moveCmd)
	if not moveCmd then return self.MOVE_RESPONSE.NO_MOVE end

	-- idunno why i have animation stuff here inside the doMove routine
	-- it assumes a bit about the child class
	-- maybe separate this into a 'walkable' child class?
	-- pass it to the Player and whoever else needs it
	-- because blocks dont walk, but they do move
	--
	if moveCmd == SIDE.LEFT then self:setLeft(true) end
	if moveCmd == SIDE.RIGHT then self:setLeft(false) end

	local newDest
	
	if moveCmd == SIDE.UP then			newDest = vec2:new(self:getPos()):addTo(0,-.5)
	elseif moveCmd == SIDE.DOWN then	newDest = vec2:new(self:getPos()):addTo(0,.5)
	elseif moveCmd == SIDE.LEFT then	newDest = vec2:new(self:getPos()):addTo(-.5,0)
	elseif moveCmd == SIDE.RIGHT then	newDest = vec2:new(self:getPos()):addTo(.5,0)
	end

	-- we were told to go somewhere
	if not newDest then return self.MOVE_RESPONSE.NO_MOVE end

	local blocked = false
	-- check it for collision
	-- local newBox = self.bbox + newDest
	-- now check collision along the side we're moving
	-- or just do two tests off the center

	-- check map bounds
	if newDest[1] < 0.4 or
		newDest[2] < 0.4 or
		newDest[1] > mapGetWidth() - 0.4 or
		newDest[2] > mapGetHeight() - 0.4
	then
		if self:hitEdge(newDest) then
			blocked = true
if self:isa(Egg) then print('blocked by world') end
		end
	end

	-- subtract .25 from the new dest
	-- then step .5 in either dir and map type test
	if not blocked then
		local typeUL = mapGetType(newDest[1] - .25, newDest[2] - .25, 0)
		local typeUR = mapGetType(newDest[1] + .25, newDest[2] - .25, 0)
		local typeLL = mapGetType(newDest[1] - .25, newDest[2] + .25, 0)
		local typeLR = mapGetType(newDest[1] + .25, newDest[2] + .25, 0)
		
		-- if our callback says 'stop at the wall' then do so
		if self:hitWorld(newDest, typeUL, typeUR, typeLL, typeLR) then
			blocked = true
		end
	end
	
	-- still not hit? check for objects ...
	if not blocked then
		for i,e in pairs(_ents) do
			if e ~= self and
				e:isa(LoloBase) and
				(e.bbox + e.destPos):touches(self.bbox + newDest)
			then
				-- see what happens when we hit it
				-- return true if we can't move into it
				local response = self:hitObject(e, newDest, moveCmd)
				if response == self.HIT_RESPONSE.STOP then
					-- we immediately fall out of the method on this case
					-- which is good, cuz the ent just might die
					blocked = true
					break
				end
				
				if response == self.HIT_RESPONSE.TEST_OBJECT then
					-- see if it can be hit
					if e:startPush(self, newDest, moveCmd) then
						blocked = true	-- no more tests if we cant move
						break
					end
				end
			end
		end
	end
	
	-- now some objects are solid
	-- if you step halfway on them then you're blocked
	-- other objects (hearts) are step-on-able
	-- only if you're right over them do you get them (equal discretized positions?)
	-- now hearts(park dollarsign) do not trigger until the movement phase is over
	-- blocks will as soon as you start (assuming they can be pushed)
	-- and they need to know where you're coming from ...

	-- if we hit something then see if it can be stepped on or pushed
	-- if stepped on then go for that, pushed, likewise
	-- execute via OOP stuff
	-- if so then clear the moveFrac
	if blocked then
		self.moveFrac = nil
		return self.MOVE_RESPONSE.WAS_BLOCKED
	end
	-- otherwise continue

	self.destPos:set(newDest)
	self.moveFrac = 0
	-- snap the dest
	self.destPos:mulTo(2):floor():mulTo(0.5)

	-- store the source
	self.srcPos:set(self:getPos())
	return self.MOVE_RESPONSE.DID_MOVE
end

function LoloMove:think()
	LoloMove.super.think(self)
	if not self:isActive() then return end

	if not self.moveFrac then
		-- do our move
		-- even if its nil, pass moveCmd
		self:doMove(self.moveCmd)
	else
		self.moveFrac = self.moveFrac + getDeltaTime() * self.speed
		if self.moveFrac >= 1 then
			self.moveFrac = nil
			self:setPos(self.destPos)

			-- see if we ran over anything
			for i,e in pairs(_ents) do
				if e ~= self
				and e:isa(LoloBase)
				and (e.bbox + e.destPos):touches(self.bbox + self.destPos)
				then
					-- see if it can be stepped on
					e:endPush(self, self.destPos)
				end
			end
		else
			local v = vec2:new():lerp(self.srcPos, self.destPos, self.moveFrac)
			-- directly set the entity position. otherwise we'll be calling the LoloBase setpos
			-- makes me think maybe we should be calling LoloBase.setPos something else ...
			Entity.setPos(self, v)
		end
	end
end
